
import Expector
import Extractions
import time
from scipy import linalg
from VTKHelperFunctions import *
from LineIdentify import *
from basics import *
from MLab import diff
from FITS import *
import sys
import TwistedClient

INITIAL_COMMAND = join(sys.argv[1:])

CHIPKEYS = ["chip", "slmin", "slmax", "start", "end", "npixels"]
COEFKEYS = ["coef_disp", "coef_xdisp", "coef_sag", "coef_tilt", "coef_sagit", "coef_slen"]
FLIPOFF = 1.5

def polynom(x,o):
   x = asarray(x,Float64)
   p = zeros((o+1,)+x.shape,Float64)
   for j in range(o+1): p[j] = power(x,j)
   return p

class SMF(DictType):
        SMFKEYS = ["type","id","ra","dec","x","y"]
        fn = [str,str,str,str,float,float]
        def ReadFile(self,smf_file):
            for key in self.SMFKEYS: self[key] = []
            f = open(smf_file,"r")
            for l in f:
               data = l.split()
               if (data[0] == "HOLE" and data[7] == "1") or (data[0] == "SLIT"):
                  [self[key].append(self.fn[self.SMFKEYS.index(key)](data[self.SMFKEYS.index(key)])) for key in self.SMFKEYS]
            f.close()
            for key in self.SMFKEYS[4:]: self[key] = array(self[key])
        def GetObject(self,object):
            i = self["id"].index(object)
            vals = [self[key][i] for key in self.SMFKEYS]
            return vals

class Object(DictType):
        def __repr__(self): return self["id"]
        def SetMapping(self,mapping):
            self.mapping = mapping
            self.dewar = self.mapping.dewar
            self.nodshuffle = self.mapping.xdisp
            if "LDSS" in self.dewar: self.ldss = True
            else: self.ldss = False
            self.woffset = 0.0
        def GetPixels(self,wavelens,verb=0,usemid=0,trace=None):
            self.c = []
            self.w = []
            self.x = []
            self.y = []
            self.y0 = []
            self.y1 = []
            self.x0 = []
            self.x1 = []
            for c in range(self["nobs"]):
               if verb: print wavelens,self["start"][c],self["end"][c]
               use = between(wavelens,self["start"][c],self["end"][c],inc=1)
               if verb: print use
               if sum(use):
                   w = compress(use,wavelens)
                   if trace:
                     w0,w1 = self.mapping.wavelens
                     wspan = (w1-w0)/2.0
                     wmidd = (w1+w0)/2.0
                     delta = -dot(trace,legendre((w-wmidd)/wspan,len(trace)-1))
                     print mean(delta)
                   else:
                     delta = 0.0
                   xcen = add.reduce(self["coef_disp"][c][::,NewAxis]*polynom(w,self.mapping.ord_disp),0)
                   ycen = add.reduce(self["coef_sag"][c][::,NewAxis]*polynom(w,self.mapping.ord_sag),0)
                   tilt = add.reduce(self["coef_tilt"][c][::,NewAxis]*polynom(w,self.mapping.ord_tilt),0)
                   slnt = add.reduce(self["coef_slen"][c][::,NewAxis]*polynom(w,self.mapping.ord_len),0)
                   sagt = add.reduce(self["coef_sagit"][c][::,NewAxis]*polynom(w,self.mapping.ord_sagit),0)
                   sagsn = where(sagt>=0.0,1.0,-1)
                   sagt = abs(sagt)
                   rag = slnt*slnt/(8.0*sagt)
                   ycen = ycen + delta
                   slength = self["slmax"][c] - self["slmin"][c]
                   scatl = slnt/slength
                   slmx = self["slmax"][c]*scatl
                   slmn = self["slmin"][c]*scatl
                   ymid = ycen+(slmx+slmn)/2.-delta
                   y0 = ymid-0.5*slnt
                   y1 = ymid+0.5*slnt
                   x0 = xcen + tilt*slmn/slnt
                   x1 = xcen + tilt*slmx/slnt
                   self.c.append(self["chip"][c]*ones(w.shape))
                   self.w.append(w)
                   self.x.append( xcen)
                   if usemid: self.y.append( ymid)
                   else: self.y.append( ycen)
                   self.x0.append( x0)
                   self.x1.append( x1)
                   self.y0.append( y0)
                   self.y1.append( y1)
                   self.xcen = xcen
                   self.ymid = ymid
            if len(self.c):
               self.c = concatenate(self.c)
               self.w = concatenate(self.w)
               self.x = concatenate(self.x)-1
               self.y = concatenate(self.y)-1
               self.y0 = concatenate(self.y0)-1
               self.y1 = concatenate(self.y1)-1
               self.x0 = concatenate(self.x0)-1
               self.x1 = concatenate(self.x1)-1
               if self.nodshuffle:
                 self.y, self.x, self.x0, self.x1, self.y0, self.y1 = \
                 self.x, self.y, self.y0, self.y1, self.x0, self.x1
        def GetGrid(self,c,wavelens=None,verb=0,trace=None):
            c0,x0,x1,y0,y1 = self.corners
            xmin,xmax = extrema([x0,x1])
            ymin,ymax = extrema([y0,y1])
            xmin,ymin = array([xmin,ymin]).astype(Float32)
            xmax,ymax = array([xmax+1,ymax+1]).astype(Float32)
            if wavelens:
              nw = len(wavelens)
            else:
              nw = int(xmax-xmin+1)/2
              dw = (self["end"][c] - self["start"][c])/nw
              wavelens = arange(self["start"][c],self["end"][c]+dw/2.0,dw,Float64)
            if trace:
              w0,w1 = self.mapping.wavelens
              wspan = (w1-w0)/2.0
              wmidd = (w1+w0)/2.0
              delta = -dot(trace,legendre((wavelens-wmidd)/wspan,len(trace)-1))
            else:
              delta = 0.0
            xcen = add.reduce(self["coef_disp"][c][::,NewAxis]*polynom(wavelens,self.mapping.ord_disp),0)
            ycen = add.reduce(self["coef_sag" ][c][::,NewAxis]*polynom(wavelens,self.mapping.ord_sag),0)
            slnt = add.reduce(self["coef_slen"][c][::,NewAxis]*polynom(wavelens,self.mapping.ord_len),0)
            tilt = add.reduce(self["coef_tilt"][c][::,NewAxis]*polynom(wavelens,self.mapping.ord_tilt),0)
            slit_len = abs(self["slmax"][c] - self["slmin"][c])
            slmx = self["slmax"][c]*slnt/slit_len
            slmn = self["slmin"][c]*slnt/slit_len
            scatl = slnt/slit_len
            ycen = ycen + delta
            if c0 > 4:
               ymid = ycen-(slmx+slmn)/2. - delta
               if self.nodshuffle:
                 yzero = (ymax-xcen)
               else:
                 yzero = (ymax-ycen)
               xa = yzero[NewAxis,::] - (0+arange(self.shape[0]))[::,NewAxis]
               if self.nodshuffle:
                 self.yloc = ymid[-1] - xmin
                 self.yob = ycen[-1] - xmin
               else:
                 self.yloc = ymax - ymid[0]
                 self.yob = ymax - ycen[0]
            else:
               ymid = ycen+(slmx+slmn)/2. - delta
               if self.nodshuffle:
                 yzero = (ycen-xmin)
               else:
                 yzero = (ycen-ymin)
               xa = (0+arange(self.shape[0]))[::,NewAxis] - yzero[NewAxis,::]
               if self.nodshuffle:
                  self.yloc = xmax - ymid[0]
                  self.yob = xmax - ycen[0]
               else:
                  self.yloc = ymid[-1] - ymin
                  self.yob = ycen[-1] - ymin
            slitpos = scatl[NewAxis,::] * xa
            c1a = self.yloc - slit_len/2.0 + 1
            c1b = self.yloc + slit_len/2.0 + 1
            if c0 > 4:
               if self.nodshuffle:
                 ypos = (ycen[-1] + slitpos) - xmin
               else:
                 ypos = ymax - (ycen[0] + slitpos)
            else:
               if self.nodshuffle:
                 ypos = xmax - ycen[0] + slitpos
               else:
                 ypos = ycen[-1] + slitpos - ymin
            if self.nodshuffle:
               xpos = (xcen-ymin) + tilt*slitpos/slnt
            else:
               xpos = (xcen-xmin) + tilt*slitpos/slnt
            xmid = (xcen - (slmx+slmn)/2.*tilt/slnt)
            if c0 > 4:
                if self.nodshuffle:
                   ypos = ypos - (xmax-xmin)
                   xpos = (ymax-ymin) - xpos
                   xmid = (ymax-ymin) - xmid
                else:
                   ypos = (ymax-ymin) - ypos
                   xpos = (xmax-xmin) - xpos
                   xmid = (xmax-xmin) - xmid
                yref = ypos[::,0]
            else:
                if self.nodshuffle:
                   ypos = (xmax-xmin) - ypos
                yref = ypos[::,-1]
            xref = xpos[xpos.shape[0]/2,NewAxis,::]
            txpos = 1*transpose(xpos)
            typos = 1*transpose(ypos)
#            qdump("typos.fits",typos)
#            qdump("txpos.fits",txpos)
#            exit()
            dy = (yref - typos)
            if c0 > 4 or self.ldss:
               dx = (xref - xpos)
            else:
               dx = (xpos - xref)
            ydist = Expector.fitmap(2,self.mapping.ord_sag,typos.flat,txpos.flat,dy.flat,(self.shape[1],self.shape[0]))
            yfit  = Expector.getmap(2,self.mapping.ord_sag,ydist,typos.flat,txpos.flat,(self.shape[1],self.shape[0]))
            yfit = reshape(yfit,dy.shape)
            yrms = rms(yfit.flat-dy.flat*self.shape[0])
            self.yinfo = [[ydist,yrms],2,self.mapping.ord_sag]
            ybor,xbor = indices(xpos.shape,Float32)
            xrect = Expector.fitmap(2,self.mapping.ord_tilt,xpos.flat,ybor.flat,dx.flat,(self.shape[0],self.shape[1]))
            xfit  = Expector.getmap(2,self.mapping.ord_tilt,xrect,xpos.flat,ybor.flat,(self.shape[0],self.shape[1]))
            xfit = reshape(xfit,dx.shape)
            xrms = rms(xfit.flat-dx.flat*self.shape[1])
            self.xinfo = [[xrect,xrms],2,self.mapping.ord_tilt]
            xco = xcen
            wcen = add.reduce(self["coef_xdisp"][c][::,NewAxis]*polynom(xco,self.mapping.ord_disp),0)
            if self.woffset:
              wcen = wcen + self.woffset[0]
            if c0 > 4 or self.ldss: xco = xmax-xco
            xc, xw = (xco[0]+xco[-1])/2.0, diff(extrema(xco))[0]/2.0
            wbasis = legendre((xco-0.5-xc)/xw,self.mapping.ord_disp)
            wsol = svdfit(transpose(wbasis),wcen)
            wrms = rms(wcen-add.reduce(wsol[::,NewAxis]*wbasis,0))
            self.winfo = [[wsol,wrms],self.mapping.ord_disp,0]
        def GetCorners(self,chip,verb=0):
            c0 = self["chip"][chip]
            if verb: print "Defining area for %s in chip %d." % (self["id"],c0)
            self.GetPixels(array([self["start"][chip],self["end"][chip]]),verb=0)
            if len(self.x):
               x0,x1 = extrema(concatenate([self.x0,self.x1]))
               y0,y1 = extrema(concatenate([self.y0,self.y1]))
               return c0,x0,x1,y0,y1
            else: return None

        def SetY(self,chip,pos_error=None,ref=0,maskx=0,verb=0):
            if verb: print "Wavelengths range from %.3f to %.3f." % (self["start"][chip], self["end"][chip])
            self.corners = self.GetCorners(chip,verb=verb)
            if not self.corners: return
            c0,x0,x1,y0,y1 = self.corners
            xmin,xmax = extrema([x0,x1])
            ymin,ymax = extrema([y0,y1])
            xmin,ymin = array([xmin,ymin]).astype(Float32)
            xmax,ymax = array([xmax+1,ymax+1]).astype(Float32)
            if verb: print "x0=%8.2f x1=%8.2f y0=%8.2f y1=%8.2f" % (x0,x1,y0,y1)
            nw = 10*self.mapping.ord_disp+1
            dw = (self["end"][chip] - self["start"][chip])/nw
            wavelens = arange(self["start"][chip],self["end"][chip]+dw/2.0,dw,Float64)
            xcen = add.reduce(self["coef_disp"][chip][::,NewAxis]*polynom(wavelens,self.mapping.ord_disp),0)
            ycen = add.reduce(self["coef_sag" ][chip][::,NewAxis]*polynom(wavelens,self.mapping.ord_sag),0)
            ycen = [ycen-ymin,ymax-ycen][c0>4]
            y = [ycen[-1],ycen[0]][c0 > 4]
            if ref and 0:
               off = self.yoff
            else:
               if pos_error:
                  if c0 > 4: off = pos_error[1]
                  else: off = pos_error[0]
                  off = off + sum([pow(maskx,j-1)*pos_error[j] for j in range(2,len(pos_error))])
               else: off = 0.0
            self["yap"][chip] = y + off
            if verb: print "Using offset of %.1f for chip %d ---> final position = %.1f" % (off,c0,self["yap"][chip])

        def GetGain(self,images,chip,smf=None,verb=0):
            if verb: print "Wavelengths range from %.3f to %.3f." % (self["start"][chip], self["end"][chip])
            self.corners = self.GetCorners(chip,verb=verb)
            if not self.corners: return
            c0,x0,x1,y0,y1 = self.corners
            if verb: print "x0=%8.2f x1=%8.2f y0=%8.2f y1=%8.2f" % (x0,x1,y0,y1)
            use = []
            for image in images:
                im = image+"c%d.fits" % (c0)
                if verb: print "Opening %s." % (im)
                F = FITS(im)
                section = concatenate([clip(extrema([y0,y1]),1,F["naxis1"]),clip(extrema([x0,x1]),1,F["naxis2"])])
                section = concatenate([section[::2], section[1::2]+1]).astype(Int32)
                data = F.readsec(section[:2],section[2:])
                data = transpose(data)
                data = data.astype(Float32)
                F.close()
                use.append(data.flat)
            s1,s2 = use
            e = s1-s2
            fl = sum(s1**2)
            top = mean(s1) + mean(s2)
            bot = 1.49*mad(e)
#            bot = rms(e)
#            bot = bwt(e)[1]
            gain = top/bot**2
            return gain

        def GetRelPos(self,image,chip,verb=0,interact=0):
            if verb: print "Wavelengths range from %.3f to %.3f." % (self["start"][chip], self["end"][chip])
            self.corners = self.GetCorners(chip,verb=verb)
            if not self.corners: return -999
            c0,x0,x1,y0,y1 = self.corners
            xmin,xmax = extrema([x0,x1])
            ymin,ymax = extrema([y0,y1])
            xmin,ymin = array([xmin,ymin]).astype(Float32)
            xmax,ymax = array([xmax+1,ymax+1]).astype(Float32)
            if verb: print "x0=%8.2f x1=%8.2f y0=%8.2f y1=%8.2f" % (x0,x1,y0,y1)
            im = image+"c%d.fits" % (c0)
            if verb: print "Opening %s." % (im)
            F = FITS(im)
            section = concatenate([clip(extrema([y0,y1]),1,F["naxis1"]),clip(extrema([x0,x1]),1,F["naxis2"])])
            section = concatenate([section[::2], section[1::2]+1]).astype(Int32)
            data = F.readsec(section[:2],section[2:])
            data = transpose(data).astype(Float32)
            if c0 > 4: data = 1*data[::-1,::-1]
            F.close()
            self.shape = data.shape
            self.GetGrid(chip,verb=verb)
            ymap = Expector.getmapping(self.yinfo[1],self.yinfo[2],self.yinfo[0][0],data,ydist=1,verb=0)
            data = Expector.rebin(data,0*ymap,ymap,data=0,cubic=0)
#            qdump("data.fits",data)
            nw = 10*self.mapping.ord_disp+1
            dw = (self["end"][chip] - self["start"][chip])/nw
            wavelens = arange(self["start"][chip],self["end"][chip]+dw/2.0,dw,Float64)
            xcen = add.reduce(self["coef_disp"][chip][::,NewAxis]*polynom(wavelens,self.mapping.ord_disp),0)
            ycen = add.reduce(self["coef_sag" ][chip][::,NewAxis]*polynom(wavelens,self.mapping.ord_sag),0)
            ycen = [ycen-ymin,ymax-ycen][c0>4]
            data = VTKMedian(data,101,3,numret=1)
            yco,xco = indices(data.shape,Float32)
            ndata = add.reduce(greater(data,0),0)
            gdata = greater(ndata,getpct(ndata,95)-2)
            data = data * gdata
            ypos = sum(yco.flat*data.flat)/sum(data.flat)
            prof = add.reduce(data,1)
            xa = arange(prof.shape[0])
            x,p = compress(prof,[xa,prof],1)
            if len(p) < 10: return -999
            d = diff(p)
            b = take(p,argsort(abs(d)))[len(p)/10]
            x,p = compress(greater(p,b),[x,p],1)
            if len(p) < 10: return -999
            params = max(p), median(ycen), 0.3, sort(p)[len(p)/10]
            sol = leastsq(gaussian,params,args=(x,p,1+0*getsqrt(p)),full_output=1)
            self.ypos = sol[0][1]
            f = gaussian(sol[0],x,p)
            if verb: print
            self.ypred = [ycen[-1],ycen[0]][c0 > 4]
            self.yoff = self.ypos - self.ypred
            print "ID,C,YPRED,YPOS = (%s,%3d,%8.2f,%8.2f)" % (self["id"],c0,self.ypred,self.ypos)
            if interact:
                qdump("data.fits",data,verb=0)
                verbex("dfits -nj data.fits")
            return self.yoff

        def GetImages(self,images,suffices,chip,linelist=None,subdir=None,smf=None,verb=0,write=0):
            if verb: print "Wavelengths range from %.3f to %.3f." % (self["start"][chip], self["end"][chip])
            self.corners = self.GetCorners(chip,verb=verb)
            if not self.corners: return
            c0,x0,x1,y0,y1 = self.corners
            if verb: print "x0=%8.2f x1=%8.2f y0=%8.2f y1=%8.2f" % (x0,x1,y0,y1)
            datum = smf.GetObject(self["id"])
            for image in images:
                if suffices: suffix = suffices[images.index(image)]
                else: suffix = ""
                im = image+"_c%d.fits" % (c0)
                if verb: print "Opening %s." % (im)
                F = FITS(im)
                section = concatenate([clip(extrema([y0,y1]),1,F["naxis1"]),clip(extrema([x0,x1]),1,F["naxis2"])])
                section = concatenate([section[::2], section[1::2]+1]).astype(Int32)
                data = F.readsec(section[:2],section[2:])
                if not self.nodshuffle:
                   data = transpose(data)
                if c0 > 4: data = data[::-1,::-1]
                data = data.astype(Float32)
                if "flat" in suffix:
                   zpr = add.reduce(data,1)/data.shape[1]
                   lz = nonzero(less(zpr,0))
                   if len(lz) > 5:
                       print "Fixing Hole in the Flat!"
                       lpr = take(data,lz,0)
                       for jz in lz: data[jz] = (data[jz] - min(data[jz])).astype(Float32)
                       zpr = add.reduce(data,1)/data.shape[1]
                       lpr = take(data,lz,0)
                       mpr = mean(lpr)
                       zof = median(concatenate([(zpr[lz[0]-1]-mpr), (zpr[lz[-1]+1]-mpr)]))
                       for jz in lz: data[jz] = (data[jz] + zof).astype(Float32)
                onaxes = list(data.shape)
                onaxes.reverse()
                if write:
                   output = "%s_%s_c%d.fits" % (self["id"],suffix,c0)
                   if subdir: output = "%s/%s" % (subdir,output)
                self.shape = data.shape
                if image == images[0]:
                   self.GetGrid(chip,wavelens=None,verb=verb)
                   nw = section[3]-section[1]
                   dw = (self["end"][chip]- self["start"][chip])/nw
                   wavelens = arange(self["start"][chip], self["end"][chip]+dw/2.0,dw,Float32)
                   ycen = self.yob
                   slit_len = abs(self["slmax"][chip] - self["slmin"][chip])
                   c1a = self.yloc - slit_len/2.0 + 0
                   c1b = self.yloc + slit_len/2.0 + 1
                   ymap = Expector.getmapping(self.yinfo[1],self.yinfo[2],self.yinfo[0][0],data,ydist=1,verb=0)
                   xmap = Expector.getmapping(self.xinfo[1],self.xinfo[2],self.xinfo[0][0],data,ydist=0,verb=0)
                   if write:
                      ydist = output.replace(".fits","-y-dist.fits")
                      xrect = output.replace(".fits","-x-rect.fits")
                   if linelist and datum[0] not in ["HOLE"]:
                      wdata = Expector.rebin(data,xmap,ymap,data=0,cubic=0,wrap=0,constant=0)
                      wdata = getpct(wdata,40)
                      try:
                         for wp in [len(self.winfo[0][0])]:
                            werr, wsol2, wrms, wspec = Expector.wcalib(0,wdata,self.winfo[0][0],
                            order=len(self.winfo[0][0]),fwhm=3,\
                            w1=self["start"][chip], w2=self["end"][chip],arc=1,skyamp=0,
                            nim=2,tele="",lamps="",gran="",\
                            dich=0,redsid=0,ll=linelist,d0=1,wfft=0,\
                            bth=15,wzero=wp,wtol=1.0,verb=verb)
                            self.woffset = wsol2 - self.winfo[0][0]
                            print self.woffset
                            self.winfo[0] = [wsol2,werr]
                      except "LackOfFeaturesError":
                         print "Problem... continuing with original solution."
                      if write and os.path.exists("temp2.fits"):
                         os.rename("temp2.fits",output.replace(".fits","-1D.fits"))
                if write:
                   slit_len = abs(self["slmax"][chip] - self["slmin"][chip])
                   c1a = self.yloc - slit_len/2.0 + 1
                   c1b = self.yloc + slit_len/2.0 + 1
                   if verb: print "Writing %s." % (output)
                   G = FITS(output,readonly=0,create=1)
                   G.copyheader(F)
                F.close()
                if write:
                   G.resize(onaxes)
                   c1a,c1b = clip([c1a,c1b],1,G["naxis2"]).astype(Int32)
                   G.modkey("nslits",1)
                   G.modkey("csect1a",c1a)
                   G.modkey("csect1b",c1b)
                   G.putdata(data)
                   Expector.writesol(1,0,self.yinfo[0],G,"y0xfm",self.yinfo[1],self.yinfo[2])
                   Expector.writesol(1,0,self.xinfo[0],G,"x0xfm",self.xinfo[1],self.xinfo[2])
                   Expector.writesol(1,0,self.winfo[0],G,"w0xfm",self.winfo[1],self.winfo[2])
                   G.modkey("ydcor",0)
                   G.modkey("xdcor",0)
                   G.modkey("y-dist",os.path.basename(ydist))
                   G.modkey("x-rect",os.path.basename(xrect))
                   if image == images[0]:
                      if verb: print "Opening %s." % (ydist)
                      if verb: print "Opening %s." % (xrect)
                      Gy = FITS(ydist,readonly=0,create=1)
                      Gx = FITS(xrect,readonly=0,create=1)
                      Gy.copyheader(G)
                      Gx.copyheader(G)
                      Gy.putdata(ymap)
                      Gx.putdata(xmap)
                      Gy.close()
                      Gx.close()
                      if verb: print "Closing %s." % (ydist)
                      if verb: print "Closing %s." % (xrect)
                   G.close()
                   if verb: print "Closing %s." % (im)
            print

        def SkyFit(self,images,chip,trace=trace,smf=None,verb=0,skycut=10,skypass=1,write=0,kx=3,dkx=1.0,ky=1,ap=4,adu=1.3,rn=3.0):
            if verb: print "Wavelengths range from %.3f to %.3f." % (self["start"][chip], self["end"][chip])
            self.corners = self.GetCorners(chip,verb=verb)
            if not self.corners: return
            c0,x0,x1,y0,y1 = self.corners
            if verb: print "x0=%8.2f x1=%8.2f y0=%8.2f y1=%8.2f" % (x0,x1,y0,y1)
            datum = smf.GetObject(self["id"])
            for image in images:
                im = image+"_c%d.fits" % (c0)
                if verb: print "Opening %s." % (im)
                F = FITS(im)
                section = concatenate([clip(extrema([y0,y1]),1,F["naxis1"]),clip(extrema([x0,x1]),1,F["naxis2"])])
                section = concatenate([section[::2], section[1::2]+1]).astype(Int32)
                print section
                data = F.readsec(section[:2],section[2:])
                if not self.nodshuffle:
                   data = transpose(data)
                if c0 > 4: data = data[::-1,::-1]
                data = data.astype(Float32)
                onaxes = list(data.shape)
                onaxes.reverse()
                self.shape = data.shape
                if image == images[0]: self.GetGrid(chip,wavelens=None,verb=verb)
                nw = section[3]-section[1]
                dw = (self["end"][chip]- self["start"][chip])/nw
                wavelens = arange(self["start"][chip], self["end"][chip]+dw/2.0,dw,Float32)
                ycen = self.yob
                if trace:
                  w0,w1 = self.mapping.wavelens
                  wspan = (w1-w0)/2.0
                  wmidd = (w1+w0)/2.0
                  delta = - dot(trace,legendre((wavelens-wmidd)/wspan,len(trace)-1))
                  ycen = ycen + delta
                slit_len = abs(self["slmax"][chip] - self["slmin"][chip])
                print self.yloc
                c1a = self.yloc - slit_len/2.0
                c1b = self.yloc + slit_len/2.0
                print c1a,c1b,self.yloc
                if image == images[0]:
                   ymap = Expector.getmapping(self.yinfo[1],self.yinfo[2],self.yinfo[0][0],data,ydist=1,verb=0)
                   xmap = Expector.getmapping(self.xinfo[1],self.xinfo[2],self.xinfo[0][0],data,ydist=0,verb=0)
#                out1 = image+"_m_c%d.fits" % (c0)
#                out2 = image+"_o_c%d.fits" % (c0)
                gus = image+"_s_c%d.fits" % (c0)
                yind,xind = indices(xmap.shape,Float32)
                if image == images[0]:
                   yindr = Expector.reversebin(yind,xmap,ymap,data=0,constant=9999)
                   xindr = Expector.reversebin(xind,xmap,ymap,data=0,constant=9999)
                yuse0 = between(yindr,c1a,c1b,inc=1)*1.0
                ouse0 = logical_not(between(yindr,ycen-ap,ycen+ap))*1.0
                use0 = greater(data,0.0)*yuse0*ouse0
                duse = Expector.rebin(data,xmap,ymap)
                if c0 == 2:
                   qdump("xmap.fits",xmap)
                   qdump("ymap.fits",ymap)
                   qdump("data.fits",data)
                   qdump("yuse.fits",yuse0*99999)
                   qdump("ouse.fits",ouse0*99999)
                   qdump("duse.fits",duse)
                   exit()
                use = use0
                for spass in range(skypass):
                   if spass == skypass-1: ukx,uky = kx,ky
                   else: ukx,uky = kx,max([0,ky-2])
                   xs,ys,zs = compress(use.flat,[xindr.flat,yindr.flat,data.flat])
                   xs,ys,zs = take( [xs,ys,zs], argsort(xs), 1)
                   ws = divz(1,sqrt(zs/adu+pow(rn/adu,2)))
                   if skycut and not spass:
                     mzs = VTKMedian(zs,min([31,int(slit_len)-2*int(ap+1)]),1,1)
                     dzs = zs-mzs
                     szs = dzs*ws
                     kzs = less(abs(szs),skycut)
                     xs,ys,zs,ws = compress(kzs, [xs,ys,zs,ws], 1)
                   ukx1,uky1 = ukx + 1, uky + 1
                   if uky: xs1,ys1,zs1,ws1 = xs,ys,zs,ws
                   else: xs1,ys1,zs1,ws1 = Expector.adjspac([xs,ys,zs,ws])
                   xmin,xmax = extrema(xs1)
                   ymin,ymax = extrema(ys1)
                   tx = arange(xmin,xmax+dkx/2.0,dkx,Float32)
                   itx = diff(searchsorted(xs1,tx))
                   utx = greater(itx,2*ukx+2)
                   tx = compress(utx,tx[:-1]+diff(tx)/2.0)
                   print "fitting b-spline (%d,%d) to %d points using %d knots." % (ukx,uky,len(xs1),len(tx))
                   if uky:
                     tx = concatenate([(ukx-1)*[tx[0]],tx,(ukx-1)*[tx[-1]]])
                     ty = concatenate([uky1*[ymin],uky1*[ymax]])
                     tck = Expector.bisplrep(xs1,ys1,zs1,w=ws1,s=0,task=-1,kx=ukx,ky=uky,\
                        tx=tx,ty=ty,xb=xmin,xe=xmax,yb=ymin,ye=ymax,quite=1)
                     sky = reshape(bisplev2(ravel(xindr),ravel(yindr),tck),data.shape)
                     sky = where(between(sky,-1e24,1e24),sky,0.0)
                   else:
                     tck = Expector.splrep(xs1,zs1,w=ws1,s=0,task=-1,k=ukx,\
                        t=tx,xb=xmin,xe=xmax,quite=1)
                     sky = zeros(data.shape,Float64)
                     [put(sky[jy],argsort(xindr[jy]),splev(take(xindr[jy],argsort(xindr[jy])),tck)) for jy in range(xindr.shape[0])]
                     sky = where(between(sky,-1e24,1e24),sky,0.0)
                   if spass < skypass-1:
                     sob = data-sky
                     use = use0*less(divz(abs(sob),getsqrt(data)),skycut)
#                if verb: print "Writing to %s and %s." % (out1,out2)
                if verb: print "Writing to %s." % (gus)
                sob = where(yuse0,sky,0)
                oob = where(yuse0,data-sky,0)
                use = yuse0
                if c0 > 4:
                  sob = sob[::-1,::-1]
                  oob = oob[::-1,::-1]
                  use = use[::-1,::-1]
                if not self.nodshuffle:
                   sob = transpose(sob)
                   oob = transpose(oob)
                   use = transpose(use)
                section = section.tolist()
                G = FITS(gus,readonly=0)
                oldoob = G.readsec(section[:2]+[0],section[2:]+[0])[0]
                oldsob = G.readsec(section[:2]+[1],section[2:]+[1])[0]
                oob = where(use,oob,oldoob)
                sob = getsqrt((oob+sob)/adu+pow(rn/adu,2))
                sob = where(use,sob,oldsob)
                G.writesec(oob,section[:2]+[0],section[2:]+[0])
                G.writesec(sob,section[:2]+[1],section[2:]+[1])
                G.close()
#                G = FITS(out1,readonly=0)
#                oldsob = G.readsec(section[:2],section[2:])
#                sob = where(use,sob,oldsob)
#                G.writesec(sob,section[:2],section[2:])
#                G.close()
#                G = FITS(out2,readonly=0)
#                oldoob = G.readsec(section[:2],section[2:])
#                oob = where(use,oob,oldoob)
#                G.writesec(oob,section[:2],section[2:])
#                G.close()

        def Rebin(self,images,chip,dw,trace=trace,smf=None,verb=0,logb=0):
            if verb: print "Wavelengths range from %.3f to %.3f." % (self["start"][chip], self["end"][chip])
            self.corners = self.GetCorners(chip,verb=verb)
            if not self.corners: return
            c0,x0,x1,y0,y1 = self.corners
            if verb: print "x0=%8.2f x1=%8.2f y0=%8.2f y1=%8.2f" % (x0,x1,y0,y1)
            datum = smf.GetObject(self["id"])
            length = self["length"]
            outputs = []
            for image in images:
                im = image+"_c%d.fits" % (c0)
                if verb: print "Opening %s." % (im)
                F = FITS(im)
                section = concatenate([clip(extrema([y0,y1]),1,F["naxis1"]),clip(extrema([x0,x1]),1,F["naxis2"])])
                section = concatenate([section[::2], section[1::2]+1]).astype(Int32)
                data = F.readsec(section[:2],section[2:])
                data = transpose(data)
                if c0 > 4: data = data[::-1,::-1]
                data = data.astype(Float32)
                onaxes = list(data.shape)
                onaxes.reverse()
                self.shape = data.shape
                if image == images[0]: self.GetGrid(chip,wavelens=None,verb=verb)
                nw = section[3]-section[1]
                delw = (self["end"][chip] - self["start"][chip])/nw
                wavelens = arange(self["start"][chip], self["end"][chip]+delw/2.0,delw,Float32)
                ycen = self.yob
                w0,w1 = self.mapping.wavelens
                slit_len0 = int(abs(self["slmax"][0] - self["slmin"][0]))
                slit_len = abs(self["slmax"][chip] - self["slmin"][chip])
                c1a = int(self.yloc - slit_len0/2.0 + 0.5)
                c1b = int(self.yloc + slit_len0/2.0 + 0.5)
                c1a,c1b = clip([c1a,c1b],0,data.shape[0])
                if image == images[0]:
                   ymap = Expector.getmapping(self.yinfo[1],self.yinfo[2],self.yinfo[0][0],data,ydist=1,verb=0)
                   xmap = Expector.getmapping(self.xinfo[1],self.xinfo[2],self.xinfo[0][0],data,ydist=0,verb=0)
                   if trace:
                     wspan = (w1-w0)/2.0
                     wmidd = (w1+w0)/2.0
                     delta = - dot(trace,legendre((wavelens-wmidd)/wspan,len(trace)-1))
                     ymap = ymap + delta
                   wxfm=[0,self.winfo[0][0],w0,w1,dw,logb]
                   xmap,good = Expector.wavemapping(xmap,wxfm)
                   ymap = Expector.wavemapping(ymap,wxfm,ydist=1)[0]
                data = Expector.rebin(data,xmap,ymap,data=1,cubic=0)[c1a:c1b].astype(Float32)
                print data.shape[0],slit_len0
                if data.shape[0] < slit_len0:
                  data2 = zeros((slit_len0,data.shape[1]),Float32)
                  data2[:data.shape[0]] = data
                  data = data2
                outputs.append(data)
            print len(outputs)
            print [d.shape for d in outputs]
            print
            return outputs

        def Pipeline(self,images,suffix,chip,subdir=None,pos=None,smf=None,refkey="s",verb=0):
            datum = smf.GetObject(self["id"])
            rawflat = suffix.index("flat")
            rawspec = suffix.index("spec")
            if subdir: os.chdir(subdir)
            if verb: print "Wavelengths range from %.3f to %.3f." % (self["start"][chip], self["end"][chip])
            self.corners = self.GetCorners(chip,verb=verb)
            if not self.corners: return
            c0,x0,x1,y0,y1 = self.corners
            if verb: print "x0=%8.2f x1=%8.2f y0=%8.2f y1=%8.2f" % (x0,x1,y0,y1)
            loc = [pos[2][j] for j in range(len(pos[0])) if pos[0][j] == self["id"] and pos[1][j] == c0][0]
            flat = "%s_%s_c%d.fits" % (self["id"],images[rawflat],c0)
            spec = "%s_%s_c%d.fits" % (self["id"],images[rawspec],c0)
            if refkey in self["id"]:
               specf = spec
            else:
               command = "modrect",flat,"-kx 3","-ky 0","-dkx 2","-ydist","-xdist"
               verbex(join(command))
               blaze = flat.replace(".fits","m.fits")
               norm = flat.replace(".fits","n.fits")
               specf = spec.replace(".fits","f.fits")
               command = "efits",flat,blaze,"\"newdivz(i1,i2)\"",norm
               verbex(join(command))
               command = "efits",spec,norm,"\"newdivz(i1,i2)\"",specf
               verbex(join(command))
            specm = specf.replace(".fits","m.fits")
            speco = specf.replace(".fits","o.fits")
            skym = specf.replace(".fits",".sky")
            F = FITS(specf)
            c1a, c1b = F["csect1a"], F["csect1b"]
            F.close()
            f = open(skym,"w")
            if c1a+3 < loc-2: print >> f, c1a+3,loc-2
            if c1b-3 > loc+2: print >> f, loc+2, c1b-3
            f.close()
            command = "skyrect",specf,"-kx 3","-ky 1","-dkx 1.00","-skyint 8192",\
                "-ydist","-xdist","-skysuf sky","-check 5","-pct 50","-skycut 10","-g 1","-rn 5"
            verbex(join(command))
            command = "efits",specf,specm,"\"(i1-i2)\"",speco
            verbex(join(command))
            if subdir: os.chdir("../")

        def TestRect(self,images,suffix,subdir=None,pos=None,smf=None,refkey="s",dw=2.0,verb=0):
            datum = smf.GetObject(self["id"])
            if "flat" in suffix:
              rawflat = suffix.index("flat")
            else:
              rawflat = -1
            rawspec = suffix.index("spec")
            if subdir: os.chdir(subdir)
            done = []
            locs = []
            for chip in range(self["nobs"]):
               if verb: print "Wavelengths range from %.3f to %.3f." % (self["start"][chip], self["end"][chip])
               self.corners = self.GetCorners(chip,verb=verb)
               if not self.corners: return
               c0,x0,x1,y0,y1 = self.corners
               if verb: print "x0=%8.2f x1=%8.2f y0=%8.2f y1=%8.2f" % (x0,x1,y0,y1)
               if pos:
                 loc = [pos[2][j] for j in range(len(pos[0])) if pos[0][j] == self["id"] and pos[1][j] == c0][0]
               else:
                 loc = 0.0
               locs.append(loc)
               if rawflat >=0: flat = "%s_%s_c%d.fits" % (self["id"],images[rawflat],c0)
               spec = "%s_%s_c%d.fits" % (self["id"],images[rawspec],c0)
               if refkey in self["id"]:
                  specf = spec
               else:
                  if rawflat >=0:
                    blaze = flat.replace(".fits","m.fits")
                    norm = flat.replace(".fits","n.fits")
                  specf = spec.replace(".fits","f.fits")
               specm = specf.replace(".fits","m.fits")
               speco = specf.replace(".fits","o.fits")
               specr = specf.replace(".fits","r.fits")
               skym = specf.replace(".fits",".sky")
               command = "unrect",specf,"-ydist","-xdist","-wdist","-w1 %f -w2 %f -dw %f" % tuple(tuple(self.mapping.wavelens)+(dw,))
               verbex(join(command),dummy=0)
               F = FITS(specr)
               c1a, c1b = array([F["csect1a"]-1, F["csect1b"]]).astype(Int32)
               rdata = F.data()[c1a:c1b]
               F.close()
               done.append(rdata)
               rect = specr.replace( "%s_%s_c%d" % (self["id"],images[rawspec],c0), "%s_%s" % (self["id"],images[rawspec]))
            my = max([r.shape[0] for r in done])
            mx = max([r.shape[1] for r in done])
            rout = zeros((my,mx),Float32)
            for j in range(len(done)):
                r = done[j]
                s = floor(locs[0] - locs[j] + 0.5)
                sr = VTKImageShift(r,0,s,numret=1,interp=0)
                rout[:r.shape[0],:r.shape[1]] = rout[:r.shape[0],:r.shape[1]] + sr
            G = FITS(rect,create=1,template=specr)
            onaxes = [mx,my]
            G.resize(onaxes)
            G.modkey("nslits",1)
            G.modkey("csect1a",1)
            G.modkey("csect1b",my)
            G.putdata(rout)
            G.close()
            if subdir: os.chdir("../")

        def Profile(self,images,suffix,c,subdir=None,pos=None,smf=None,refkey="s",verb=0,hap=3,ho=1,lg=0):
            chip = self["chip"][c]
            objdat = [(pos[0][j],pos[1][j],pos[2][j]) for j in range(len(pos[0])) if pos[0][j] == self["id"] and pos[1][j] == chip]
            id,c0,loc = objdat[0]
            print id,c0,loc
            spec = "%s_%s_c%d.fits" % (self["id"],suffix,c0)
            if refkey in self["id"]: specf = spec
            else: specf = spec.replace(".fits","f.fits")
            specm = specf.replace(".fits","m.fits")
            speco = specf.replace(".fits","o.fits")
            os.chdir(subdir)
            F = FITS(speco)
            c1a, c1b = F["csect1a"], F["csect1b"]
            F.close()
            i = Extractions.input_frame(speco,1.0,5.0)
            data = i.GetPixels(loc,hap,0)
            woco,uw,uy,uv,um,us,ue,uq,uq0 = data
            wcent = mean(self.mapping.wavelens)
            wspan = diff(self.mapping.wavelens)[0]/2.0
            lbase = (uw-wcent)/wspan
            lbasis = legendre(lbase,lg)
            adu = 1.0
            wuse = divz(um**2*greater(um,-3*getsqrt(uv)),uq+um*greater(um,0)/adu+power(uv-um,2))
            self.ad[c] = svdfit(transpose(wuse[NewAxis,::]*lbasis),wuse*uy)
            print self.ad[c]
            tb = dot(self.ad[c],lbasis)
            uy = uy - tb
            wsig = hap/2.0
            wcen = self.ad[c][0]/2.0
            def fitgp(params):
               s = 10**params[0]
               kernel = -0.5*power((uy-params[2])/s,2)
               p = params[1]*getexp(kernel) + params[3]
               return (uv-p)*mean(uw)
            wfit = leastsq(fitgp,[log10(wsig),getpct(uv,90),wcen,getpct(uv,10)],factor=0.1,xtol=1e-4,ftol=1e-4,gtol=1e-4,epsfcn=1e-4)[0]
            self.wsig[c] = 10**wfit[0]
            self.counts[c] = wfit[1]/sqrt(2*pi*self.wsig[c]**2)
            print wfit
            print self.counts[c]
            print "wsig_f =", self.wsig[c]
            wt = greater(uv,-2*getsqrt(uq))
            self.pb[c] = Extractions.Hdecomp(0.0,ho,uv,wt,getsqrt(uq),0,uy,self.wsig[c],0)
            print self.pb[c]
            os.chdir("../")


class Mapping:
        def __init__(self,mapfile,trace=None):
          self.mapfile = mapfile
          self.trace = trace
          self.ReadMap()
          self.Basic()
          self.GetObjectList()
        def ReadMap(self):
          self.mapdata = open(self.mapfile,"r").readlines()
        def Basic(self):
          self.xdisp = int(self.mapdata[0].split()[-1])
          self.ord_disp,self.ord_sag,self.ord_tilt,self.ord_sagit,self.ord_len = \
                        map(int,self.mapdata[1].split()[-5:])
          self.telscale = float(self.mapdata[2].split()[2])
          self.wavelens = array(map(float,self.mapdata[3].split()[-2:]))
          self.dewar = self.mapdata[4].split()[2]
          print "Mapping covers between %.3fA to %.3fA using %s" % (tuple(self.wavelens)+(self.dewar,))
        def GetObjectList(self):
           self.objects = [l.split()[2] for l in self.mapdata if l[:4] in ["SLIT","END"]]
           self.types = [int(l.split()[3]) for l in self.mapdata if l[:4] in ["SLIT"]]
           self.positions = [self.mapdata.index(l) for l in self.mapdata if l[:4] in ["SLIT","END"]]
           self.number = len(self.objects)
           self.objcache = {}
        def Recover(self,target):
           j = self.objects.index(target)
           otype = self.types[j]
           p = self.positions[j]
           if j < self.number-1:
              pe = self.positions[j+1]
              objmap = self.mapdata[p:pe]
           else:
              objmap = self.mapdata[p:]
           objmap = [l for l in objmap if l != "\n"]
           objdict = Object()
           self.objcache[target] = objdict
           objdict.SetMapping(self)
           chipdata = [map(eval,l.split()[1:]) for l in objmap if l.split()[0] == "CHIP"]
           chippos = [objmap.index(l) for l in objmap if l.split()[0] == "CHIP"]
           objdict["id"] = target
           objdict["type"] = otype
           objdict["nobs"] = len(chipdata)
           objdict["yap"] = zeros((objdict["nobs"],),Float32)
           objdict["length"] = [float(l.split()[-1]) for l in objmap if l.split()[0] == "LENGTH"]
           if objdict["length"]: objdict["length"] = objdict["length"][0]
           else: objdict["length"] = 0
           objdict.counts = [None for j in range(objdict["nobs"])]
           objdict.pb = [None for j in range(objdict["nobs"])]
           objdict.ad = [None for j in range(objdict["nobs"])]
           objdict.wsig = [None for j in range(objdict["nobs"])]
           for key in CHIPKEYS: objdict[key] = []
           for key in COEFKEYS: objdict[key] = []
           [objdict[key].append(chipdata[i][CHIPKEYS.index(key)]) for key in CHIPKEYS for i in range(objdict["nobs"])]
           [objdict[COEFKEYS[k]].append( array(map(eval,objmap[chippos[i]+1+k].split()))) for k in range(len(COEFKEYS)) for i in range(objdict["nobs"])]
           if self.trace:
              for c in range(objdict["nobs"]):
                  olength = objdict["slmax"][c] - objdict["slmin"][c]
                  w0,w1 = self.wavelens
                  wspan = (w1-w0)/2.0
                  wmidd = (w1+w0)/2.0
                  ow0,ow1 = objdict["start"][c], objdict["end"][c]
                  osw = (ow1-ow0)/100
                  scw = arange(ow0,ow1+osw/2,osw,Float64)
                  sp = polynom(scw,2)
                  ydelta = -dot(self.trace,legendre((scw-wmidd)/wspan,len(self.trace)-1))
                  new_sag = svdfit(transpose(sp),ydelta)
                  yd2 = dot(new_sag,sp)
                  objdict["coef_sag"][c][:3] = objdict["coef_sag"][c][:3] + new_sag

class Profile:
     def __init__(self,profile):
         f = open(profile,"r")
         l = f.readlines()
         f.close()
         self.sig = eval(l[1])
         self.ad = asarray(eval(l[4]))
         i = 7
         self.pb = l[i].strip()
         while "]" not in self.pb: i,self.pb=i+1,self.pb+l[i+1].strip()
         self.pb = asarray(eval(self.pb))
         self.ho = len(self.pb)-1
         self.lg = len(self.ad)-1

class Extraction:
     def __init__(self,Mapping,object,w1,w2,dw,subdir,posdat,suffices,profiles,scales,refkey,kx=3,ky=0,spass=1,cut=0,hap=3):
         self.object = object
         self.mapping = Mapping
         self.scales = scales
         self.kx = kx
         self.ky = ky
         self.cut = cut
         self.spass = spass
         self.w1 = w1
         self.w2 = w2
         self.dw = dw
         self.hap = hap
         self.profiles = profiles
         self.posdat = posdat 
         self.suffices = suffices
         self.ref = refkey in self.object
         time0 = time.time()
         os.chdir(subdir)
         done = [self.DoChip(chip) for chip in 1+arange(8)]
         done = [d for d in done if d]
         if len(done):
            name = "%s_1D.fits" % (self.object)
            command = ["efits"]+done
            command += ["\"sum(allimages)\"",name]
            verbex(join(command))
         time1 = time.time()
         print "Total time was %.3fs." % (time1-time0)
         os.chdir("../")
     def DoChip(self,chip):
         f,l = self.GetInfo(chip)
         if not len(f): return None
         if self.ref:
             spec = ["%s_%s_c%do.fits" % (self.object,s,chip) for s in self.suffices]
             extra = ["%s_%s_c%d.fits" % (self.object,s.replace("spec","flat"),chip) for s in self.suffices]
         else:
             spec = ["%s_%s_c%dfo.fits" % (self.object,s,chip) for s in self.suffices]
             extra = ["%s_%s_c%dm.fits" % (self.object,s.replace("spec","flat"),chip) for s in self.suffices]
         images = [spec[j] for j in f]
         Fs = [FITS(im) for im in images]
         f = [j for j in range(len(Fs)) if Fs[j]["skyim"] != "N/A"]
         [F.close() for F in Fs]
         print f
         if not f: return None
         images = [spec[j] for j in f]
         extras = [extra[j] for j in f]
         locs = asarray([l[j] for j in f]).flat
         sigs = asarray([self.profiles[j].sig for j in f]).flat
         pbs = [self.profiles[j].pb for j in f]
         hos = [self.profiles[j].ho for j in f]
         if self.ref:
            ads = [0*self.profiles[j].ad for j in f]
            lgs = [self.profiles[j].lg for j in f]
         else:
            ads = [0*self.profiles[j].ad for j in f]
            lgs = [0*self.profiles[j].lg for j in f]
         lgran = [self.mapping.wavelens for j in f]
         sc = [self.scales[j] for j in f]
         Fs = [FITS(im) for im in images]
         egain = [F["newgain"] for F in Fs]
         enoise = [F["enoise"] for F in Fs]
         [F.close() for F in Fs]
         Inputs = [Extractions.input_frame(images[j],egain[j],enoise[j],extra=extras[j]) for j in range(len(images))]
         if self.hap < 0: hwids = abs(self.hap)*sigs
         else: hwids = self.hap*ones(sigs.shape)
         print locs
         print sigs
         print hwids
         print sc
         name = "%s_c%d" % (self.object,chip)
         Extractions.GetObject(name,locs,hwids,Inputs,w1=self.w1,w2=self.w2,dw=self.dw,\
                kx=self.kx,ky=self.ky,lg=lgs[0],lgdata=ads,lgran=lgran,ho=hos[0],hodata=pbs,\
                insigs=sigs,inscales=sc,cut=self.cut,spass=self.spass)
         name1d = name+"_1D.fits"
         return name1d
     def GetInfo(self,chip):
         j = [nonzero([(self.object == i) for i in p[0]]) for p in self.posdat]
         l = [[self.posdat[i][2][k] for k in j[i] if self.posdat[i][1][k] == chip] for i in range(len(j))]
         f = nonzero([len(q) for q in l])
         return f,l

def gaussian(p,x,y,dy=None):
    a,c,s,b = p
    s = 10**s
    k = -0.5*power((x-c)/s,2)
    g = a*getexp(k)+b
    r = y-g
    if dy: return divz(y-g,dy)
    else: return g

def SpawnTwisted(objects):
   commands = ["%s -o %s %s" % (os.path.basename(sys.argv[0]),object,INITIAL_COMMAND) for object in objects]
   Q = TwistedClient.TwistedVerbexQueue(commands)
   Q.Run()
   exit()



