
import time
from FITS import *
from Multipack import *
from VTKHelperFunctions import *
from ReadSex import *
from masks import *
from basics import *

NA = NewAxis

def QuickShift(i,x,y):
    s = VTKImageShift(i,x,y,interp=0,numret=1)
#    s = shiftim(i,-x,-y)
    return s

class Observation:
    def __init__(self,image,wt=None,scale=0.125,pakey="rotangle",saturate=30000,sigsuf=None):
       self.image = image
       self.catalog = image.replace(".fits",".cat")
       if sigsuf: self.sigimage = image.replace(".fits",sigsuf+".fits")
       else: self.sigimage = None
       self.template = image.replace(".fits","master.fits")
       self.imbases = image.replace(".fits","bases.fits")
       self.bimage = image.replace(".fits","bg.fits")
       self.nimage = image.replace(".fits","noise.fits")
       self.psfgrid = image.replace(".fits","grid.fits")
       self.masked = image.replace(".fits","mask.fits")
       self.difference = image.replace(".fits","diff.fits")
       self.segmap = "."+os.path.basename(self.image.replace(".fits","seg.fits"))
       self.wt = wt
       self.bg = 0.0
       self.scale = scale
       self.saturate = saturate
       f = FITS(self.image)
       self.epoch = f["epoch"]
       self.exptime = f["exptime"]
       if self.epoch == "N/A": self.epoch = -99
       if self.exptime == "N/A": self.exptime = 1.0
       self.ra = f["ra"]
       self.dec = f["dec"]
       self.pa = f[pakey]
       if self.pa == "N/A": self.pa = 0.0
       if type(self.ra) == StringType: self.ra = 15.0*scanSexagesimal(self.ra)
       if type(self.dec) == StringType: self.dec= scanSexagesimal(self.dec)
       f.close()
       self.data = None
       self.master = None
    def __repr__(self):
       return self.image
    def sex(self,sexdir="./sex/",thresh=5.0):
       if sexdir[-1] != "/": sexdir += "/"
       self.sexcom = ["sex",self.image]
       self.sexcom += ["-c "+sexdir+"default.sex"]
       self.sexcom += ["-PARAMETERS_NAME "+sexdir+"default.param"]
       if self.wt: self.sexcom += ["-WEIGHT_IMAGE "+self.wtimage]
       if self.wt: self.sexcom += ["-WEIGHT_TYPE MAP_RMS"]
       self.sexcom += ["-DETECT_MINAREA 5"]
       self.sexcom += ["-DETECT_THRESH "+str(thresh)]
       self.sexcom += ["-ANALYSIS_THRESH "+str(thresh)]
       self.sexcom += ["-PIXEL_SCALE "+str(self.scale)]
       self.sexcom += ["-SEEING_FWHM "+str(0.8/self.scale)]
       self.sexcom += ["-CATALOG_NAME",self.catalog]
       self.sexcom += ["-CHECKIMAGE_TYPE SEGMENTATION"]
       self.sexcom += ["-CHECKIMAGE_NAME "+self.segmap]
       self.sexcom += ["-FILTER_NAME "+sexdir+"gauss_3.0_5x5.conv"]
       self.sexcom += ["-STARNNW_NAME "+sexdir+"default.nnw"]
       self.sexcom += ["-SATUR_LEVEL "+str(self.saturate)]
       self.sexcom += ["-VERBOSE_TYPE QUIET"]
       self.sexcom = join(self.sexcom)
       verbex(self.sexcom)
       self.thresh = thresh
    def readcat(self):
       self.sexdata = readsex(self.catalog)[0]
    def compute(self,nmax=40):
       self.nmax = nmax
       x = self.sexdata["X_IMAGE"]
       y = self.sexdata["Y_IMAGE"]
       if "MAG_BEST" in self.sexdata: m = self.sexdata["MAG_BEST"]
       elif "MAG_AUTO" in self.sexdata: m = self.sexdata["MAG_AUTO"]
       q = self.sexdata["FLAGS"]
       x,y,m,q = compress(less_equal(q,4),[x,y,m,q],1)
       a = self.pa*pi/180
       r = self.ra*pi/180
       d = self.dec*pi/180
       o = argsort(m)[:self.nmax]
       x,y,m = take([x,y,m],o,1)
       dx = x[NewAxis,::] - x[::,NewAxis]
       dy = y[NewAxis,::] - y[::,NewAxis]
       ds = sqrt(power(dx,2) + power(dy,2))
       da = arctan2(dy,dx)*180/pi
       self.x = x
       self.y = y
       self.m = m
       self.dx = dx
       self.dy = dy
       self.ds = ds
       self.da = da
    def objmatch(self,nord=0,perr=2,aerr=1.0):
       print "Matching",self.master,"to",self
       if "sexdata" not in dir(self.master):
          self.master.master = self.master
          self.master.readcat()
          self.master.compute(nmax=self.nmax)
       dx = self.dx[::,::,NA,NA] - self.master.dx[NA,NA,::,::]
       dy = self.dy[::,::,NA,NA] - self.master.dy[NA,NA,::,::]
       ds = sqrt(dx**2+dy**2)
       da = self.da[::,::,NA,NA] - self.master.da[NA,NA,::,::]
#       use = less(abs(dx),perr)*less(abs(da),aerr)
       use = less(abs(ds),perr)*less(abs(da),aerr)
       suse = add.reduce(add.reduce(use,3),1)
       if self.verb:
          for suse1 in suse: print suse1
#       guse = greater(suse,self.x.shape[0]/5)
       guse = greater(suse,max(suse.flat)/2)
       i = [j for j in range(self.x.shape[0]) if sum(guse[j])]
       m = [argmax(guse[j]) for j in range(self.x.shape[0]) if sum(guse[j])]
       x0,y0 = take([self.x,self.y],i,1)
       x1,y1 = take([self.master.x,self.master.y],m,1)
       xshift,xscat = bwt(x0-x1)
       xscat = max([1.0,xscat])
       yshift,yscat = bwt(y0-y1)
       yscat = max([1.0,yscat])
       keep = less(abs(x0-x1-xshift),3*xscat)*less(abs(y0-y1-yshift),3*yscat)
       x0,y0,x1,y1 = compress( keep, [x0,y0,x1,y1], 1)
       wt = ones(x0.shape,Float32)

       for iter in range(3):
          if iter:
             allx0 = array(self.sexdata["X_IMAGE"],Float64)
             ally0 = array(self.sexdata["Y_IMAGE"],Float64)
             if "MAG_BEST" in self.sexdata: allm0 = array(self.sexdata["MAG_BEST"])
             elif "MAG_AUTO" in self.sexdata: allm0 = array(self.sexdata["MAG_AUTO"])
             allq0 = array(self.sexdata["FLAGS"])
             allx1 = array(self.master.sexdata["X_IMAGE"],Float64)
             ally1 = array(self.master.sexdata["Y_IMAGE"],Float64)
             if "MAG_BEST" in self.sexdata: allm1 = array(self.master.sexdata["MAG_BEST"])
             elif "MAG_AUTO" in self.sexdata: allm1 = array(self.master.sexdata["MAG_AUTO"])
             allq1 = array(self.master.sexdata["FLAGS"])
             if nord == -1:
                ix = allx0 + xshift
                iy = ally0 + yshift
             else:
                basis = abasis(nord,allx0,ally0,rot=[0,1][nord==0])
                ixy = add.reduce(sol[NewAxis,::]*basis,1)
                ix,iy = ixy[:len(allx0.flat)], ixy[len(allx0.flat):]
             delx = ix[::,NA] - allx1[NA,::]
             dely = iy[::,NA] - ally1[NA,::]
             dels = sqrt(power(delx,2) + power(dely,2))
             delq = less_equal(allq0,4)[::,NA]*less_equal(allq1,4)[NA,::]
             dels = where(delq,dels,9999)
             okay = less(dels,tol)
             ui0 = [j for j in range(delx.shape[0]) if min(dels[j]) < 1.0]
             ui1 = [argmin(dels[j]) for j in range(delx.shape[0]) if min(dels[j]) < 1.0]
             x0,y0,m0 = take([allx0,ally0,allm0],ui0,1)
             x1,y1,m1 = take([allx1,ally1,allm1],ui1,1)
             f0 = power(10,-0.4*(m0-min(m0)))
             f1 = power(10,-0.4*(m1-min(m1)))
             wt = sqrt(f0+f1) * 0.0 + 1
          if nord == -1:
             xshift = sum(wt*(x1-x0))/sum(wt)
             yshift = sum(wt*(y1-y0))/sum(wt)
             sol = array([xshift,yshift,1.0,0.0])
             ix = x0 + xshift
             iy = y0 + yshift
          else:
             wt = concatenate([wt,wt])
             basis = abasis(nord,x0,y0,rot=[0,1][nord==0])
             sol = svdfit(basis*wt[::,NA],concatenate([x1,y1])*wt)
             ixy = add.reduce(sol[NewAxis,::]*basis,1)
             ix,iy = ixy[:len(x0.flat)], ixy[len(x0.flat):]
          print "Pass %d, with %d number of objects." % (iter+1,len(x0))
          print sol
          delx = ix-x1
          dely = iy-y1
          dels = sqrt(power(delx,2) + power(dely,2))
          scx = bwt(delx)[1]
          scy = bwt(dely)[1]
          print "Biweight estimate for scatter (x,y) = (%.5f,%.5f)" % (scx,scy)
          tol = 3.0*bwt(dels)[1]
       self.sol = sol
       self.nord = nord
    def imread(self):
        if self.data == None:
           print "Now reading frames for %s" % (self)
           f = FITS(self.image)
           self.data = f.data()
           self.naxis1,self.naxis2 = f.naxes()
           f.close()
           f = FITS(self.segmap)
           self.seg = f.data()
           f.close()
           if self.sigimage:
               print "Reading",self.sigimage
               f = FITS(self.sigimage)
               self.sigimage = f.data()
               f.close()
           if self.master: self.master.imread()
    def mktemplate(self,registered,sol=None,mimage=None):
        self.oy,self.ox = indices((self.naxis2,self.naxis1),Float32)
        if registered:
           self.timage = self.master.data
           self.mseg = greater(self.master.seg,0.0).astype(Float32)
        else:
           if not sol: sol = self.sol
           if not mimage: mimage = self.master.data
           print "Constructing transformation..."
           ix,iy = mbasis(sol,self.ox,self.oy,rot=[0,1][self.nord<=0])
           self.tx = ix - self.ox
           self.ty = iy - self.oy
           print "Transforming..."
           self.timage = VTKImageTransform(mimage,self.tx,self.ty,numret=1,cubic=0,interp=1)
           mseg = greater(self.master.seg,0.0).astype(Float32)
           self.mseg = VTKImageTransform(mseg,self.tx,self.ty,numret=1,cubic=0,interp=1,constant=-99999)
    def mkweight(self):
        print "Working on mask and weights..."
        seg = greater(self.seg,0.0).astype(Float32)
        mseg = greater(self.mseg,0.0).astype(Float32)
        gwid = max([2,1*self.pwid])
        if self.verb>=3: opendev("/xs",nx=2,ny=1,land=0,asp=0.5,size=10,ask=0)
        for uk in range(5):
           if uk:
              ukeep = 1.*between(self.data,bg-4*r,bg+2*r)
              tkeep = 1.*between(self.timage,tbg-4*r2,tbg+2*r2)
              ukeep = 1.0*equal(VTKConvolve(ukeep,k=5,numret=1),1.0)
              tkeep = 1.0*equal(VTKConvolve(tkeep,k=5,numret=1),1.0)
              udata = compress(ukeep.flat,self.data.flat)
              utimage = compress(tkeep.flat,self.timage.flat)
           else:
              udata = self.data.flat
              utimage = self.timage.flat
           bg = GaussianBG(udata,99,verb=self.verb)
           if self.verb>=3: subtitle(self.image,ch=2)
           tbg = GaussianBG(utimage,99,verb=self.verb)
           if self.verb>=3: subtitle(self.master.image,ch=2)
           d1 = VTKSubtract(udata,bg)
           d2 = VTKSubtract(utimage,tbg)
           r = 1.49*median(abs(d1))
           r2 = 1.49*median(abs(d2))
           print "Using %d: BG=%.3f with sigma=%.3f" % (len(udata),bg,r)
           print "Using %d: tBG=%.3f with sigma=%.3f" % (len(utimage),tbg,r2)
        if self.verb>=3: closedev()
        self.bg = bg
        self.tbg = tbg
        qdump(self.bimage,self.data-bg,self.image)
        print "Cutting below 5-sigma: %.3f" % (bg-5*r)
        wt = VTKMultiply(seg, VTKMultiply(mseg, less(VTKGauss(equal(self.data,0.0).astype(Float32),gwid,numret=1),0.02)))
        swt = less(VTKGauss(less(self.data,bg-5*r).astype(Float32),gwid,numret=1),0.02).astype(Float32)
        twt = less(VTKGauss(less(self.timage,tbg-5*r2).astype(Float32),gwid,numret=1),0.02).astype(Float32)
        wt = VTKMultiply(greater(VTKGauss(wt,gwid,numret=1),0.02).astype(Float32),VTKMultiply(swt,twt))
        insidex = between(self.ox,2*gwid,self.naxis1-2*gwid).astype(Float32)
        insidey = between(self.oy,2*gwid,self.naxis2-2*gwid).astype(Float32)
        if self.xwin: insidex = VTKMultiply(insidex,between(self.ox,self.xwin[0],self.xwin[1]).astype(Float32))
        if self.ywin: insidey = VTKMultiply(insidey,between(self.oy,self.ywin[0],self.ywin[1]).astype(Float32))
        wt = VTKMultiply(wt,VTKAnd(insidex, insidey))
        swt = VTKGreaterEqual(self.data,self.saturate)
        twt = VTKGreaterEqual(self.timage,self.master.saturate)
        swt = VTKDilate(VTKOr(swt,twt),5,5,numret=1)
        swt = VTKGauss(swt,gwid,numret=1)
        wt = VTKMultiply(wt, VTKLessEqual(swt,0.02))
        if self.sigimage:
           self.noise = self.sigimage
        else:
           n2 = VTKSubtract(self.data,bg)
           self.noise = VTKSqrt(VTKAdd(VTKAbs(n2),pow(r,2)))
        qdump(self.nimage,self.noise,self.image)
        self.invnoise = VTKInvert(self.noise)
        wt = VTKMultiply(wt, self.invnoise)
        wt = wt * between(self.ox,2*self.pwid,self.naxis1-2*self.pwid) * between(self.oy,2*self.pwid,self.naxis2-2*self.pwid)
        if self.pwid > -1: wt = VTKIslandRemoval(wt,1.0,0.0,max([3,self.pwid])**2,numret=1)
        self.wt = wt.astype(Float32)
    def mkmatch(self,preserve):
        wtflat = self.wt.flat
        data = (self.data-self.bg)
        timage = (self.timage-self.tbg) * (self.exptime/self.master.exptime)
        print "Using %d pixels." % (sum(1.0*greater(wtflat,0)))
        if self.pwid == -1:
           flux = sum(wtflat**2*data.flat)/sum(wtflat**2*timage.flat)
           print "Flux ratio = %.4f" % (flux)
           self.match = flux*timage
        else:
           pful = int(2*self.pwid+1)
           owid = self.pwid + 1.0
           pful2 = pful*pful
           pbox = arange(-self.pwid,self.pwid+0.5,1.0).astype(Int32)
           self.ii = array([i for i in pbox for j in pbox if (i*i+j*j)<=owid**2 or 0])
           self.jj = array([j for i in pbox for j in pbox if (i*i+j*j)<=owid**2 or 0])
           ll = len(self.ii)
           print "Constructing bases."
           it0 = time.time()
           if self.rev:
              f0 = compress(wtflat,wtflat*timage.flat)
              d0 = compress(wtflat,data.flat)
              n0 = compress(wtflat,self.invnoise.flat)
              basisimage = data
           else:
              f0 = compress(wtflat,wtflat*data.flat)
              d0 = compress(wtflat,timage.flat)
              n0 = compress(wtflat,self.invnoise.flat)
              basisimage = timage
           cwt = compress(wtflat,wtflat)
           coy, cox = compress(wtflat,[self.oy.flat,self.ox.flat]).astype(Int32)
           flatimage = basisimage.flat
           basis = [cwt*take(flatimage,(coy+self.jj[k])*self.naxis1+(cox+self.ii[k])) for k in range(ll) \
                  if not sys.stdout.write(".") and not sys.stdout.flush()]
           nb = len(basis)
           if self.skyoff: basis += [ones(f0.shape,Float64)]
           if self.spatial:
               self.oxl = (self.ox-self.naxis1/2.0)/(self.naxis1/2.0)
               woxl = compress(wtflat,self.oxl.flat)
               self.oyl = (self.oy-self.naxis2/2.0)/(self.naxis2/2.0)
               woyl = compress(wtflat,self.oyl.flat)
               xbasis = [b*woxl for b in basis[:nb]]
               ybasis = [b*woyl for b in basis[:nb]]
               basis += xbasis
               basis += ybasis
           it1 = time.time()
           basis = asarray(basis)
           print "Bases constructed in %.4fs." % (it1-it0), basis.shape
           print "Decomposition."
           it0 = time.time()
           du,dv,dw = singular_value_decomposition(transpose(basis))
           it1 = time.time()
           print "Decomposition in %.4fs." % (it1-it0)
           bases = []
           for sol1a in dw:
               psf2 = zeros((pful,pful),Float64)
               for k in range(ll):
                   py = self.jj[k]+self.pwid
                   px = self.ii[k]+self.pwid
                   pr = sqrt(self.jj[k]**2 + self.ii[k]**2)
                   psf2[py,px] = sol1a[k]
               bases.append(psf2)
           bases = asarray(bases)
           npsx = int(sqrt(bases.shape[0]))+1
           npsy = bases.shape[0]/npsx + 1
           imbases = zeros( (npsy*(bases.shape[1]+5), npsx*(bases.shape[2]+5)), Float64)
           imb1 = 0
           for j in range(npsy):
             for i in range(npsx):
                y0 = j*(bases.shape[1]+5)
                y1 = y0 + bases.shape[1]
                x0 = i*(bases.shape[2]+5)
                x1 = x0 + bases.shape[2]
                imbases[y0:y1,x0:x1] = bases[imb1,::-1,::]
                imb1 += 1
                if imb1 == len(bases): break
             if imb1 == len(bases): break
           imbases = imbases[::-1]
           qdump(self.imbases,imbases,self.image)
#           if self.verb: print dv
           sol1 = transpose(dw)
           if self.mcut and self.mcut < len(dv):
               print "Ratio of first eigenvalue with last ten:"
               rdv = dv[0]/dv
               print rdv[-10:]
               qdump("rdv.fits",rdv)
               dv[self.mcut:] = 0 * dv[self.mcut:]
               sol2 = divz(1,dv[:self.mcut])
               sol3 = dot(transpose(du[::,:self.mcut]),f0)
               self.psf1 = dot(sol1[::,:self.mcut], (sol2*sol3))
               fitted = add.reduce(self.psf1[::,NewAxis]*basis,0)
               resid = fitted - f0
               vc = dv[0]/dv[self.mcut-1]
               umean = mean(resid)
               urms = sqrt(mean(power(resid,2)))
               uchi2 = sum(power(resid,2))
               uchi2u = uchi2 / (len(resid)-self.mcut)
               print "(CUT,MEAN,RMS,CHI2,RCHI2) = (%9.1f,%12.6f,%12.6f,%12.6f,%12.6f)" % (vc,umean,urms,uchi2,uchi2u)
           else:
               print "Ratio of first eigenvalue with last ten:"
               rdv = dv[0]/dv
               print rdv[-10:]
               qdump("rdv.fits",rdv)
               cdv = zeros(rdv.shape,Float64)
               if self.vcut == 0: ucv = [0,1]
               else: ucv = [1]
               for cv in ucv:
                  if cv:
                      if self.vcut == 0:
                         qdump("cdv.fits",cdv)
                         tck = splrep(arange(len(dv)),rdv-cdv,k=1,s=0)
                         vloc = sproot(tck,mest=len(dv))
                         vloc = int(vloc[0])
                         self.vcut = 10*rdv[vloc]
                      rdv = [self.vcut]
                      print "Cutting at %1f." % (self.vcut)
                  for jvc in range(len(rdv)):
                      vc = rdv[jvc]
                      udv = 1.0*greater_equal(dv,max(dv)/vc)
                      jcut = int(sum(udv))
                      if cv: print "Using %d out of %d." % (sum(udv),len(udv))
                      sol2 = divz(1,dv[:jcut])
                      sol3 = dot(transpose(du[::,:jcut]),f0)
                      self.psf1 = dot(sol1[::,:jcut], (sol2*sol3))
                      fitted = zeros(f0.shape,Float64)
                      [add(fitted,p1*b1,fitted) for p1,b1 in zip (self.psf1,basis)]
#                      fitted = add.reduce(self.psf1[::,NewAxis]*basis,0)
                      resid = fitted - f0
                      umean = mean(resid)
                      urms = sqrt(mean(power(resid,2)))
                      uchi2 = sum(power(resid,2))
                      uchi2u = uchi2 / (len(resid)-jcut)
                      if not cv: cdv[jvc] = urms
                      print "(CUT,MEAN,RMS,CHI2,RCHI2) = (%9.1f,%12.6f,%12.6f,%12.6f,%12.6f)" % (vc,umean,urms,uchi2,uchi2u)
           del sol1,sol2,sol3,du,dv,dw,basis
           if self.skyoff: print "Sky=",self.psf1[ll]
           fluxrat = sum(self.psf1[:ll])
           qdump("psf1.fits",self.psf1)
           print "Flux ratio=",fluxrat
           if preserve:
              print "Preserving flux?"
              self.psf1[:ll] = self.psf1[:ll] / fluxrat
              fluxrat = sum(self.psf1[:ll])
              print "Flux ratio=",fluxrat
           if self.verb:
             self.psf2 = zeros((pful,pful),Float64)
             pxs = []
             pys = []
             prs = []
             pfs = []
             for k in range(ll):
                 py = self.jj[k]+self.pwid
                 px = self.ii[k]+self.pwid
                 pr = sqrt(self.jj[k]**2 + self.ii[k]**2)
                 self.psf2[py,px] = self.psf1[k]
                 prs.append(pr)
                 pxs.append(self.ii[k])
                 pys.append(self.jj[k])
                 pfs.append( self.psf2[py,px])
             prsf = transpose([pxs,pys,prs,pfs])
             psffile = open("psf.dat","w")
             [psffile.write("%f %f %f %f\n" % tuple(p)) for p in prsf]
             psffile.close()
             for p in self.psf2:
                fmt = pful*"  %7.4f"
                print fmt % tuple(p)
           symm = False
           if symm:
              self.psf2 = (self.psf2+ self.psf2[::-1,::]+ self.psf2[::,::-1]+ self.psf2[::-1,::-1])/4.0
              for k in range(ll):
                  py = self.jj[k]+self.pwid
                  px = self.ii[k]+self.pwid
                  self.psf1[k] = self.psf2[py,px]
           self.match = zeros(self.data.shape,Float64)
           self.grid = zeros(self.data.shape,Float64)
           dgridx, dgridy = int(self.naxis1/32), int(self.naxis2/32)
           gridin = logical_not(fmod(self.ox+dgridx,dgridx*2.0)) * logical_not(fmod(self.oy+dgridy,dgridy*2.0))
           sys.stdout.write("Constructing matched image.")
           it0 = time.time()
           for k in range(ll):
              sys.stdout.write(".");sys.stdout.flush()
              add(self.match, self.component(k,matching=basisimage), self.match)
              add(self.grid, self.component(k,matching=gridin,grid=1), self.grid)
           it1 = time.time()
           print "done.\nConstructed matched image in %.4fs." % (it1-it0)
    def component(self,k,matching,grid=0):
        ll = len(self.ii)
        if self.skyoff: s = 1
        else: s = 0
        ko = k - s*(k>=ll)
        p, i, j = self.psf1[k], self.ii[ko%ll], self.jj[ko%ll]
        if self.spatial:
           if ll <= ko < 2*ll: p = p*self.oxl
           elif 2*ll <= ko < 3*ll: p = p*self.oyl
        if k == ll and self.skyoff: return (not grid and p) or 0.0
        else: return p*QuickShift(matching,i,j)
    def GoCatGo(self,master,verb=0,rev=0,xwin=None,ywin=None,\
        skyoff=0,nonlinear=0,pwid=0,perr=5.0,nmax=40,\
        nord=0,spatial=0,mcut=0,vcut=1e-2,subt=0,registered=0,preserve=0):
        self.master=master
        self.skyoff=skyoff
        self.nonlinear=nonlinear
        self.pwid=pwid
        self.xwin=xwin
        self.ywin=ywin
        self.spatial=spatial
        self.rev=rev
        self.vcut=vcut
        self.mcut=mcut
        self.verb=verb
        if not registered:
           self.readcat()
           self.compute(nmax=nmax)
           self.objmatch(nord=nord,perr=perr)
        self.imread()
        self.mktemplate(registered)
        self.mkweight()
        qdump(self.template,(self.timage-self.tbg)*(self.exptime/self.master.exptime)-self.bg,self.image)
        qdump(self.masked,greater(self.wt,0)*(self.data-self.bg),self.image)
        if subt: qdump(self.difference,(self.data-self.bg)-(self.timage-self.tbg)*(self.exptime/self.master.exptime),self.image)
        self.mkmatch(preserve)
        qdump(self.template,self.match,self.image)
        if self.pwid > -1: qdump(self.psfgrid,self.grid)
        if subt:
           if self.rev: qdump(self.difference,(self.match+self.tbg)-self.timage,self.image)
           else: qdump(self.difference,self.data-(self.match+self.bg),self.image)
        print

