
from basics import *
from sets import Set
from Expector import *
from LDP import *
from FITS import *
from Multipack import *
from Mestimators import *
import os

pkeys = ["subsec1", "subsec1m", "subsec1r", "subsec1b", "subsec1n",
        "subsec1X", "subsec1y", "subsec1w", "subsec1gh", "fitted1"]
dkeys = ["ds","ms","rs","bs","ns",
         "xs","ys","ws","ps","fs"]
ldkeys = dkeys + ["dws","ts","qs"]


def OneIndex(dfile,verb=0):
    d = FITS(dfile,readonly=1)
    objects = []
    while not len(objects) or objects[-1] != "N/A":
       objects.append(d["OBJ%04d" % (len(objects)+1)].strip())
    d.close()
    objects = objects[:-1]
    if verb: print "Read %d object names from %s" % (len(objects),dfile)
    return objects

def ArrayToDict(X,t):
    a = {}
    for jkey,dkey in enumerate(dkeys): a[dkey] = X[jkey]
    a["dws"] = Derivative(a["xs"],a["ws"])
    a["ts"] = t*ones(a["xs"].shape,Float64)
    return a

def Derivative(x,w):
    wa = sort(list(Set(w.flat)))
    xa = array([median(compress(equal(w.flat,w1),x.flat)) for w1 in wa])
    xa,wa = take([xa,wa],argsort(xa),1)
    xtck = splrep(xa,log(wa),s=0,k=3)
    dlwdx = zeros(x.flat.shape,Float64)
    order = argsort(x.flat)
    put(dlwdx,order,splev(take(x.flat,order),xtck,1))
    dws = reshape(dlwdx*w.flat,x.shape)
    return abs(dws)

def ReadOneFile(dfile,objno):
    print "Reading from",dfile,
    qs = None
    if "odc" in dfile:
       qfile = dfile.replace("odc","oqc")
       if os.path.exists(qfile):
          d = FITS(qfile,readonly=1)
          qd = d.data()
          qw = qd[3][objno-1]
          qq = qd[5][objno-1]
          qf = qd[6][objno-1]
          d.close()
          qs = (qw,qq,qf)
    d = FITS(dfile,readonly=1)
    etime = d["exptime"]
    if d["nshuffle"]:
      pa = -(d["rotangle"]-133.55) * pi/180
      nr = d["nod-ra"]/0.2
      nd = d["nod-dec"]/0.2
      nsh = nd*cos(pa) + nr*sin(pa)
      print nr,nd,pa,nsh,
    else: nsh = 0
    if nsh: nf = 2
    else: nf = 1
    hap = d["hap"]
    corners = []
    hdu = nf*(objno-1)+2
    print hdu,
    d.goto(hdu)
    slitobj = d["slitobj"]
    print slitobj,
    rchi2 = d["rchi2"]
    print "%.4f" % (rchi2),
    pm1 = []
    while d["pm1"+str(len(pm1)+1)] != "N/A": pm1.append(d["pm1"+str(len(pm1)+1)])
    pm1 = eval("["+"".join(pm1)+"]")[:1]
    pm2 = []
    while d["pm2"+str(len(pm2)+1)] != "N/A": pm2.append(d["pm2"+str(len(pm2)+1)])
    pm2 = eval("["+"".join(pm2)+"]")[:1]
    pmh = []
    while d["pmh"+str(len(pmh)+1)] != "N/A": pmh.append(d["pmh"+str(len(pmh)+1)])
    pmh = eval("["+"".join(pmh)+"]")
    corners.append(eval(d["corners"]))
    adata = d.data()
    if nsh:
       d.goto(hdu+1)
       slitobj = d["slitobj"]
       print slitobj
       corners.append(eval(d["corners"]))
       bdata = d.data()
    d.close()
#    qdump("/tmp/a.fits",adata[0])
#    qdump("/tmp/b.fits",bdata[0])
#    exit()
    o = {"t":etime,"h":hap,"nf":nf,"nsh":nsh,"corners":corners,"rchi2":rchi2,"qs":qs}
    o["pm1"] = asarray(pm1)
    o["pm2"] = asarray(pm2)
    o["pmh"] = asarray(pmh)
    o["a"] = adata
    if nsh:
       o["b"] = bdata
    return o

def ReadOneObject(object,robjects,objno):
    print object
    robject = robjects[object]
    print robject
#    objno = [1+dobjects[dfile].index(object) for dfile in robject]
    print objno
    data = {}
    for r,o in zip(robject,objno):
       data[r] = ReadOneFile(r,o)
    for r in data:
       d = data[r]
       d["a"] = ArrayToDict(d["a"],d["t"])
       if d["nsh"]: d["b"] = ArrayToDict(d["b"],d["t"])
       if not d["qs"]:
           d["haveqe"] = False
           d["a"]["qs"] = ones(d["a"]["ds"].shape,Float32)
           if d["nsh"]: d["b"]["qs"] = ones(d["a"]["ds"].shape,Float32)
       else:
           d["haveqe"] = True
           ws = d["a"]["ws"]
           if d["nsh"]:
             wsb = d["b"]["ws"]
           qw,qq,qf = d["qs"]
           qd = {}
           [qd.setdefault(qw1,(qq1,qf1)) for qw1,qq1,qf1 in zip(qw,qq,qf)]
           ws0 = ws[0]
           qs0 = array([(ws1 in qd and qd[ws1][0] or 0.0) for ws1 in ws0])
           qf0 = array([(ws1 in qd and qd[ws1][1] or 0.0) for ws1 in ws0])
           qs = asarray(len(ws)*[qs0])
           qf = asarray(len(ws)*[qf0])
           d["a"]["qs"] = qf*qs
           if d["nsh"]:
               qsb = asarray(len(wsb)*[qs0])
               qfb = asarray(len(wsb)*[qf0])
               d["b"]["qs"] = qfb*qsb
#           qdump("/tmp/%sa.fits" % ("ps"),d["a"]["ps"])
#           qdump("/tmp/%sb.fits" % ("ps"),d["b"]["ps"])
#           for k in ["ds","ms","rs","bs","fs","ns"]:
#              d["a"][k] = divz(d["a"][k],qs)
#              if d["nsh"]:
#                  d["b"][k] = divz(d["b"][k],qsb)
    return data

WBinsDict = {}

def GetWBins(nbin=1.0,minbin=0.0,w1=0.0,w2=0.0):
    if (nbin,minbin,w1,w2) not in WBinsDict:
       mw = arange(3800,11.0e3,5.0,Float32)
       if w1: mw = concatenate([mw,[w1]])
       if w2: mw = concatenate([mw,[w2]])
       mw = sort(list(Set(mw)))
       wxbasis = legendre((mw-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
       mx = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
#       for (a,b) in zip(mw,mx): print a,b
       if w1: mx0 = mx[searchsorted(mw,w1)]
       else: mx0 = mx[0]
       mx = mx - mx0
       lmw = log(mw)
       mxtck = splrep(mx,lmw,s=0,k=1)
       mdwdx = splev(mx,mxtck,1)*mw
#       mdwdx = clip(mdwdx,minbin,1e3)
       if w1: bin0 = w1
       else: bin0 = mw[0]+mdwdx[0]/2.0
       if w2: wmax = w2
       else: wmax = LDPDISP["lambda"][-1]+nbin*mdwdx[-1]/2.
       def getx(www,xxx=-1):
           wxbasis = legendre((www-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
           p = (LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ - mx0)
#           raw_input(str((p,xxx)))
           if xxx == -1: return p
           else: return p - xxx
       bins = [bin0]
       xins = [getx(bin0)]
#       print nbin
       while bins[-1] < wmax:
#          print xins
#          print bins,wmax
          new = fsolve(getx,bins[-1],args=(xins[-1]+nbin,))
#          raw_input(str(new))
          xins.append(xins[-1]+nbin)
          bins.append(new)
       wbins = array(bins)
       if w2: wbins = compress(less_equal(wbins,w2),wbins)
       xbins = getx(wbins)
       dbins = nbin*splev(xbins,mxtck,1)*wbins
#       dbins = nbin * clip(dbins,minbin,1e3)
       WBinsDict[(nbin,minbin)] = wbins,dbins
    return WBinsDict[(nbin,minbin)]


def oldGetWBins(nbin=1.0,minbin=0.0,w1=0.0,w2=0.0):
    if (nbin,minbin,w1,w2) not in WBinsDict:
       mw = LDPDISP["lambda"]
       if w1: mw = concatenate([mw,[w1]])
       if w2: mw = concatenate([mw,[w2]])
       mw = sort(list(Set(mw)))
       wxbasis = legendre((mw-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
       mx = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
       mx,mw = take([mx,mw],argsort(mx),1)
       if w1:
          mx0 = mx[searchsorted(mw,w1)]
       else:
          mx0 = mx[0]
       mx = mx - mx0
       lmw = log(mw)
       mxtck = splrep(mx,lmw,s=0,k=3)
       mdwdx = splev(mx,mxtck,1)*mw
       mdwdx = clip(mdwdx,minbin,1e3)
       if w1: bin0 = w1
       else: bin0 = mw[0]+mdwdx[0]/2.0
       if w2: wmax = w2+nbin*mdwdx[searchsorted(mw,w2)]/2.
       else: wmax = LDPDISP["lambda"][-1]+nbin*mdwdx[-1]/2
       bins = [bin0]
       while bins[-1] < wmax:
          bw0 = 1e38
          bw = bins[-1]
          co = 0
          while abs(bw-bw0)/bw > 1e-6:
             bw0 = bw
             co += 1
             if len(bins) > 1:
                wxbasis = legendre((array([bins[-1],bw])-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
                bx = mean(LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ - mx0)
             else:
                wxbasis = legendre((bw-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
                bx = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ - mx0
             dwdxb = splev(bx,mxtck,1)*bw
             dwdxb = clip(dwdxb,minbin,1e3)
             bw = bins[-1]+nbin*dwdxb/2.
             print co,bw,bw0,dwdxb
             if bw > 9000: raw_input(str(bins))
          bins.append( bins[-1]+nbin*dwdxb)
       wbins = array(bins)
       if w2: wbins = compress(less_equal(wbins,w2),wbins)
       wxbasis = legendre((wbins-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
       xbins = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ - mx0
       dbins = splev(xbins,mxtck,1)*wbins
       dbins = nbin * clip(dbins,minbin,1e3)
       WBinsDict[(nbin,minbin)] = wbins,dbins
    return WBinsDict[(nbin,minbin)]

def SingleBin(w,dw,d,h,nsh=1,t=0,M=L2,pw=1,scut=5.0):
    if scut > 100:
       adw = [1.0]
    else:
       adw = [2.0,1.0]
    for jdw,ndw in enumerate(adw):
       keep = less(abs(d["ws"]-w),ndw*dw/2.)
#       print w,ndw*dw/2.,len(keep.flat),sum(keep.flat),
       keep = keep*greater(d["qs"],0)
       if not sum(keep.flat):
           print w,ndw*dw/2.,len(keep.flat),
           print sum(keep.flat),
           print "not enough!"
           return asarray(6*[0.0])
       e0 = {}
       for dkey in ldkeys: e0[dkey] = compress(keep.flat,d[dkey].flat)
       if "yc" in e0:
          keep = less_equal(abs(e0["yc"]),h+0.5)
       else:
          yc = e0["ys"].flat[argmax(e0["ps"])]
          keep = less_equal(abs(e0["ys"]-yc),h+0.5)
       e = {}
       for dkey in ldkeys:
          e[dkey] = compress(keep.flat,e0[dkey].flat)
       dt = e["ts"]
       dd = abs(e["dws"])
       y = e["ys"] - yc
       p = e["ps"]
       f = e["fs"]
       s = e["ds"]
       n = e["ns"]
       qe = e["qs"]
       r0 = (s-f)/n
       er = M(r0/scut,w=1)
       k = 1/(dt*dd*p)
       s = s*k/qe*1e20
       n = n*k/qe*1e20
       v = n**2
       q = divz(1.0,v)
       spec1 = divz(sum(er*q*s),sum(er*q))
       nois1 = divz(1,getsqrt(divz(sum(er*q*(divz(1,v))),sum(er*q))))
       if jdw == 0:
          Me = less(abs(r0),scut)
       else:
#          nfac = bwt((s-spec2/k)/n)[1]
          nfac = 1.0
#          print nfac
#          Me = less(abs(r0),scut)
          Me = Fair((s-spec2/k)/n/nfac,w=1)
#          Me = Huber2((s-spec2/k)/n/nfac,w=1)
#          Me = L1((s-spec2/k)/n/nfac,w=1)
#          Me = L2((s-spec2/k)/n/nfac,w=1)
#          Me = M(r0/scut,w=1)
       qM = power(p,pw)*q
       #qM = q
       spec2 = divz(sum(qM*s),sum(qM))
       nois2 = divz(1,sqrt(divz(sum(qM/v),sum(qM))))
       scat = divz(sum(q*(s-spec2)**2),sum(q))
       scat = sqrt(scat)
    if 5000 < w < 8000 and nois2 == 0:
       print "here",w,dw,e["ys"].shape
       print e["ys"]
    apkeep = less_equal(abs(y),h)
    A = 2.0*h/1e20
    R = A*t
    if nsh != 0: R = 2*R
    spec2 = spec2 * R
    nois2 = nois2 * R
    spec1 = spec1 * R
    nois1 = nois1 * R
    eds,ens,edws,er0 = compress(apkeep,[e["ds"],e["ns"],e["dws"],r0],1)
    eds = eds/edws
    ens = ens/edws
    wt = divz(1,ens**2*(1+er0**2))
    spec0 = divz(sum(wt*eds),sum(wt))*A
    nois0 = sqrt(divz(sum(wt*ens**2),sum(wt))*A)
    pkg = asarray([spec0,nois0,spec1,nois1,spec2,nois2])
    return pkg

def FitOneObject(data,nbin=1,output="Spectra",name=None,inter=0,scut=5.0,w1=0,w2=0,hap=0):
    w1a = w1
    w2a = w2
    ro = data.keys()
    ro.sort()
#    ro = [r for r in ro if data[r]["rchi2"] < 2]
    nsh = data[ro[0]]["nsh"]
    haveqe = data[ro[0]]["haveqe"]
    unsh = [0,1][nsh!=0]
    if not hap: h = mean([data[r]["h"] for r in ro])
    else: h = hap

#    if inter: pdev = "/xs"
#    else: pdev = "%s/%s.ps/vcps"

    npit = 4
#    Mp = Fair
#    Mp = lambda r,w: 1.0/(1+r**2)
    Mp = lambda r,w: 1.0
    if scut > 100:
      pbins = [1.0,1.0,1.0,1.0]
#      pbins = [2.0,1.5,1.0,1.0]
    else:
      pbins = [2.0,1.5,1.0,1.0]
    for pit,pbin,pMp in zip(range(npit),pbins,[Mp,Mp,Mp,Mp]):
       for r in ro:
          for jab,ab in enumerate("ab"[:unsh+1]):
              ds = data[r][ab]["ds"]
              fs = data[r][ab]["fs"]
              ns = data[r][ab]["ns"]
              bs = data[r][ab]["bs"]
              qs = data[r][ab]["qs"]
              mbs = median(bs)
              rs = divz(ds-fs,ns)**2
              brs = divz(bs-mbs,ns)**2
              wrs = greater(rs,scut**2)
#              wrs = logical_or(greater(rs,scut**2),greater(brs,0.1))
#              qdump("/tmp/dbs.fits",divz(bs-mbs,ns))
#              [qdump("/tmp/%s.fits" % (key),data[r][ab][key]) for key in ldkeys]
#              exit()
              ls = 1.0*less(VTKConvolve(1.0*wrs,k=3,numret=1),0.25)
              data[r][ab]["ls"] = ls

       tt = 0
       t = []
       if inter > 1:
          nx=(1+unsh)
          ny=len(ro)
          if ny > 4:
             nx = nx + 2*(1+unsh)
             ny = 6
          if nx==ny==1:psize = 8
          else: psize=min([nx*5,16])
          opendev(PGPLOT_DEV,nx=nx,ny=ny,size=psize,land=0,asp=0.5*ny/nx,ask=1)
          for r in ro:
             for jab,ab in enumerate("ab"[:unsh+1]):
                corners = data[r]["corners"][jab]
                ds = data[r][ab]["ds"]
                fs = data[r][ab]["fs"]
                ps = data[r][ab]["ps"]
                ns = data[r][ab]["ns"]
                ws = data[r][ab]["ws"]
                xs = data[r][ab]["xs"]
                ys = data[r][ab]["ys"]
                ls = data[r][ab]["ls"]
                qs = data[r][ab]["qs"]
                rs = divz(ds-fs,ns)**2
                gz = ps
                wrz = 1/(1+rs)

                for js,us in enumerate([ds,ds-fs]):
                   if js == 0: cpgsvp(0.1,0.95,0.525,0.90)
                   else: cpgsvp(0.1,0.95,0.15,0.525)
                   newbox(add(extrema(xs.flat),[-0.5,2]),add(extrema(ys.flat),[-0.5,-2.0]),ch=2.5,manvp=1,xlab=(js==1),ylab=2,era=(js==0))
                   if js == 0:
                     subtitle("%s (%s)" % (r,ab),ch=2.5)

                   xfm = array([xs.flat[0]-0.5,1.0,0.0,ys.flat[0]-0.5,0.0,1.0])
#                   mds = divz(sum(wrz.flat*gz.flat*ps.flat*ds.flat),sum(wrz.flat*gz.flat*ps.flat))
#                   mds = median(ns.flat)
                   mds = 1.49*median(abs(compress(ps.flat>1e-4,us.flat)))

                   us = where(ns,us,-999)

                   if mds: a1,a2 = -5*mds,5*mds
                   else: a1,a2 = -100,100
                   cpggray(us,us.shape[1],us.shape[0],1,us.shape[1],1,ds.shape[0],a1,a2,xfm)

                   cpgsci(5)
                   cpgsls(4)
                   cont = array([data[r]["pm1"]])
                   cpgcont(ys,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                   cpgsci(1)
                   cpgsls(1)

                   cpgsci(7)
                   cpgsls(2)
                   cont = array([max(ps.flat)/2.0])
                   cpgcont(ps,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                   cpgsci(1)
                   cpgsls(1)

                   if js:
                      cpgsci(8)
                      cpgsls(4)
                      cont = array([w1,5000,6000,7000,8000,9000,1e4,w2])
                      cpgcont(ws,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                      cpgsci(1)
                      cpgsls(1)

                   else:
                      cpgsci(4)
                      cpgsls(1)
                      cont = array([scut**2])
                      cpgcont(rs,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                      cpgsci(1)
                      cpgsls(1)

                      cpgsci(2)
                      cpgsls(1)
                      cont = array([1.0])
                      cpgcont(ls,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                      cpgsci(1)
                      cpgsls(1)


          curs = ["O",(0,0)]
          while curs[0] != "g":
             if curs[0] in "qQI": os._exit(0)
             else: curs = getcurs(curs[1][0],curs[1][1])
          closedev()

       d = {}
       for r in ro:
          for ab in "ab"[:unsh+1]:
             t.append(data[r]["t"])
          d[r] = {}
          for dkey in data[r]["a"]:
             d[r][dkey] = []
             for ab in "ab"[:unsh+1]:
                d[r][dkey] = concatenate([d[r][dkey],data[r][ab][dkey].flat])
          print r, len(d[r]["ps"]),sum(greater(d[r]["ps"],1e-3)),
          print sum(greater(d[r]["ns"],0)),
          print sum(greater(d[r]["ds"],-3*d[r]["ns"])),
          keep = greater(d[r]["ps"],1e-8)*greater(d[r]["ns"],0)*greater(d[r]["ds"],-3*d[r]["ns"])*d[r]["ls"]
          keep = keep*not_equal(d[r]["ds"],0)
          keep = keep*greater(d[r]["qs"],0)
#          keep = keep*less(abs(d[r]["ys"]-data[r]["pm1"]),h+1)
          print sum(keep)
          for dkey in ldkeys: d[r][dkey] = compress(keep,d[r][dkey])
          order = argsort(d[r]["ws"])
          for dkey in ldkeys: d[r][dkey] = take(d[r][dkey],order)
    
       tt = sum(t)
       A = 2.0*h
       print t
#       exit()

       w1b,w2b = extrema([extrema(data[r][ab]["ws"].flat) for r in ro for ab in "ab"[:unsh+1]])

       wbins,dbins = GetWBins(pbin*nbin,minbin=1.0,w1=w1,w2=w2)
       print pMp

       individ = {}
       norms = {}
       scales = {}
       weights = {}
       if pit: pw = 2
       else: pw = 2
       for r in ro:
          individ[r] = transpose([SingleBin(wb,db,d[r],h,nsh=unsh,t=data[r]["t"],M=pMp,pw=pw,scut=scut) for wb,db in zip(wbins,dbins)])
#          print individ[r][4]
#          print individ[r][5]
          for x in range(3):
             individ[r][2*x] = individ[r][2*x]/((unsh+1)*data[r]["t"])
             individ[r][2*x+1] = individ[r][2*x+1]/((unsh+1)*data[r]["t"])
          norms[r] = divz(sum(divz(individ[r][4]*between(wbins,5500.0,8000.0),(1e20*individ[r][5])**2)),\
                          sum(divz(between(wbins,5500.0,8000.0),(1e20*individ[r][5])**2)))

       print [norms[r] for r in ro]
       unorms = [norms[r] for r in ro]
       gnorms = compress(greater(unorms,0),unorms)
       if gnorms: mnorm = median(gnorms)
       else: mnorm = 0.0
       for r in ro:
          norms[r] = (norms[r] > mnorm/4 and norms[r]) or 0.0
       nnorm = [norms[r] for r in ro]
       mnorm = mean(compress(nnorm,nnorm))

       print [norms[r] for r in ro]
       sumweights = 0.0
       nro = 0
       tt = 0.0
       for r in ro:
          if mnorm:
             norms[r] = divz(mnorm,norms[r])
             if norms[r]:
                tt = tt + data[r]["t"]*(unsh+1)
                nro = nro + 1
          else:
             norms[r] = 1.0
          scales[r] = norms[r]*data[r]["t"]*(unsh+1)
#          print individ[r][5]
          weights[r] = divz(1,sqrt(scales[r])*individ[r][5]**2)
#          print weights[r]
          sumweights = sumweights + weights[r]
       print [norms[r] for r in ro]
       sumnorms = sum([norms[r] for r in ro])
       sum2norms = sum([norms[r]**2 for r in ro])
       sumweights = divz(sumweights, nro)
       for r in ro:
          weights[r] = divz(weights[r],sumweights)
       print [median(weights[r]) for r in ro]
       print nro

       print A,tt
       sbins2 = zeros(wbins.shape,Float32)
       ebins2 = zeros(wbins.shape,Float32)
       bbins2 = zeros(wbins.shape,Float32)

       if haveqe and scut < 100:
          YHOrder = int(h/2+1)
          WHOrder = 1
       else:
          YHOrder = 2*int(h-1)
          WHOrder = 2

#       for x in range(3):
#          exec("sbins%d = divz(sum([(weights[r]*scales[r]*individ[r][%d]) for r in ro]),tt)" % (x,2*x))
#          exec("ebins%d = divz(sqrt(sum([weights[r]*(scales[r]*individ[r][%d])**2 for r in ro])),tt)" % (x,2*x+1))
#
       if pbin == 1.0 or 1:
         ds = []
         bs = []
         vs = []
         ns = []
         fs = []
         ps = []
         ws = []
         dws = []
         qs = []
         tx = None
         for r in ro:
            if mnorm and norms[r]:
               print r
               ws2 = []
               dws2 = []
               ds2 = []
               fs2 = []
               bs2 = []
               vs2 = []
               ns2 = []
               ps2 = []
               qs2 = []
               for ab in "ab"[:unsh+1]:
                  if "yc" in data[r][ab]:
                     keep = less_equal(abs(data[r][ab]["yc"].flat),h)
                  else:
                     ws1 = data[r][ab]["ws"].flat
                     ys1 = data[r][ab]["ys"].flat
#                     wxbasis = legendre((ws1-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
#                     dcx0 = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
                     m1 = data[r]["pm1"]
                     if len(shape(m1)) == 0: m1 = [m1]
#                     m1b = polynom(ws1/1e4,len(m1)-1)
#                     m1b = polynom(ws1/1e4,len(m1)-1)
                     m1b = legendre((1e4/ws1-LDPCOEFC)/LDPCOEFS,len(m1)-1)
                     dy1 = dot(m1,m1b)
                     yc1 = ys1 - dy1
                     keep = less_equal(abs(yc1.flat),h)
                  keep = keep * (data[r][ab]["ls"].flat)
                  keep = keep * greater(data[r][ab]["ns"].flat,0.0)
                  keep = keep * greater(data[r][ab]["ws"].flat,0)
                  keep = keep * greater(data[r][ab]["qs"].flat,0)
#                  keep = keep * between(data[r][ab]["ws"].flat,w1,w2,inc=1)
                  keep = keep * greater(data[r][ab]["ps"].flat,1e-8)
                  keep = keep * greater(data[r][ab]["ds"].flat,-3* data[r][ab]["ns"].flat)
                  keep = keep * not_equal(data[r][ab]["ds"].flat,0)
#                  keep = keep * not_equal(data[r][ab]["fs"].flat,0)
#                  keep = keep * less(abs(data[r][ab]["ds"].flat-data[r][ab]["fs"].flat),scut*data[r][ab]["ns"].flat)
                  ds1 = compress(keep,data[r][ab]["ds"].flat)
                  qs1 = compress(keep,data[r][ab]["qs"].flat)
                  bs1 = compress(keep,data[r][ab]["bs"].flat)
                  ns1 = compress(keep,data[r][ab]["ns"].flat)
                  ws1 = compress(keep,data[r][ab]["ws"].flat)
                  dws1 = compress(keep,data[r][ab]["dws"].flat)
                  fs1 = compress(keep,data[r][ab]["fs"].flat)
                  ps1 = compress(keep,data[r][ab]["ps"].flat)
                  ms1 = pMp((ds1-fs1)/(scut*ns1),w=1)
                  ds1 = ds1*norms[r]/ps1/dws1/data[r]["t"]/qs1
                  fs1 = fs1*norms[r]/ps1/dws1/data[r]["t"]/qs1
#                  ns1 = ns1*sqrt(norms[r]/ps1/dws1/data[r]["t"]/qs1)
                  ns1 = ns1*norms[r]/ps1/dws1/data[r]["t"]/qs1
                  bs1 = bs1*norms[r]/dws1/data[r]["t"]/qs1
                  vs1 = ns1**2
                  ns1 = divz(ms1,ns1)
#                  ns1 = ns1*sqrt(data[r]["t"]*qs1*ps1*dws1/norms[r])
                  ws1,dws1,ds1,bs1,vs1,ns1,fs1,ps1,qs1 = take([ws1,dws1,ds1,bs1,vs1,ns1,fs1,ps1,qs1],argsort(ws1),1)
                  if not tx: tx = sort(list(Set(ws1)))
                  ws2 = concatenate([ws2,ws1])
                  dws2 = concatenate([dws2,dws1])
                  qs2 = concatenate([qs2,qs1])
                  ds2 = concatenate([ds2,ds1])
                  fs2 = concatenate([fs2,fs1])
                  bs2 = concatenate([bs2,bs1])
                  vs2 = concatenate([vs2,vs1])
                  ns2 = concatenate([ns2,ns1])
                  ps2 = concatenate([ps2,ps1])
               ws2,dws2,ds2,bs2,vs2,ns2,fs2,ps2,qs2 = take([ws2,dws2,ds2,bs2,vs2,ns2,fs2,ps2,qs2],argsort(ws2),1)
               ws = concatenate([ws,ws2])
               dws = concatenate([dws,dws2])
               qs = concatenate([qs,qs2])
               ds = concatenate([ds,ds2])
               bs = concatenate([bs,bs2])
               vs = concatenate([vs,vs2])
               ns = concatenate([ns,ns2])
               ps = concatenate([ps,ps2])
               fs = concatenate([fs,fs2])

         if len(ws):
            ws,dws,ds,bs,vs,ns,fs,ps,qs = take([ws,dws,ds,bs,vs,ns,fs,ps,qs],argsort(ws),1)
            ws0 = 1*ws
            ws,dws,ds,bs,vs,ns,fs,ps,qs = adjspac([ws,dws,ds,bs,vs,ns,fs,ps,qs])
            xb,xe = extrema(ws)
            print xb,xe
            use = between(ws,w1,w2,inc=1)*greater(ps,1e-4)
            dfit = fs
#            dfit = divz(VTKGauss(ds*ns**2,1*(A*sumnorms/2),numret=1),\
#                      VTKGauss(ns**2,1*(A*sumnorms/2),numret=1))

            if scut < 100: wit = 4
            else: wit = 1
            wns = ns
            kx = 3
            for jm in range(wit):
               dresids = divz(ds-dfit,sqrt(vs))
               dresid = compress(between(ws,w1,w2),dresids)
               rchi2 = sum(dresid**2)/(len(dresid)-len(tx))
               print "Reduced chi^2 =",rchi2
               if jm and scut < 100 and haveqe: wms = Lp(dresids,w=1)
               else: wms = ones(dresids.shape,Float32)
               if haveqe: wns = ns * wms
               else: wns = ns * wms
               wns = where(wns,wns,min(compress(wns,wns))*1e-10)
               tx = wbins
               ntx = diff(searchsorted(ws0,tx))
               tx = wbins[:-1]+dbins[:-1]/2.0
#               tx = wbins[:-1]
               utx = 1.0*greater(ntx,median(compress(ntx,ntx))/2.)
               tx = compress(utx,tx)
               tx = compress(between(tx,xb,xe,inc=1),tx)
               if len(tx) < len(wbins)/2.:
                   print "Giving up! Not enough data in this one!",
                   print len(tx),len(wbins)
                   return
               stck = splrep(ws,ds,w=wns,s=None,task=-1,k=kx,t=tx,xb=xb,xe=xe)
               dfit = splev(ws0,stck)

            dresids = divz(ds-dfit,sqrt(vs))
            dresid = compress(between(ws,w1,w2),dresids)
            rchi2 = sum(dresid**2)/(len(dresid)-len(tx))
            print "Reduced chi^2 =",rchi2

            aresid = sqrt(divz(VTKGauss(wns*dresids**2,2*(A*sumnorms),numret=1), VTKGauss(wns,2.*(A*sumnorms),numret=1)))

            etck = splrep(ws,vs,w=wns,s=None,task=-1,k=kx,t=tx,xb=xb,xe=xe)
            btck = splrep(ws,bs,w=wns,s=None,task=-1,k=kx,t=tx,xb=xb,xe=xe)
#            atck = splrep(ws,aresid,w=ns,s=None,task=-1,k=kx,t=tx,xb=xb,xe=xe)

            sbinss = A*splev(wbins,stck)*between(wbins,xb,xe)
            bbinss = A*splev(wbins,btck)*between(wbins,xb,xe)
#            afit = splev(wbins,atck)*between(wbins,xb,xe)
            efit = splev(wbins,etck)*between(wbins,xb,xe)

            ebinss = A*getsqrt(efit/A/sumnorms)

#            ebinss = sqrt(1./(sumnorms-1.))*getsqrt(A*efit)
#            rbinss = divz(sbinss-VTKMedian(sbinss,5,1,1),ebinss)
#            print sum(rbinss**2)/len(rbinss)

#            ebinss = A*getsqrt(afit*splev(wbins,etck)*between(wbins,xb,xe)/sumnorms/A)

#            ebinss = sqrt((sumnorms-1)/(sumnorms+1.))*getsqrt(efit*between(wbins,xb,xe)/A)/sumnorms
#            ebinss = sqrt((sumnorms-1)/(sumnorms+1.))*getsqrt(efit)
#            ebinss = getsqrt(A*afit*efit)
#            ebinss = afit

            gtx = greater(VTKImageShift(concatenate([utx,[utx[-1]]]),-0.5,0.0,numret=1,interp=1),0.25)

            sbins2 = sbinss*gtx
            ebins2 = ebinss*gtx
            bbins2 = bbinss*gtx

       else:
          stck = splrep(wbins,sbins2/dbins,s=0,k=3)

       for r in ro:
          if mnorm and norms[r]:
             d[r]["fs0"] = d[r]["fs"]
             ps = d[r]["ps"]
             rfs = (data[r]["t"]) * d[r]["dws"] * splev(d[r]["ws"],stck) * d[r]["qs"]
             if norms[r]: rfs = rfs/norms[r]
             d[r]["fs"] = rfs * ps
             for jab,ab in enumerate("ab"[:unsh+1]):
                ws = data[r][ab]["ws"].flat
                ss = zeros(ws.shape,Float64)
                order = argsort(ws.flat)
                put(ss,order,splev(take(ws,order),stck))
                ss = ss * between(ws,w1,w2,inc=1)
                ss = reshape(ss,data[r][ab]["ds"].shape)
                ss = ss * data[r][ab]["dws"] * data[r]["t"]
                ss = ss * data[r][ab]["qs"]
                data[r][ab]["rfs"] = ss
                if norms[r]: data[r][ab]["rfs"] = data[r][ab]["rfs"] / norms[r]
                data[r][ab]["fs0"] = data[r][ab]["fs"]
                data[r][ab]["fs"] = data[r][ab]["rfs"] * data[r][ab]["ps"]
             if pit < npit-1:
                print "Let's try a new set of moments for",r
                ws = concatenate([data[r][ab]["ws"].flat for ab in "ab"[:unsh+1]])
                dws = concatenate([data[r][ab]["dws"].flat for ab in "ab"[:unsh+1]])
                xs = concatenate([data[r][ab]["xs"].flat for ab in "ab"[:unsh+1]])
                ys = concatenate([data[r][ab]["ys"].flat for ab in "ab"[:unsh+1]])
                ds = concatenate([data[r][ab]["ds"].flat for ab in "ab"[:unsh+1]])
                bs = concatenate([data[r][ab]["bs"].flat for ab in "ab"[:unsh+1]])
                ns = concatenate([data[r][ab]["ns"].flat for ab in "ab"[:unsh+1]])
                ls = concatenate([data[r][ab]["ls"].flat for ab in "ab"[:unsh+1]])
                rs = concatenate([data[r][ab]["rfs"].flat for ab in "ab"[:unsh+1]])
                fs = concatenate([data[r][ab]["fs"].flat for ab in "ab"[:unsh+1]])
                ps = concatenate([data[r][ab]["ps"].flat for ab in "ab"[:unsh+1]])
                qs = concatenate([data[r][ab]["qs"].flat for ab in "ab"[:unsh+1]])
                keep = ls*greater(ds,-3*ns)*greater(ns,0)*not_equal(ds,0)*not_equal(qs,0)
                ws,dws,xs,ys,ds,bs,ns,fs,ps,rs,qs = [compress(keep,a) for a in (ws,dws,xs,ys,ds,bs,ns,fs,ps,rs,qs)]
                order = argsort(ws)
                ws,dws,xs,ys,ds,bs,ns,fs,ps,rs,qs = [take(a,order) for a in (ws,dws,xs,ys,ds,bs,ns,fs,ps,rs,qs)]
                es = divz((ds-fs)/ns)

#                wxbasis = legendre((ws-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
#                dcx0 = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ

#                m1b = polynom(ws/1e4,len(data[r]["pm1"])-1)
                m1b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,len(data[r]["pm1"])-1)
                m1y1 = dot(data[r]["pm1"],m1b)

                wt = getsqrt(ds)*less(es**2,scut**2)*less(abs(ys-m1y1),h+1)*between(ws,w1,w2,inc=1)

                if scut > 100: W1Order = max([1,len(data[r]["pm1"]) - 1])
                else: W1Order = max([0,len(data[r]["pm1"]) - 1])
#                m1b = polynom(ws/1e4,W1Order)
                m1b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W1Order)
                m1 = svdfit(transpose(m1b)*wt[::,NewAxis],ys*wt)
                print m1
                data[r]["pm1"] = m1
                dy1 = dot(data[r]["pm1"],m1b)
                yc = ys - dy1

#                wt = getsqrt(ds)*less(es**2,scut**2)*less(abs(yc),min([h+1,2.0*data[r]["pm2"][0]]))
                wt = getsqrt(ds)*less(es**2,scut**2)*between(ws,w1,w2,inc=1)
#                swt = wt*less(abs(yc),h+1)
                swt = wt*less(abs(yc),min([h+1,2.5*data[r]["pm2"][0]]))
                if scut > 100: W2Order = max([1,len(data[r]["pm2"]) - 1])
                else: W2Order = max([0,len(data[r]["pm2"]) - 1])
#               m2b = polynom(ws/1e4,W2Order)
                m2b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W2Order)
                for s2p in range(10):
                   m2 = svdfit(transpose(m2b)*swt[::,NewAxis],power(yc,2)*swt)
                   sm2 = clip(getsqrt(dot(m2,m2b)),1,999)
                   swt = wt*less(abs(yc),2.5*sm2)
                data[r]["pm2"] = m2
                dy2 = clip(getsqrt(dot(m2,m2b)),1,999)
                print m2,quartiles(dy2)
                yn = divz(yc,dy2)
                gy = exp(-0.5*power(yn,2))
                mh = None
#                WHOrder,YHOrder,mh = 0,0,array([sqrt(pi)])

#                try:
                if 1:
                   wt = between(ws,w1,w2,inc=1)*not_equal(ds,0.0)*less(es**2,scut**2)*less(abs(yc),h)/ns
#                   wt = wt*less(abs(yn),2.5)
                   print h,YHOrder
                   mhyb = hermite(yn,YHOrder,norm=1)
                   uds = ds
                   if scut < -100e99:
                      mhyb = mhyb[1:]
                      uds =  (ds-rs*sqrt(pi)*gy)
#                   mhwb = polynom(ws/1e4,WHOrder)
                   mhwb = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,WHOrder)
#                   rs = splev(ws,stck)*dws*qs*(data[r]["t"]/norms[r])
                   rgw = rs*gy*wt
                   mhb = [(rgw*hyb1*hwb1) for hyb1 in mhyb for hwb1 in mhwb]
                   mh = svdfit(transpose(mhb),uds*wt)
#                   mhb = [(hyb1*hwb1) for hyb1 in mhyb for hwb1 in mhwb]
#                   gh = dot(mh,mhb)*gy
#                   if scut < -100e99: gh = gh + sqrt(pi)*gy
#                except:
#                   mh = None
                data[r]["pmh"] = mh
                print mh

                for ab in "ab"[:unsh+1]:
                   ws = data[r][ab]["ws"].flat
                   ys = data[r][ab]["ys"].flat
#                   wxbasis = legendre((ws-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
#                   dcx0 = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
                   m1b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W1Order)
#                   m1b = polynom(ws/1e4,W1Order)
                   dy1 = dot(data[r]["pm1"],m1b)
                   yc = ys - dy1
                   data[r][ab]["yc"] = reshape(yc, data[r][ab]["ys"].shape)
#                   m2b = polynom(ws/1e4,W2Order)
                   m2b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W2Order)
                   dy2 = clip(getsqrt(dot(data[r]["pm2"],m2b)),1,999)
                   yn = divz(yc,dy2)
                   gy = exp(-0.5*power(yn,2))
                   ps = gy
                   if data[r]["pmh"]:
                      mhyb = hermite(yn,YHOrder,norm=1)
                      if scut < -100e99: mhyb = mhyb[1:]
#                      mhwb = polynom(ws/1e4,WHOrder)
                      mhwb = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,WHOrder)
                      mhb = [(gy*hyb1*hwb1) for hyb1 in mhyb for hwb1 in mhwb]
                      ps = dot(data[r]["pmh"],mhb)
                      if scut < -100e99: ps = ps + sqrt(pi)*gy
                   ps = reshape(ps, data[r][ab]["ps"].shape)
                   yc = reshape(yc, data[r][ab]["ps"].shape)
#                   qdump("psold.fits",data[r][ab]["ps"])
#                   qdump("psnew.fits",ps)
                   data[r][ab]["ps"] = ps

                   ws = data[r][ab]["ws"][0]
                   wdir = median(diff(ws)) < 0
                   if wdir: ws = ws[::-1]
                   rfs = splev(ws,stck)
                   if wdir: rfs = rfs[::-1]
                   rfs = rfs*between(ws,xb,xe,inc=1)*data[r][ab]["dws"][0]*data[r][ab]["qs"][0]*data[r]["t"]/norms[r]
                   fs = rfs[NewAxis,::] * data[r][ab]["ps"]
                   use = between(data[r][ab]["ws"],w1,w2,inc=1)
                   data[r][ab]["fs"] = fs*use
                   use = use*less_equal(abs(yc),h)*data[r][ab]["ls"]*data[r][ab]["ns"]
                   x = divz(data[r][ab]["ds"] - data[r][ab]["fs"], data[r][ab]["ns"]).flat
                   x = compress(use.flat,x)
                   x2 = sum(power(x,2))
                   ndeg = len(x.flat)-len(tx)-len(data[r]["pmh"])-len(data[r]["pm1"])-len(data[r]["pm2"])
                   print x2,len(x.flat),ndeg,x2/ndeg
#                   fs =  data[r][ab]["fs"]
#                   qdump("data.fits",data[r][ab]["ds"])
#                   qdump("modelold.fits",data[r][ab]["fs0"])
#                   qdump("model.fits",data[r][ab]["fs"])
##                   qdump("model.fits",fs)
#                   qdump("noise.fits",data[r][ab]["ns"]*reshape(less(abs(yc),h+1),data[r][ab]["ns"].shape))
#                   qdump("wave.fits",data[r][ab]["ws"])
#                   data[r][ab]["fs"] = fs
#                   data[r][ab]["fs"] = data[r][ab]["rfs"] * data[r][ab]["ps"]

       if pbin == 1.0 and nro and 1:
          model = {}
          dymod = 1.0
          for ab in "ab"[:unsh+1]:
#             ws = data[r][ab]["ws"]
#             ys = data[r][ab]["ys"]
             ws = wbins
             ys = arange(-2*h,2*h,dymod,Float32)
             ws,ys = ws[NewAxis,::]*ones(ys.shape)[::,NewAxis],ys[::,NewAxis]*ones(ws.shape)[NewAxis,::]
             ws = ws.flat
             yc = ys.flat
#             wxbasis = legendre((ws-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
#             dcx0 = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
             W2Order = len(data[r]["pm2"]) - 1
             m2b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W2Order)
#             m2b = polynom(ws/1e4,W2Order)
             pm2 = sum([norms[r]*data[r]["pm2"] for r in ro if norms[r]])
             dy2 = clip(getsqrt(dot(data[r]["pm2"],m2b)),1,999)
             yn = divz(yc,dy2)
             gy = exp(-0.5*yn**2)
             ps = gy
             pmh = sum([norms[r]*data[r]["pmh"] for r in ro if norms[r] if data[r]["pmh"]])
             if pmh:
                mhyb = hermite(yn,YHOrder,norm=1)
                if scut < -100e99: mhyb = mhyb[1:]
#                mhwb = polynom(ws/1e4,WHOrder)
                mhwb = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,WHOrder)
                mhb = asarray([(gy*hyb1*hwb1) for hyb1 in mhyb for hwb1 in mhwb])
                ps = dot(pmh,mhb)
                if scut < -100e99: ps = ps + sqrt(pi)*gy
             ps = reshape(ps, ys.shape)
#             ws = data[r][ab]["ws"][0]
#             qs = data[r][ab]["qs"][0]
#             wdir = median(diff(ws)) < 0
#             dws = data[r][ab]["dws"][0]
#             ws,dws = take([ws,dws],argsort(ws),1)
#             tu = data[r]["t"]
             rfs = splev(wbins,stck)
#             rfs = rfs*between(ws,xb,xe,inc=1)*qs*dws
             fs = rfs[NewAxis,::] * ps
#             if wdir: fs = fs[::,::-1]
#             ds = data[r][ab]["ds"]
#             qdump("model.fits",fs*between(ws,w1,w2))
#             qdump("data.fits",ds)
#             qdump("model.fits",where( between(data[r][ab]["ys"],-h,h,inc=1), data[r][ab]["ds"], -9999))
#             qdump("data.fits",where( between(data[r][ab]["ys"],-h,h,inc=1), data[r][ab]["ds"], -9999))
             sbins2 = 2*h*divz(sum(fs*dymod),sum(ps*dymod))

       if inter:
          opendev("/xs",nx=1,ny=2,size=9,land=0,asp=1.0)

          newbox([LDPDISP["lambda"][0],LDPDISP["lambda"][-1]],getmar((dbins*sbins2+dbins*ebins2),50,150),ch=2)
          xlabel("\gl (\A)",ch=2)
          ylabel("PV weighted Counts/s",ch=2)
          subtitle("%s (p=%d)" % (name,1+pit),ch=2)
          rci = 1
          for r in ro:
             if norms[r]:
                rci += 1
                wp,dwp,sp,dsp = compress(weights[r],[wbins,dbins,dbins*individ[r][4],dbins*individ[r][5]],1)
                drawpt(wp,norms[r]*sp,ps=1,ch=2,ci=rci)
                drawpterr(wp,dwp/2.,norms[r]*sp,norms[r]*dsp,ch=2,ci=rci)
          drawpt(wbins,dbins*sbins2,ps=1,ch=2,ci=1)
          drawpterr(wbins,dbins/2.,dbins*sbins2,dbins*ebins2,ch=2,ci=1)

          newbox([LDPDISP["lambda"][0],LDPDISP["lambda"][-1]],getmar((sbins2+2*ebins2)/2.,50,400),ch=2)
          xlabel("\gl (\A)",ch=2)
          ylabel("PV weighted Counts/A/s",ch=2)
          rci = 1
          for r in ro:
             if norms[r]:
                rci += 1
                wp,dwp,sp,dsp = compress(weights[r],[wbins,dbins,individ[r][4],individ[r][5]],1)
                drawpt(wp,norms[r]*sp,ps=1,ch=2,ci=rci)
                drawpterr(wp,dwp/2.,norms[r]*sp,norms[r]*dsp,ch=2,ci=rci)
          drawpt(wbins,sbins2,ps=1,ch=2,ci=1)
          drawpterr(wbins,dbins/2.,sbins2,ebins2,ch=2,ci=1)

          if inter > 0:
             curs = ["O"]
             while curs[0] != "g":
                if curs[0] in "qQI": os._exit(0)
                else: curs = getcurs(0,0)
          closedev()

    results = {}
    results["w"] = wbins
    results["dw"] = dbins
    results["s"] = sbins2*tt
    results["ds"] = ebins2*tt
    results["b"] = bbins2*tt
    results["tt"] = tt
    results["A"] = A

#    if not haveqe:
#       sbins2 = divz(sbins2*dbins)
#       ebins2 = divz(ebins2*dbins)
#       bbins2 = divz(bbins2*dbins)

    D = asarray([wbins,dbins,sbins2,ebins2,bbins2])

    naxes = array(D.shape)[::-1].tolist()
    F = FITS("%s/%s.fits" % (output,name),create=1)
    F.newhead("T",-32,len(naxes),naxes)
    F.modkey("object", name)
    F.modkey("exptime", tt)
    F.modkey("aperture", A)
    F.modkey("mnorm", mnorm)
    [F.modkey("norm%d" % (j+1), "(%s,%f)" % (r,norms[r])) for j,r in enumerate(ro)]
    F.putdata(D)
    F.close()

    return results

def PullOneGalaxy(object,robjects,objno,nbin=1,output="Spectra",inter=False,scut=5,w1=0,w2=0,hap=0):
   data = ReadOneObject(object,robjects,objno)
   t0 = time.time()
   FitOneObject(data,nbin=nbin,output=output,name=object,inter=inter,scut=scut,w1=w1,w2=w2,hap=hap)
   t1 = time.time()
   print "Elapsed time was %.3fs" % (t1-t0)

