#!/usr/bin/env python

from math import atan2
from sys import stdout
from os import remove
import time
import tempfile
from ephem import *
from masks import *
from MLab import *
from basics import *
from VTKHelperFunctions import *
from FITS import *
from NEWFIRM import JTRANS
import SimpleQueue
from sets import Set

NSCALE = 0.4

class newfirm:
    def __init__(self,record,keys):
       simplify = lambda l: l.replace(" ",",")
       record = simplify(record)
       simplify = lambda l: l.replace(",,",",")
       while record.count(",,"):
          record = simplify(record)
       record = record.replace(",',","',")
       record = record[:-3]+"'"
       record = record.replace("'0,[s]'","0.0")
       record = record.replace(",[s]'","'")
       record = record.split(",")
       record[1:] = map(eval, record[1:])
       self.info = {}
       [self.info.setdefault(key,rec) for rec,key in zip(record,keys)]
       self.base = os.path.basename(self.info["Filename"]).replace(".fits","")
       self.name = os.path.basename(self.info["Filename"])
       self.bdir = os.path.dirname(self.info["Filename"])
       self.ra = hrrad(scanSexagesimal(self.info["RA"]))
       self.dec = degrad(scanSexagesimal(self.info["DEC"]))
       rao = self.info["RAOFF"]
       deco = self.info["DECOFF"]
       rint = 60.0
       ira0 = int((raddeg(self.ra)*3600-rao)/rint+0.5)
       idec0 = int((raddeg(self.dec)*3600-deco)/rint+0.5)
       self.ra0 = ira0*rint/3600.0/15.0
       self.dec0 = idec0*rint/3600.0
       self.RA0 = formatSexagesimal(self.ra0,4,3600).strip()
       self.DEC0 = formatSexagesimal(self.dec0,4,3600).strip()
       if self.DEC0[0] != "-": self.DEC0 = "+"+ self.DEC0
       if self.RA0.find(":") == 1: self.RA0 = "0"+self.RA0
       self.type = self.info["OBSTYPE"]
       self.mjd = self.info["MJD-OBS"]
       self.when = int(self.mjd)
       if self.type =="flat": pref = "flt"
       elif self.type =="object": pref = "obj"
       elif self.type =="focus": pref = "foc"
       else: pref = "unk"
       if pref == "obj": self.object = "%s%s%s_%s" % (pref,self.RA0,self.DEC0,self.when)
       elif pref == "flt": self.object = self.info["OBJECT"].lower()
       elif pref == "foc": self.object = "focus"
       self.filter = self.info["FILTER"]
       self.jno = int(self.base[-4:])
       self.exptime = self.info["EXPTIME"]
       self.asecs = self.info["RAOFF"]
       self.dsecs = self.info["DECOFF"]
       self.newfile = self.base
       self.skys=[]
       self.groupt=[]
       self.groups=[]
       self.groupe=[]
       self.groupx=[]
       self.groupy=[]
       self.refgroups=[]
       self.sets=0

    def __repr__(self): return self.newfile

    def order(self,no):
       self.order = no

    def group(self,t,refdir=None):
       # obsoleted! use FindGroups!
       dx,dy= normalcoord(degrad(t.ra) % (2*pi),degrad(t.dec),\
               degrad(self.ra) % (2*pi),degrad(self.dec), conv=180.0*3600.0/pi)
       m0 = abs(dx) < 300
       m1 = abs(dy) < 300
       m2 = (self.object == t.object)
       m3 = (self.filter == t.filter)
       if m0 and m1 and m2 and m3:
           if refdir: self.refgroups.append( refdir+"/"+t.newfile)
           else:
              self.groups.append( t.newfile)
              self.groupe.append( t.exptime)
              self.groupx.append( -dx/NSCALE)
              self.groupy.append( dy/NSCALE)

    def clean(self,t):
       if t != self and t.groups == self.groups: t.groups=[]

    def set(self,t):
       m0 = less(abs(self.mjd - t.mjd),0.4)
       m1 = (self.type == t.type)
       m2 = (self.object == t.object)
       m3 = (self.filter == t.filter)
       if m0 and m1 and m2 and m3: self.sets += 1

def readdb(db):
   f=open(db,"r")
   df = f.readlines()
   f.close()
   header = df[1][1:].replace(" ","").replace("''","','")
   keys = eval(header)
   j = nonzero([(l[0] not in ["#","["]) for l in df])
   df = [df[k] for k in j]
   t = [newfirm(df[j],keys) for j in range(len(df))]
   return t

def reduce2groups(t):
   [[i.clean(x) for x in t] for i in t]
   u = nonzero([i.groups != [] for i in t])
   t = [t[j] for j in u]
   return t

def CollateDB(db,targ,filt):
  print "Reading database",db
  t = readdb(db)

  targets = list(Set([i.object for  i in t]))
  targets.sort()

  groups = {}
  for target in targets:
    u = nonzero([(target in i.object) for i in t])
    tu = [t[j] for j in u]
    if filt: tu = [i for i in tu if i.filter == filt]
    if not targ or targ == target:
        print "Target %30s [%5d]" % (target,len(tu))
    [[i.set(x) for x in tu] for i in tu]
    [tu[j].order(j) for j in range(len(tu))]
    groups[target] = tu
  if targ in groups:
     return groups[targ]
  else:
     raise "Target Not Found!"

if __name__ == "__main__":
    db = argv[1]
    if argv[2:]: targ = argv[2]
    else: targ = None

    CollateDB(db,targ,None)

