#!/usr/bin/env python

from basics import *
from scipy.special import *
from MLab import diff
from Mestimators import *
from sets import Set
import time
from VTKHelperFunctions import *
from FITS import *
from pgplot import *
from NEWFIRMDB import *
from SimpleQueue import *
#from TwistedClient import *
from masks import *
import cPickle

nfdbfile = os.getenv("NEWFIRMDB")
nftarget = os.getenv("NEWFIRMTARG")
nffilter = os.getenv("NEWFIRMFILTER")

if nfdbfile:
   NEWFIRMDataset = CollateDB(nfdbfile,nftarget,nffilter)
   if nftarget and os.path.basename(os.getcwd()) != nftarget:
      if not os.path.exists(nftarget):
          os.mkdir(nftarget)
      nfmods = glob("NEWFIRM*.py*")
      flats = glob("Flatc?.fits")
      masks = glob("maskc?.fits")
      print "Changing to directory",nftarget
      os.chdir(nftarget)
      links = [".Twisted","sex"]+nfmods+[nfdbfile,nfdbfile.replace(".db","")]
      links = links + flats + masks
      for l in links:
#        if os.path.exists(l): os.remove(l)
        if os.path.exists("../"+l) and not os.path.exists(l):
            print "Creating symlink to",l
            os.symlink("../"+l,l)

SEXINTKEYS = ["NUMBER","VECTOR_ASSOC"]

def readsex(filename,verb=0):
   fu = open(filename,"r")
   icol = []
   keys = []
   fmts = []
   data = {}
   n=0
   for l in fu.xreadlines():
     if l[0] == "#":
       w = split(l)
       icol.append(int(w[1])-1)
       keys.append(w[2])
       if w[2] in SEXINTKEYS: fmts.append("%6d")
       elif w[2] == "FLAGS": fmts.append("%4d")
       elif w[2] == "CLASS_STAR": fmts.append("%5.2f")
       elif w[2] == "X_IMAGE": fmts.append("%9.3f")
       elif w[2] == "Y_IMAGE": fmts.append("%9.3f")
       else: fmts.append("%9.4f")
       data[w[2]] = []
       n+=1
     else:
       w = split(l)
       [data[keys[j]].append(w[j]) for j in range(n)]
   fu.close()
   if verb:
     print "Found Keys="
     print keys
     print "Number of Objects = %d" % ( len(data[keys[0]]))
   for k in keys:
     data[k] = array([[float,int][k in SEXINTKEYS](d) for d in data[k]])
   return data

class Exposure:
   def __init__(self,N):
      self.N = N
      self.fits = N.info["Filename"]
      self.name = N.name.replace(".fits","")
      self.raoff = N.info["RAOFF"]
      self.decoff = N.info["DECOFF"]
      self.base = N.base
      self.expno = N.jno
      self.mfits1 = "mode1/"+N.name.replace(".fits","cD.fits")
      self.mfits2 = "mode2/"+N.name.replace(".fits","cD.fits")
      self.sfits1 = "sky1/"+N.name.replace(".fits","cD.fits")
      self.sfits2 = "sky2/"+N.name.replace(".fits","cD.fits")
      self.mask1 = "mask1/"+N.name.replace(".fits","cD.fits")
      self.pers1 = "pers1/"+N.name.replace(".fits","cD.fits")
      self.cat1 = "cat1/"+N.name.replace(".fits","cD.fits").replace(".fits",".cat")
      self.cat2 = "cat2/"+N.name.replace(".fits","cD.fits").replace(".fits",".cat")
      self.scale = NSCALE
   def __repr__(self):
      return self.base
   def ReadModes(self,first=True,second=False,data=0):
      chips = [1,2,3,4]
      m = asarray([self.ReadMode(chip,first=first,second=second,data=data) for chip in chips])
      return m
   def ReadMode(self,chip,first=True,second=False,data=0):
      if first:
        mfits = self.mfits1.replace("D",str(chip))
      elif second:
        mfits = self.mfits2.replace("D",str(chip))
      G = FITS(mfits)
      cent = G["mode"]
      if cent == "N/A": cent = G["median"]
      scat = G["scat"]
      G.close()
      if not data: return cent,scat
      if data:
         G = FITS(self.fits)
         d = G.data()
         G.close()
         return cent,scat,d
   def Job(self,func,args):
      os.nice(15)
      chips = [1,2,3,4]
      dicts = [{"args":(self.N,chip),"keywords":args} for chip in chips]
      if "verb" in args and args["verb"]:
         [apply(func,dict["args"],dict["keywords"]) for dict in dicts]
      else:
         print func.__name__
         self.queue = TwistedObjectQueue(func.__name__,dicts)
         self.queue.repeatable = False
         self.queue.Run(forks=False,sleeptime=0.02)
         print self.queue.results
   def Modes(self,first=True,second=False,msh=10,cat=None,histo=0,mag=3,verb=1):
      args = {"first":first,"second":second,"msh":msh,"cat":cat,"mag":mag,
        "histo":histo,"verb":verb}
      self.Job(NEWFIRMMode,args)
   def Mode(self,chip,first=True,second=False,msh=10,cat=None,histo=0,mag=3,verb=1):
      start = time.time()
      F = FITS(self.fits)
      F.goto(1+chip)
      data = F.data()
      data = data[:2048,:2048].astype(Float32)
      F.close()
      flatf = "Flatc%d.fits" % (chip)
      maskf = "maskc%d.fits" % (chip)
      if os.path.exists(maskf):
         F = FITS(maskf)
         maskd = F.data()
         F.close()
      else:
         maskd = None
      if os.path.exists(flatf):
         F = FITS(flatf)
         flatd = F.data()
         F.close()
         data = divz(data,flatd)
      if cat:
        mask,pers = DrawMask(chip,cat,self,data.shape,mag=mag,verb=verb)
        if not os.path.exists("mask1"): os.mkdir("mask1")
        if not os.path.exists("pers1"): os.mkdir("pers1")
        sfits1 = self.sfits1.replace("D",str(chip))
        mask1 = self.mask1.replace("D",str(chip))
        pers1 = self.pers1.replace("D",str(chip))
        print "Writing %s" % (mask1)
        G = FITS(mask1,create=1,template=sfits1)
        G.putdata(mask)
        G.close()
        print "Writing %s" % (pers1)
        G = FITS(pers1,create=1,template=sfits1)
        G.putdata(pers)
        G.close()
        mask = logical_and(mask,pers)
      else:
        mask = not_equal(data,0)
      if maskd: mask = logical_and(mask,maskd)
      if (first and not histo) or not msh:
         kind = "median"
         cent, scat = self.QuickMedian(data.flat,mask=mask,verb=verb)
      else:
         kind = "mode"
         cent, scat = self.FindMode(data.flat,msh=msh,mask=mask,histo=histo,verb=verb)
      finish = time.time()
      ctime = finish-start
      if first:
        mfits = self.mfits1.replace("D",str(chip))
        if not os.path.exists("mode1"): os.mkdir("mode1")
      elif second:
        mfits = self.mfits2.replace("D",str(chip))
        if not os.path.exists("mode2"): os.mkdir("mode2")
      print "Writing %s[%d,%d] using (%s,scat) = (%.1f,%.1f), computed in %.2fs" % \
        (mfits,data.shape[1],data.shape[0],kind,cent,scat,ctime)
      G = FITS(mfits,create=1,readonly=0)
      F = FITS(self.fits)
      G.copyheader(F)
      G.delkey("extend")
      G.delkey("nextend")
      F.goto(1+chip)
      bad = ["xtension","bitpix","naxis","naxis1","naxis2","pcount","gcount",
        "extname","inherit","extver",]
      [G.modkey(key,F[key]) for key in F.keys() if key not in bad]
      F.close()
      G.modkey("bitpix",-32)
      G.resize( [data.shape[1], data.shape[0]])
      G.modkey(kind,cent)
      G.modkey("scat",scat)
      mdata = VTKDivide(data,cent)
      G.putdata(mdata)
      G.close()
      return cent,scat
   def QuickMedian(self,s,mask=None,verb=0):
      if mask: s = compress(mask.flat,s)
      s = sort(s)
      me = s[len(s)/2]
      sc = (s[int(0.84*len(s))] - s[int(0.16*len(s))])/2.
      return me, sc
   def FindMode(self,s,th=0,sat=65e3,mask=None,msh=10,verb=0,histo=0):
      mode = 0.0
      scat = 0.0
      s = compress(greater(abs(s),th),s)
      if sat: s = compress(less(s,sat),s)
      if mask: s = compress(mask.flat,s)
      s = sort(s)
      if not msh: msh=10
      if histo:
         me = s[len(s)/2]
         ms = sqrt(me)
         ml = me-3.0*ms
         mh = me+1.0*ms
         md = ms/msh
         bins = arange(ml,mh,md,Float)
         mi = diff(searchsorted(s,bins))
         mb = diff(bins)/2. + bins[:-1]
         lmi = log(mi)
         bases = polynom(mb,2)
         sol = linear_least_squares(transpose(bases),lmi)[0]
         print sol
         scat = sqrt(-0.5/sol[2])
         mode = pow(scat,2)*sol[1]
         print me,mode,scat
         if verb:
            lmi1 = max(lmi)
            mi1 = max(mi)
            xm=[mb[0],mb[-1]]
            ym=[lmi1-2,lmi1+0.3]
            newbox(xm,ym,ch=2.0)
            xlabel("Counts",ch=2.0)
            ylabel("ln N",ch=2.0)
            drawln(mb,lmi,ls=1,hist=1)
            drawln([mode,mode],ym,ls=1)
            drawln([mode-scat,mode-scat],ym,ls=2)
            drawln([mode+scat,mode+scat],ym,ls=2)
            raw_input("hit <cr> ")
      else:
         x = VTKShrinkN(0.5+arange(len(s)),msh,usemean=1,numret=1)/msh
         s = VTKShrinkN(s,msh,usemean=1,numret=1)
         p = x/(1.*len(s))
         m = s[3*len(s)/8]
         n = ndtri(p)
#         n = (pdtri(m,1-p)-m)/sqrt(m)
         q = (s[int(0.84*len(s)+0.5)]-s[int(0.16*len(s)+0.5)])/2.
         guess = [m,q]
         w = exp(-0.5*(p-0.5)**2/pow(0.25,2))
         def fitp(g):
           y = g[0]+g[1]*n
           r = s-y
           u = r*w
           if verb: print g,len(s)
           return u
         result = leastsq(fitp,guess)
         if verb: print result
         mode,scat = result[0]
      return mode,scat
   def SkySubtracts(self,NDS,first=1,second=0,nf=3,verb=0,flat=0):
      args = {"NDS":NDS,"first":first,"second":second,"nf":nf,"verb":verb,"flat":flat}
      self.Job(NEWFIRMSkySubtract,args)
   def SkySubtract(self,chip,Exposures,first=1,second=0,nf=3,verb=0,flat=0):
      start = time.time()
      if first:
         if not os.path.exists("sky1"): os.mkdir("sky1")
         mfits = self.mfits1.replace("D",str(chip))
         sfits = self.sfits1.replace("D",str(chip))
         ne = [j for j in range(len(Exposures)) if Exposures[j].base == self.base][0]
         if flat:
            ne30 = len(Exposures)
            skyframe = "sky1/Flatc%d.fits" % (chip)
         else:
            ne30 = (ne / 30 * 30) + 1
            skyframe = "sky1/FirstSky%04dc%d.fits" % (ne30,chip)
         nf = 0
         if (ne+1) == ne30: await = False
         else: await = True
         if await: print "Need to await",skyframe
      elif second:
         await = False
         if not os.path.exists("sky2"): os.mkdir("sky2")
         mfits = self.mfits2.replace("D",str(chip))
         sfits = self.sfits2.replace("D",str(chip))
         pfits = self.pers1.replace("D",str(chip))
         skyframe = "sky2/SecondSky%04dc%d.fits" % (self.expno,chip)
      if first and (await or os.path.exists(skyframe)):
        print "Using",skyframe
        while not os.path.exists(skyframe):
           time.sleep(1.0)
           print "Waiting for",skyframe
        ssize = 0
        while ssize != os.path.getsize(skyframe):
           ssize = os.path.getsize(skyframe)
           time.sleep(0.1)
           print "Waiting for completion of",skyframe
        F = FITS(skyframe)
        S = F.data()
        F.close()
        use = None
      elif nf:
        sfits1 = self.sfits1.replace("D",str(chip))
        ne = [j for j in range(len(Exposures)) if Exposures[j].base == self.base][0]
        F = FITS(sfits1)
        xpos1 = F["xdiff1"]
        ypos1 = F["ydiff1"]
        F.close()
        sfitsE = lambda E: E.sfits1.replace("D",str(chip))
        F = [FITS(sfitsE(E)) for E in Exposures]
        xdiff1 = array([f["xdiff1"] for f in F])
        ydiff1 = array([f["ydiff1"] for f in F])
        [f.close() for f in F]
        ds = sqrt((xpos1-xdiff1)**2 + (ypos1-ydiff1)**2)
        off = nonzero(greater(ds,2.5))
        neo = clip(searchsorted(off,ne),0,len(off)-1)
        off = off[max([0,neo-nf]):min([neo+nf+1,len(off)])]
        use = [Exposures[o] for o in off]
      elif flat:
        use = Exposures
      else:
        use = Exposures[ne30-1:ne30-1+30]
      if use:
         print "Computing",skyframe
         print "From",use
         S = self.MedianFrames(chip,use,first,second,verb=verb)
         if flat:
            S = S * between(S,0.5,1.2)
            S1 = abs(S-1)
            Ss = VTKMedian(S1,3,3,numret=1)
            Sg = greater(divz(S1,Ss),5).astype(Float32)
            Sgx = add.reduce(Sg,0)
            Sgy = add.reduce(Sg,1)
            Sgx1,Sgx2 = bwt(Sgx)
            Sgy1,Sgy2 = bwt(Sgy)
            Sg = ones(S.shape,Float)*less(Sgx,Sgx1+5.0*Sgx2)[NewAxis,::]
            Sg = Sg*less(Sgy,Sgy1+5.0*Sgy2)[::,NewAxis]
            S = S * Sg
         G = FITS(skyframe,create=1,template=mfits)
         G.modkey("ncombine",len(use))
         G.putdata(S)
         G.close()
      cent,scat = self.ReadMode(chip,first=first,second=second)
      F = FITS(mfits)
      D = F.data()
      G = less(D,1e24)
      F.close()
      D = cent*(D-S)*G
      if second:
        F = FITS(pfits)
        D = D*F.data()
        F.close()
      G = FITS(sfits,create=1,template=mfits)
      G.putdata(D)
      G.close()
      finish = time.time()
      ctime = finish-start
      print "Writing %s, computed in %.2fs" % (sfits,ctime)
   def MedianFrames(self,chip,Frames,first,second,yskip=32,verb=0):
      mfits = lambda f: (first and f.mfits1) or (second and f.mfits2)
      F = [FITS(mfits(f).replace("D",str(chip))) for f in Frames]
      M = zeros((F[0]["naxis2"],F[0]["naxis1"]),Float32)
      if F[0]["mode"] == "N/A":
         Fmodes = array([f["median"] for f in F])
      else:
         Fmodes = array([f["mode"] for f in F])
      Fno = array([f.expno for f in Frames])
      Fnos = sort(list(Set(Fno)))
      if second:
         K = [FITS(f.mask1.replace("D",str(chip))) for f in Frames]
         L = [FITS(f.pers1.replace("D",str(chip))) for f in Frames]
      Ny = (M.shape[0]-1)/yskip+1
      for j in range(Ny):
         j0 = j*yskip
         j1 = min([j0+yskip,M.shape[1]])
         Ma = asarray([f.readsec([0,j0],[M.shape[1]-1,j1-1]) for f in F])
         if second:
            Mb = asarray([f.readsec([0,j0],[M.shape[1]-1,j1-1]) for f in K])
            Mb = Mb*asarray([f.readsec([0,j0],[M.shape[1]-1,j1-1]) for f in L])
            Ma = Mb*Ma
         Mg = not_equal(Ma,0.0)
         Ma = sort(where(Mg,Ma,1e18),0)
         Mo = add.reduce(less(Ma,1e16),0).astype(Int32)
         Mo2i = (Mo/2.).astype(Int32)
         Moi = (2*Mo2i).astype(Int32)
         Meven = equal(Mo,Moi)
         Modd = not_equal(Mo,Moi)
         aF = arange(len(F))[::,NewAxis,NewAxis]*ones(Ma.shape)
         Reven = add.reduce(Ma * Meven[NewAxis,::,::] * between(aF,Mo2i-1,Mo2i,inc=1),0)/2.
         Rodd = add.reduce(Ma * Modd[NewAxis,::,::] * equal(aF,Mo2i),0)
         Mj = where(Modd,Rodd,Reven)*greater(Mo,0)
#         if second:
#            Mn = getsqrt(Ma/Fmodes[::,NewAxis,NewAxis])
#            Md = array([newdivz(Ma[Jf]-Ma[Jf-1],sqrt(2)*Mn[Jf]) for Jf in range(len(F))])
#            Mb = Mb/sqrt(1 + Md**2)
#            Mj = newdivz(sum(Mb*Ma),sum(Mb))
         M[j0:j1] = Mj.astype(Float32)
         sys.stdout.write(".")
         sys.stdout.flush()
      print
      [f.close() for f in F]
      if second:
         [f.close() for f in K]
         [f.close() for f in L]
      return M
   def Shift(self,Exposures,first=1,second=0,verb=0):
      ne = [j for j in range(len(Exposures)) if Exposures[j].name == self.name][0]
      if ne:
         ref = Exposures[ne-1]
         refim = Exposures[ne-1].name
         yoff1 = self.raoff
         xoff1 = -self.decoff
         yoff2 = ref.raoff
         xoff2 = -ref.decoff
# THIS IS THE OPPOSITE SIGN OF WHAT IS EXPECTED
         xo = -(xoff2-xoff1)
         yo = -(yoff2-yoff1)
         xo0 = xo
         yo0 = yo
         ux1,uy1,uf1,um1,ura1,udec1 = [[] for j in range(6)]
         ux2,uy2,uf2,um2,ura2,udec2 = [[] for j in range(6)]
         ax1,ay1,af1,am1,ara1,adec1 = [[] for j in range(6)]
         ax2,ay2,af2,am2,ara2,adec2 = [[] for j in range(6)]
         for chip in [1,2,3,4]:
            if first:
               if not os.path.exists("cat1"): os.mkdir("cat1")
               cat = self.cat1.replace("D",str(chip))
               sfits = self.sfits1.replace("D",str(chip))
            elif second:
               if not os.path.exists("cat2"): os.mkdir("cat2")
               cat = self.cat2.replace("D",str(chip))
               sfits = self.sfits2.replace("D",str(chip))
            if os.path.exists(cat):
               f = open(cat,"r")
               c = f.read()
               f.close()
               if c: runsex = False
               else: runsex = True
            else: runsex = True
            if runsex: Sextractor(sfits,cat,scale=self.scale,thresh=3,saturate=10e3)
            if first:
              rcat = ref.cat1.replace("D",str(chip))
              rfits = ref.sfits1.replace("D",str(chip))
            elif second:
              rcat = ref.cat2.replace("D",str(chip))
              rfits = ref.sfits2.replace("D",str(chip))
            sex1 = readsex(cat)
            sex2 = readsex(rcat)
            x1,y1 = sex1["X_IMAGE"],sex1["Y_IMAGE"]
            x2,y2 = sex2["X_IMAGE"],sex2["Y_IMAGE"]
            w1 = sex1["FWHM_IMAGE"]
            w2 = sex2["FWHM_IMAGE"]
            a1 = sex1["A_IMAGE"]
            a2 = sex2["A_IMAGE"]
            b1 = sex1["B_IMAGE"]
            b2 = sex2["B_IMAGE"]
            f1 = sex1["FLAGS"]
            f2 = sex2["FLAGS"]
            m1 = sex1["MAG_APER"]
            m2 = sex2["MAG_APER"]
            d1 = sex1["MAG_BEST"] - m1
            d2 = sex2["MAG_BEST"] - m2
            inside1 = between(x1,100,1948)*between(y1,100,1948)
            inside2 = between(x2,100,1948)*between(y2,100,1948)
            inside1 = inside1*less(f1,16)*between(w1,3,30)*less(d1,-0.75)
            inside2 = inside2*less(f2,16)*between(w2,3,30)*less(d2,-0.75)
#            inside1 = inside1*greater(a1,1.5)*greater(b1,1.5)
#            inside2 = inside2*greater(a2,1.5)*greater(b2,1.5)
            x1,y1,f1,m1 = compress(inside1, [x1,y1,f1,m1], 1)
            x2,y2,f2,m2 = compress(inside2, [x2,y2,f2,m2], 1)
            soln = join([os.environ["PYTHONHOME"],"Database","NEWFIRMc%d.soln" % (chip)],"/")
            ra1,dec1 = xy2rd(soln,x1,y1,text=0)
            ra2,dec2 = xy2rd(soln,x2,y2,text=0)
            use1 = argsort(m1)[:15]
            use2 = argsort(m2)[:15]
            vx1,vy1,vf1,vm1,vra1,vdec1 = take([x1,y1,f1,m1,ra1,dec1],use1,1)
            vx2,vy2,vf2,vm2,vra2,vdec2 = take([x2,y2,f2,m2,ra2,dec2],use2,1)
            ux1,uy1,uf1,um1,ura1,udec1 = [concatenate([a,b]) for a,b in \
                    zip([ux1,uy1,uf1,um1,ura1,udec1], [vx1,vy1,vf1,vm1,vra1,vdec1])]
            ux2,uy2,uf2,um2,ura2,udec2 = [concatenate([a,b]) for a,b in \
                    zip([ux2,uy2,uf2,um2,ura2,udec2], [vx2,vy2,vf2,vm2,vra2,vdec2])]
            ax1,ay1,af1,am1,ara1,adec1 = [concatenate([a,b]) for a,b in \
                    zip([ax1,ay1,af1,am1,ara1,adec1], [x1,y1,f1,m1,ra1,dec1])]
            ax2,ay2,af2,am2,ara2,adec2 = [concatenate([a,b]) for a,b in \
                    zip([ax2,ay2,af2,am2,ara2,adec2], [x2,y2,f2,m2,ra2,dec2])]
         cra = mean(extrema(ura1))
         cdec = mean(extrema(udec1))
         ux1,uy1 = normalcoord(ura1,udec1,cra,cdec,conv=3600*180/pi)
         ux2,uy2 = normalcoord(ura2,udec2,cra,cdec,conv=3600*180/pi)
         x1,y1 = normalcoord(ara1,adec1,cra,cdec,conv=3600*180/pi)
         x2,y2 = normalcoord(ara2,adec2,cra,cdec,conv=3600*180/pi)
         th = 2.0
         dx1 = (ux1[::,NewAxis] - ux1[NewAxis,::])
         dy1 = (uy1[::,NewAxis] - uy1[NewAxis,::])
         dx2 = (ux2[::,NewAxis] - ux2[NewAxis,::])
         dy2 = (uy2[::,NewAxis] - uy2[NewAxis,::])
         ds1 = sqrt(dx1**2+dy1**2)
         ds2 = sqrt(dx2**2+dy2**2)
         dt1 = arctan2(dy1,dx1)*180/pi
         dt2 = arctan2(dy2,dx2)*180/pi
         ds = ds1[::,::,NewAxis,NewAxis] - ds2[NewAxis,NewAxis,::,::]
         dt = dt1[::,::,NewAxis,NewAxis] - dt2[NewAxis,NewAxis,::,::]
         th2d = clip(0.01*th*ds2[NewAxis,NewAxis,::,::],2.0,1e38)
         uds1 = less(abs(ds),th2d)
         uds2 = less(abs(dt),0.5)
         uds = uds1*uds2
#         dx = dx1[::,::,NewAxis,NewAxis] - dx2[NewAxis,NewAxis,::,::]
#         dy = dy1[::,::,NewAxis,NewAxis] - dy2[NewAxis,NewAxis,::,::]
#         ds = sqrt(dx**2+dy**2)
#         ds = between(abs(dx),0,th) * between(abs(dy),0,th)
         ns = add.reduce(add.reduce(uds,3),1)
#         for ns1 in ns: print ns1
#         print
#         kx = abs((ux1+xo)[::,NewAxis,] - ux2[NewAxis,::])
#         ky = abs((uy1+yo)[::,NewAxis,] - uy2[NewAxis,::])
#         ks = less(kx,60)*less(ky,60)
#         ns = ns*ks
#         for ns1 in ns: print ns1
         mns = ns.shape[0]/4
         mns = 0.75*max(compress(ns.flat,ns.flat))
         jm,km = transpose([(j,argmax(ns1)) for j,ns1 in enumerate(ns) if max(ns1) >= mns])
#         print take(ux2,km)-take(ux1+xo,jm)
#         print take(uy2,km)-take(uy1+yo,jm)
#         print ns.shape
         dx = take(ux2,km)-take(ux1,jm)
         dy = take(uy2,km)-take(uy1,jm)
         xb = bwt(dx)
         yb = bwt(dy)
         xo,xr = xb
         yo,yr = yb
         lmat = len(dx)
         print "%3d %4.1f %3d %10.4f+-%.4f %10.4f+-%.4f %s %s" % (0,th,lmat,xo,xr,yo,yr,ref.base,self.base)
         th = 10
         for niter in range(5):
            dx = (x2[::,NewAxis]-(x1+xo)[NewAxis,::])
            dy = (y2[::,NewAxis]-(y1+yo)[NewAxis,::])
            ds = sqrt(dx**2 + dy**2)
            okay = less(ds,th)
            j1 = argmin(ds,1)
            m2 = [j2 for j2 in range(len(x2)) if okay[j2,j1[j2]]]
            m1 = take(j1,m2)
            x1m = take(x1,m1)
            y1m = take(y1,m1)
            x2m = take(x2,m2)
            y2m = take(y2,m2)
            xb = bwt(x2m-x1m)
            yb = bwt(y2m-y1m)
            xo,xr= xb
            yo,yr= yb
            lmat = len(x2m)
            print "%3d %4.1f %3d %10.4f+-%.4f %10.4f+-%.4f %s %s" % (niter+1,th,lmat,xo,xr,yo,yr,ref.base,self.base)
            th = max([6,4*sqrt(xb[1]**2+yb[1]**2)])
            if verb:
               PGPLOT_DEV = os.getenv("PGPLOT_DEV")
               if (PGPLOT_DEV and PGPLOT_DEV[-2:] == "ps" and niter==4) \
                       or (PGPLOT_DEV and PGPLOT_DEV[-2:] != "ps"):
                   newbox(extrema(x1),extrema(y1),ch=1.5)
                   xlabel("X",ch=1.5)
                   ylabel("X",ch=1.5)
                   title = "%s-%s [%d]" % (ref.base,self.base,chip)
                   subtitle(title,ch=1.5)
                   drawpt(x2,y2,ps=17,ch=3,ci=2)
                   drawpt(x1+xo,y1+yo,ps=4,ch=2,ci=4)
                   drawpt(x1,y1,ps=4,ch=1.5,ci=7)
                   drawpt(x1+xo0,y1+yo0,ps=17,ch=1,ci=5)
         if verb:
            PGPLOT_DEV = os.getenv("PGPLOT_DEV")
            if PGPLOT_DEV and PGPLOT_DEV[-2:] != "ps": raw_input("hit <cr>")
      else:
         for chip in [1,2,3,4]:
            if first:
               if not os.path.exists("cat1"): os.mkdir("cat1")
               cat = self.cat1.replace("D",str(chip))
               sfits = self.sfits1.replace("D",str(chip))
            elif second:
               if not os.path.exists("cat2"): os.mkdir("cat2")
               cat = self.cat2.replace("D",str(chip))
               sfits = self.sfits2.replace("D",str(chip))
            if os.path.exists(cat):
               f = open(cat,"r")
               c = f.read()
               f.close()
               if c: runsex = False
               else: runsex = True
            else: runsex = True
            if runsex: Sextractor(sfits,cat,scale=self.scale,thresh=3,saturate=10e3)
         xo = 0.0
         yo = 0.0
         xr = 0.0
         yr = 0.0
         lmat = 0
         refim = Exposures[0].name
      for chip in [1,2,3,4]:
         if first:
            if not os.path.exists("cat1"): os.mkdir("cat1")
            cat = self.cat1.replace("D",str(chip))
            sfits = self.sfits1.replace("D",str(chip))
         elif second:
            if not os.path.exists("cat2"): os.mkdir("cat2")
            cat = self.cat2.replace("D",str(chip))
            sfits = self.sfits2.replace("D",str(chip))
         if os.path.exists(sfits):
            F = FITS(sfits,readonly=0)
            F.modkey("refim1",refim)
            F.modkey("lmatch1",lmat)
            F.modkey("xdiff1",xo)
            F.modkey("ydiff1",yo)
            F.modkey("xscat1",xr)
            F.modkey("yscat1",yr)
            F.close()

def Sextractor(image,catalog,sexdir="./sex/",scale=0.2,thresh=3.0,saturate=55e3,seeing=1.3):
      sexcom = ["sex",image]
      sexcom += ["-c "+sexdir+"default.sex"]
      sexcom += ["-PARAMETERS_NAME "+sexdir+"default.param"]
      sexcom += ["-WEIGHT_TYPE BACKGROUND"]
      sexcom += ["-DETECT_MINAREA 5"]
      sexcom += ["-DETECT_THRESH "+str(thresh)]
      sexcom += ["-ANALYSIS_THRESH "+str(thresh)]
      sexcom += ["-PIXEL_SCALE %f" % (scale)]
      sexcom += ["-SEEING_FWHM %f" % (seeing)]
      sexcom += ["-CATALOG_NAME /tmp/temporary.cat"]
      sexcom += ["-FILTER_NAME "+sexdir+"gauss_3.0_5x5.conv"]
      sexcom += ["-STARNNW_NAME "+sexdir+"default.nnw"]
      sexcom += ["-SATUR_LEVEL "+str(saturate)]
      verbex(" ".join(sexcom))
      verbex("mv /tmp/temporary.cat "+catalog)

def Offsets( Exposures, chip):
   mXs = []
   mYs = []
   sXs = []
   sYs = []
   xscale = Exposures[0].scale
   yscale = Exposures[0].scale
   sfits = lambda E,c: E.sfits1.replace("D",str(c))
   for c in [1,2,3,4]:
      G = [FITS(sfits(E,c)) for E in Exposures]
      gX = array([g["xdiff1"] for g in G])/xscale
      gY = array([g["ydiff1"] for g in G])/yscale
      gXs = add.accumulate(gX)
      gYs = add.accumulate(gY)
      [g.close() for g in G]
      mX0,mX1 = extrema(gXs)
      mY0,mY1 = extrema(gYs)
      sX = mX1 - mX0
      sY = mY1 - mY0
      sXs.append(sX)
      sYs.append(sY)
      mXs.append(mX0)
      mYs.append(mY0)
   aX = argmax(sXs)
   aY = argmax(sYs)
   sX = sXs[aX]
   mX0 = mXs[aX]
   sY = sYs[aY]
   mY0 = mYs[aY]
   F = [FITS(sfits(E,chip)) for E in Exposures]
   X = array([f["xdiff1"] for f in F])/xscale
   Y = array([f["ydiff1"] for f in F])/yscale
   Xs = add.accumulate(X)
   Ys = add.accumulate(Y)
   mX0,mX1 = extrema(Xs)
   mY0,mY1 = extrema(Ys)
   sX = mX1-mX0
   sY = mY1-mY0
   Xz = Xs - mX0
   Yz = Ys - mY0
   [f.close() for f in F]
   return Xz,Yz,Xs,Ys

def Job(func,args):
   os.nice(15)
   chips = [1,2,3,4]
   dicts = [{"args":(chip,),"keywords":args} for chip in chips]
   print func.__name__
   queue = TwistedObjectQueue(func.__name__,dicts)
   queue.repeatable = False
   queue.Run(forks=False,sleeptime=0.02)
   print queue.results

def Stacks(NDS,yskip=32,usemed=1,first=1,second=0,name="FirstStack.fits",justsex=0):
   args = {"NDS":NDS,"yskip":yskip,"first":first,"second":second,"name":name,"justsex":justsex,"usemed":usemed}
   Job(NEWFIRMStack,args)

def NEWFIRMStack(chip,NDS=[],yskip=32,usemed=1,first=1,second=0,name="FirstStackcD.fits",justsex=0):
   Exposures = [Exposure(N1) for N1 in NDS]
   if first:
     sfits = lambda E,c: E.sfits1.replace("D",str(c))
     mfits = lambda E,c: E.mfits1.replace("D",str(c))
   elif second:
     sfits = lambda E,c: E.sfits2.replace("D",str(c))
     mfits = lambda E,c: E.mfits2.replace("D",str(c))
     pfits = lambda E,c: E.pers1.replace("D",str(c))
   F = [FITS(sfits(E,chip)) for E in Exposures]
   exptimes = [f["exptime"] for f in F]
   [f.close() for f in F]
   cname = name.replace("D",str(chip))
   ename = cname.replace(".fits","map.fits")
   sname = cname.replace(".fits","sig.fits")
   if not justsex:
      if 1:
         xscale = Exposures[0].scale
         yscale = Exposures[0].scale
         mXs = []
         mYs = []
         sXs = []
         sYs = []
         for c in [1,2,3,4]:
            G = [FITS(sfits(E,c)) for E in Exposures]
            gX = array([g["xdiff1"] for g in G])/xscale
            gY = array([g["ydiff1"] for g in G])/yscale
            gXs = add.accumulate(gX)
            gYs = add.accumulate(gY)
            [g.close() for g in G]
            mX0,mX1 = extrema(gXs)
            mY0,mY1 = extrema(gYs)
            sX = mX1 - mX0
            sY = mY1 - mY0
            sXs.append(sX)
            sYs.append(sY)
            mXs.append(mX0)
            mYs.append(mY0)
         aX = argmax(sXs)
         aY = argmax(sYs)
         sX = sXs[aX]
         sY = sYs[aY]
         mX0 = mXs[aX]
         mY0 = mYs[aY]
         s = add((2048,2048),(sX,sY)).astype(Int32)+1
         print  "Image size = %r" % (s)
         stack = zeros(s,Float32) + 1e38
         emap = zeros(s,Float32)
         smap = zeros(s,Float32)
         if os.path.exists(cname): os.remove(cname)
         if os.path.exists(ename): os.remove(ename)
         if os.path.exists(sname): os.remove(sname)
         solnfile = join([os.environ["PYTHONHOME"],"Database","NEWFIRMc%d.soln" % (chip)],"/")
         solndat = readsoln(solnfile)
         solndat["ic"] = invsoln(solndat,2048,2048,verb=1)
         F = [FITS(sfits(E,chip)) for E in Exposures]
         X = array([f["xdiff1"] for f in F])/xscale
         Y = array([f["ydiff1"] for f in F])/yscale
         M = array([f["median"] for f in F])
         ra0 = hrrad(scanSexagesimal(F[0]["ra"]))
         dec0 = degrad(scanSexagesimal(F[0]["dec"]))
         Xd = add.accumulate(X)
         Yd = add.accumulate(Y)
         Xs = (Xd - mX0)
         Ys = (Yd - mY0)
         print "Creating",cname
         So = FITS(cname,create=1,template=F[0].name())
         So.modkey("exptime",1.0)
         print s
         So.resize([s[1],s[0]])
         So.putdata(stack)
         print "Creating",ename
         Eo = FITS(ename,create=1,template=F[0].name())
         Eo.modkey("exptime",1.0)
         Eo.resize([s[1],s[0]])
         Eo.putdata(emap)
         print "Creating",sname
         Vo = FITS(sname,create=1,template=F[0].name())
         Vo.modkey("exptime",1.0)
         Vo.resize([s[1],s[0]])
         Vo.putdata(smap)
         y,x = indices(stack.shape,Float32)
         start = time.time()
         Ny = stack.shape[0]/yskip+1
#         mF = [FITS(mfits(E,chip)) for E in Exposures]
#         if second:
#            pF = [FITS(pfits(E,chip)) for E in Exposures]
         sys.stdout.flush()
         for j in range(Ny):
            j0 = j*yskip
            j1 = min([j0+yskip,stack.shape[0]])
            if j1>j0:
#               print j0,j1,stack.shape
               dm = []
               em = []
               um = []
               vm = []
               xi0 = (x[j0:j1]-Xs[0])-solndat["center"][0]
               yi0 = (y[j0:j1]-Ys[0])-solndat["center"][1]
               ui0,vi0 = mbasis(solndat["c"],xi0,yi0)
               for i,f in enumerate(F):
                  if exptimes[i]:
                     Xd1 = Xd[i]*xscale*pi/180/3600
                     Yd1 = Yd[i]*yscale*pi/180/3600
                     ui = ui0 - Xd1
                     vi = vi0 - Yd1
                     xi,yi = mbasis(solndat["ic"],ui,vi)
                     xi = xi + solndat["center"][0]
                     yi = yi + solndat["center"][1]
                     x0,x1 = clip(add([0,1],extrema(xi.flat)),0,2047).astype(Int32)
                     y0,y1 = clip(add([0,1],extrema(yi.flat)),0,2047).astype(Int32)
                     d = f.readsec([x0,y0],[x1,y1])
                     d = where(between(d,-1e12,1e12),d,0.0)
                     mf = FITS(mfits(E,chip))
                     nd = M[i]*mf.readsec([x0,y0],[x1,y1])
                     mf.close()
                     d = divz(d,exptimes[i]).astype(Float32)
#                     if second:
#                        pf = FITS(pfits(E,chip))
#                        pd = pf.readsec([x0,y0],[x1,y1])
#                        pf.close()
                     v = (not_equal(d,0)*getsqrt(8*nd + 25**2)/8.).astype(Float32)
                     e = (not_equal(d,0)*exptimes[i]).astype(Float32)
                     ny,nx = d.shape
                     xi0 = (xi+0.5-x0).astype(Int32).flat
                     yi0 = (yi+0.5-y0).astype(Int32).flat
                     c0 = (yi0*nx + xi0)
                     t0 = clip(c0,0,len(d.flat)-1)
                     o0 = reshape(between(c0,0,len(d.flat)-1),y[j0:j1].shape)
                     d0 = reshape(take(d.flat,t0),y[j0:j1].shape)
                     v0 = reshape(take(v.flat,t0),y[j0:j1].shape)
                     e0 = reshape(take(e.flat,t0),y[j0:j1].shape)
                     u0 = between(xi,0,2047)*between(yi,0,2047)*not_equal(d0,0)*o0
                     u0 = u0.astype(Float32)
                     dm.append(d0)
                     em.append(e0)
                     um.append(u0)
                     vm.append(divz(1.0*v0,exptimes[i]))
               dm = asarray(dm)
               em = asarray(em)
               um = asarray(um)
               vm = asarray(vm)
#               qdump("/tmp/dm.fits",dm)
#               qdump("/tmp/vm.fits",vm)
#               print sort(dm.flat)[:10], sort(dm.flat)[-10:]
#               print sort(vm.flat)[:10], sort(vm.flat)[-10:]
               ums = sum(um)
               if usemed:
                  if 1:
                     dms = sort(where(um,dm,1e18),0)
                     Mo = sum(less(dms,1e16)).astype(Int32)
                     mMo = int(max(Mo.flat))
                     if mMo:
                        dms = dms[:mMo]
                        Mo2i = (Mo/2.).astype(Int32)
                        Moi = (2*Mo2i).astype(Int32)
                        Meven = equal(Mo,Moi)
                        Modd = not_equal(Mo,Moi)
                        aF = arange(len(dms))[::,NewAxis,NewAxis]
                        Reven = add.reduce(dms * Meven[NewAxis,::,::] * between(aF,Mo2i-1,Mo2i,inc=1),0)/2.
                        Rodd = add.reduce(dms * Modd[NewAxis,::,::] * equal(aF,Mo2i),0)
                        mm = where(Modd,Rodd,Reven)*greater(Mo,0)
                        sm = sqrt(divz(sum(um*vm**2),sum(um)*sum(um)))
                        em = sum(em*um)
                     else:
                        mm = zeros(dms[0].shape,Float32)
                        sm = zeros(dms[0].shape,Float32)
                        em = zeros(dms[0].shape,Float32)
                  elif 0:
                     mm = divz(sum(um*dm),sum(um))
                     sm = sqrt(divz(sum(um*vm**2),sum(um)*sum(um)))
                     em = sum(em*um)
                  if second:
                     rm = dm-mm[NewAxis,::,::]
#                     km = less(abs(rm),5*vm)
                     wm = divz(1,vm**2)
                     km = Lp(divz(rm,vm),w=1)
                     um = um*not_equal(km,0)
                     mm = divz(sum(dm*um*km*wm),sum(um*km*wm))
                     em = divz(sum(em*um*km*wm),sum(um*km*wm))
                     sm = sqrt(divz(sum((dm-mm)**2*um*km*wm),sum(um*km*wm)*sum(um*km)))
               else:
                  wm = divz(1,vm**2)
                  mm = divz(sum(dm*um*wm),sum(um*wm))
                  em = divz(sum(em*um*wm),sum(um*wm))
                  sm = sqrt(divz(sum((dm-mm)**2*um*wm),sum(um*wm)*sum(um)))
               mm = mm*greater(ums,5)
               sm = sm*greater(ums,5)
               em = em*greater(ums,5)
               stack[j0:j1] = mm.astype(Float32)
               emap[j0:j1] = em.astype(Float32)
               smap[j0:j1] = sm.astype(Float32)
               if j1:
                   dt = (time.time()-start)
                   print "Elapsed=%.2fs Remain=%.2fs" % (dt,(stack.shape[0]-j1)*dt/j1)
                   sys.stdout.flush()
#                   raw_input(" %d %d" % (j0,j1))
#               print j0,j1,stack[j0:j1].shape,[0,j0],[stack.shape[1]-1,j1]
               So.writesec(stack[j0:j1],[0,j0],[stack.shape[1]-1,j1-1])
               Eo.writesec(emap[j0:j1],[0,j0],[stack.shape[1]-1,j1-1])
               Vo.writesec(smap[j0:j1],[0,j0],[stack.shape[1]-1,j1-1])
         print
         finish = time.time()
         print "Total=%.2fs" % (finish-start)
         So.close()
         Eo.close()
         Vo.close()
         [f.close() for f in F]
#         [f.close() for f in mF]
#         if second: [f.close() for f in pF]
         sys.stdout.flush()
   catalog = cname.replace(".fits",".cat")
   saturate = 10e3 * len(Exposures) / sum(exptimes)
   Sextractor(cname,catalog,scale=NSCALE,saturate=saturate,thresh=2.5,seeing=1.4)

def DrawMask(chip,cat,E,size,mag=3.0,verb=0):
   mask = ones(size,Float32)
   pers = ones(size,Float32)
   Xz = cat["Xz"]
   Yz = cat["Yz"]
   Xs = cat["Xs"]
   Ys = cat["Ys"]
   Es = cat["exp"]
   dat = readsex( cat["cat"])
   cnaxis1,cnaxis2 = cat["naxes"]
   Ei = [i for i,E1 in enumerate(Es) if E1.name == E.name][0]
   xscale = E.scale
   yscale = E.scale
   solnfile = join([os.environ["PYTHONHOME"],"Database","NEWFIRMc%d.soln" % (chip)],"/")
   solndat = readsoln(solnfile)
   solndat["ic"] = invsoln(solndat,2048,2048,verb=1)
   N = len(dat["NUMBER"])
   i = Ei
   ip = 0
   while i >= 0 and ip <=3:
      u,v = dat["X_IMAGE"]-cnaxis1/2., dat["Y_IMAGE"]-cnaxis2/2.
      u,v = mbasis(solndat["c"],u,v)
      du = -Xs[i]*xscale*pi/180/3600
      dv = -Ys[i]*yscale*pi/180/3600
      if (i == Ei) or sqrt((Xs[i]-Xs[Ei])**2+(Ys[i]-Ys[Ei])**2) > 2.5:
#         print Ei,i,du,dv,Xs[i],Ys[i]
         xl,yl = mbasis(solndat["ic"],u+du,v+dv)
         xl = xl + 1024
         yl = yl + 1024
         if (i != Ei) and sqrt((Xs[i]-Xs[i+1])**2+(Ys[i]-Ys[i+1])**2) > 2.5: ip += 1
         for n in range(N):
            x = xl[n]
            y = yl[n]
            mb = dat["MAG_BEST"][n]
            if 0<x<2048 and 0<y<2048 and (i==Ei or mb<-8.5):
               a = dat["A_IMAGE"][n]
               b = dat["B_IMAGE"][n]
               t = dat["THETA_IMAGE"][n]*pi/180
               a = a / (1.0+0.5*ip)
               b = b / (1.0+0.5*ip)
      
               dth = (2*pi)/50
               th = arange(0,2*pi+dth,dth,Float)
               cth = cos(th)
               sth = sin(th)
               ct = cos(t)
               st = sin(t)
               xr = mag*a*cth
               yr = mag*b*sth
               xr,yr = xr*ct - yr*st, xr*st + yr*ct
               xr,yr = xr+x,yr+y
               m = len(xr)
               if verb: print x,y,a,b,t*180/pi
      
               xmin, xmax = extrema(xr).astype(Int32)
               ymin, ymax = extrema(yr).astype(Int32)
               xr,yr = xr-xmin,yr-ymin
               xr = (xr+0.5).astype(Int32)
               yr = (yr+0.5).astype(Int32)
      
               subimage = zeros((ymax-ymin,xmax-xmin),Float32)
               subim = NumToVTKImage(subimage)
      
               canvas = vtkImageCanvasSource2D()
               canvas.SetExtent(0,xmax-xmin,0,ymax-ymin,0,0)
               canvas.SetScalarType(VTK_FLOAT)
               canvas.DrawImage(0,0,subim)
               canvas.SetDrawColor(1.0)
      
               for j in range(m-1):
                  canvas.DrawSegment(xr[j],yr[j],xr[j+1],yr[j+1])
      
               xi, yi = int(x+0.5)-xmin, int(y+0.5)-ymin
               canvas.FillPixel(xi,yi)
               output = canvas.GetOutput()
               output.Update()
      
               nx = xmax-xmin
               ny = ymax-ymin
               oxmin,oxmax = clip([xmin,xmax],0,2048)
               oymin,oymax = clip([ymin,ymax],0,2048)
#######         print xmin,xmax, oxmin,oxmax, nx
#######         print ymin,ymax, oymin,oymax, ny
      
               obj = VTKImageToNum(output)
               if ymin < 0: obj = obj[:ymin]
               if xmin < 0: obj = obj[::,:xmin]
               if xmax > 2047: obj = obj[::,xmax-2047:]
               if ymax > 2047: obj = obj[ymax-2047:]
               if i == Ei:
                  submask = mask[oymin:oymax+1,oxmin:oxmax+1]
                  dmask = submask * logical_not(obj)
                  mask[oymin:oymax+1,oxmin:oxmax+1] = dmask.astype(mask.typecode())
               else:
                  submask = pers[oymin:oymax+1,oxmin:oxmax+1]
                  dmask = submask * logical_not(obj)
                  pers[oymin:oymax+1,oxmin:oxmax+1] = dmask.astype(pers.typecode())
      i -= 1

   return mask,pers

def NEWFIRMMode(N,chip,first=True,second=False,msh=10,cat=None,histo=0,mag=3,verb=1):
   E = Exposure(N)
   cen,sig = E.Mode(chip,first=first,second=second,msh=msh,cat=cat,histo=histo,mag=mag,verb=verb)
   return cen,sig

def NEWFIRMSkySubtract(N,chip,NDS=[],first=1,second=0,nf=3,verb=0):
   Exposures = [Exposure(N1) for N1 in NDS]
   u = [j for j in range(len(NDS)) if NDS[j].base == N.base][0]
   E = Exposures[u]
   E.SkySubtract(chip,Exposures,first=first,second=second,nf=nf,verb=verb)

def NEWFIRMShift(N,chip,NDS=[],first=1,second=0,nf=3,verb=0):
   Exposures = [Exposure(N1) for N1 in NDS]
   u = [j for j in range(len(NDS)) if NDS[j].base == N.base][0]
   E = Exposures[u]
   E.Shift(chip,Exposures,first=first,second=second,verb=verb)

#def nonexist(f,d):
#   return (f not in os.listdir(d)) or os.stat(d+"/"+f)[6]==0

def invsoln(solndat,naxis1,naxis2,verb=0):
    nterms = len(solndat["c"])
    x = asarray( range(0,naxis1+naxis1/nterms,naxis1/nterms) + [naxis1])
    y = asarray( range(0,naxis2+naxis2/nterms,naxis2/nterms) + [naxis2])
    x = VTKSubtract(x, solndat["center"][0])
    y = VTKSubtract(y, solndat["center"][1])
    nx = len(x); ny = len(y)
    o = ones((ny,nx),Float32)
    yy = (o*y[::,NewAxis]).flat
    xx = (o*x[NewAxis,::]).flat
    u,v = mbasis(solndat["c"],xx,yy)
    nterm = [4,6,12,20,30,42,56]
    nords = [0,1, 2, 4, 4, 4, 4]
    m = nords[nterm.index(nterms)]
    b = abasis(m,u,v)
    c = svdfit(b,concatenate([xx,yy]))
    xf,yf = mbasis(c,u,v)
    xres = VTKSubtract(xf,xx); yres = VTKSubtract(yf,yy)
    if verb:
      print "Using",nterm,"terms (order=",m,")"
      print "Inverse solution RMS=",rms(xres),rms(yres)
    return c


