
import time
from sets import Set
from MLab import *
from scipy.interpolate import *
from scipy.optimize import *
from basics import *
from pgplot import *
import gzip
import pickle
import cPickle
from scipy.special import *

DDIR = join([os.environ["PYTHONHOME"],"Database"],"/")
LFDIR = join([os.environ["PYTHONHOME"],"Database","Filters"],"/")
BDIR = join([os.environ["PYTHONHOME"],"Database","BPZ","SED"],"/")
FDIR = join([os.environ["PYTHONHOME"],"Database","ZPHOT","filters"],"/")
TDIR = join([os.environ["PYTHONHOME"],"Database","ZPHOT","templates"],"/")
COLORS = [2,8,3,4]
COLORS += 100*[1]
MAXZFORM = 3.0

localfilters = ("wirc", "filter","suprime","ACS","irac")

def Cardelli(l,Rv=3.1):
    x = 1.0/(l/1e4)
    def fn1(x):
       a =  0.574*power(x,1.61)
       b = -0.527*power(x,1.61)
       return asarray([a,b])
    def fn2(x):
       y = x-1.82
       a = 1 + 0.17699*y - 0.50447*y**2 - 0.02427*y**3 + 0.72085*y**4 \
           + 0.01979*y**5 - 0.77530*y**6 + 0.32999*y**7
       b = 1.41338*y + 2.28305*y**2 + 1.107233*y**3 - 5.38434*y**4 \
           - 0.62251*y**5 + 5.30260*y**6 - 2.09002*y**7
       return asarray([a,b])
    def fn3(x):
       s1 = (x<5.9)
       s2 = (x>=5.9)*(x<=8.0)
       def sf1(x):
          if type(x) == ArrayType:
             fa = fb = zeros(x.shape,Float)
          else:
             fa = fb = 0.0
          return asarray([fa,fb])
       def sf2(x):
          fa = -0.04473*(x-5.9)**2 - 0.009779*(x-5.9)**3
          fb = 0.2130*(x-5.9)**2 + 0.1207*(x-5.9)**3
          return fa,fb
          return asarray([fa,fb])
       fa,fb = s1*sf1(x) + s2*sf2(x)
       a = 1.752 - 0.316*x-0.104/((x-4.67)**2 + 0.341) + fa
       b = -3.090 + 1.825*x + 1.206/((x-4.62)**2 + 0.263) + fb
       return asarray([a,b])
    def fn4(x):
       a = -1.073 - 0.628*(x-8) + 0.137*(x-8)**2 - 0.070*(x-8)**3
       b = 13.670 + 4.257*(x-8) - 0.420*(x-8)**2 + 0.374*(x-8)**3
       return asarray([a,b])
    use1 = ((x<=1.1))[NewAxis,::]
    use2 = ((x>1.1)*(x<=3.3))[NewAxis,::]
    use3 = ((x>3.3)*(x<=8.0))[NewAxis,::]
    use4 = ((x>8))[NewAxis,::]
    a,b = use1*fn1(x) + use2*fn2(x) + use3*fn3(x) + use4*fn4(x)
    AlAV = a+b/Rv
    return  AlAV

def Calzetti(l,Rv=3.1):
    x = l/1e4
    ke1 = ((1.86-0.48/x)/x - 0.1)/x + 1.73
    ue1 = between(x,0.63,1.0,inc=1)
    ke2 = 2.656*(-2.156 + 1.509/x - 0.198/x**2 + 0.011/x**3) + 4.88
    ue2 = less(x,0.63)
    ke = ue1*ke1 + ue2*ke2
    return ke/Rv

class BPZTemplate:
    def __init__(self,ty):
       g = glob(BDIR+"/*.sed")
       fl = [sed for sed in g if ty in sed][0]
       self.name = fl.split("/")[-1]
       self.lam,self.flam = readtable(fl)
       self.fixedz = True
    def __repr__(self):
       return self.name

class Template:
    def __init__(self,ty):
       fl = "%s/CWW_%s_ext.sed" % (TDIR,ty)
       self.name = fl.split("/")[-1]
       self.lam,self.flam = readtable(fl)
       self.fixedz = True
    def __repr__(self):
       return self.name

class AGNTemplate:
    def __init__(self,ty):
       fl = "%s/sedAGN%d.dat" % (TDIR,ty)
       self.name = fl.split("/")[-1]
       self.lam,self.flam = readtable(fl)
       self.lam = exp10(-self.lam)*1e3
       self.flam = exp10(self.flam)
       self.fixedz = True
    def __repr__(self):
       return self.name


class SDSSAGNTemplate:
    def __init__(self,default=True,num=29,noem=[3727.0]):
       if default: fl = "%s/sdssqso.fits" % (TDIR)
       else: fl = "%s/spDR2-%03d.fit" % (TDIR,num)
       self.default = default
       self.noem = noem
       self.fits = FITS(fl)
       self.read()
       self.fits.close()
       self.name = fl.split("/")[-1]
       self.fixedz = True
    def __repr__(self):
       return self.name
    def read(self):
       if self.default:
          crval1 = self.fits["crval1"]
          cdelt1 = self.fits["cdelt1"]
          naxis1 = self.fits["naxis1"]
          crpix1 = self.fits["crpix1"]
          self.lam0 = (1+arange(naxis1)-crpix1)*cdelt1 + crval1
          self.flam0 = self.fits.row(0)
          self.lamu = arange(400.0,self.lam0[0]+cdelt1/2.,cdelt1,Float32)
          self.flamu = zeros(self.lamu.shape,Float32)
          self.lamr = arange(self.lam0[-1]+cdelt1,30e3,cdelt1,Float32)
          redlevel = median(compress(greater(self.lam0,7400),self.flam0))
          self.flamr = redlevel * ones(self.lamr.shape,Float32)
          self.lam = concatenate([self.lamu,self.lam0,self.lamr])
          self.flam = concatenate([self.flamu,self.flam0,self.flamr])
       else:
          crval1 = self.fits["crval1"]
          cdelt1 = self.fits["cd1_1"]
          naxis1 = self.fits["naxis1"]
          crpix1 = 1.0
          self.lam0 = (1+arange(naxis1)-crpix1)*cdelt1 + crval1
          self.flam0 = self.fits.row(0)
          self.lam = exp10(self.lam0)
          self.flam = self.flam0*less(self.lam,6780)
       for noem in self.noem:
          o = between(self.lam,noem-50,noem+50,inc=1)
          l,f = compress(1-o,[self.lam,self.flam])
          tck = splrep(l,f,s=0,k=1)
          self.flam = splev(self.lam,tck)

class BCTemplate:
    def __init__(self,ty="62",age=1e10,maraston=0):
       self.ty = ty
       self.maraston = maraston
       if self.maraston or self.ty == None or type(self.ty) == FloatType: self.generic_load()
       elif type(self.ty) == StringType: self.tyload()
       self.name = os.path.basename(self.fl)
       self.atten = 1.0
       self.age = age
       self.lages = getlog10(self.ages)
       self.params = None
       self.AlAv = None
       if self.age:
          if self.fixedz: self.getage(age)
          elif ty != None: self.lagelz(getlog10(age),ty)
       self.backup_age0 = Template("Scd")
       self.info = {}
    def generic_load(self):
       if self.maraston:
          self.fl = "%s/maraston2005.pck" % (DDIR)
       else:
          self.fl = "%s/bc2003_lr_ssp.pck" % (DDIR)
       if not os.path.exists(self.fl): print "Not there!";exit()
       self.bf = gzip.open(self.fl,"r")
       self.generic = cPickle.load(self.bf)
       self.bf.close()
       self.waves = self.generic["wavelengths"]
       self.ages = self.generic["ages"]
       self.metals = self.generic["metals"]
       self.tcks = None
       self.fixedz = False
    def tyload(self):
       self.fl = "%s/bc2003_lr_m%s_chab_ssp.pck" % (DDIR,self.ty)
       if not os.path.exists(self.fl): print "Not there!";exit()
       self.bf = gzip.open(self.fl,"r")
       self.waves,self.ages,self.tck = cPickle.load(self.bf)
       self.bf.close()
       self.fixedz = True
    def getage(self,age):
       if age: self.lage(log10(age))
       else: self.flam = None
    def lage(self,lage):
       self.logage = lage
       self.lam = self.waves
       if self.fixedz:
          self.flam0 = bisplev(self.lam,lage,self.tck).flat
          self.flam0 = self.flam0 * LSUN / (4*pi*pow(10*CMPPC,2))
          self.flam = 1.0*self.flam0
       else:
          self.lagelz(lage,self.ty)
       self.AlAv = None
    def lagelz(self,lage,lz):
       if (lage,lz) != self.params:
          self.params = [lage,lz]
          lage = asarray(lage)
          self.logage = lage
          lz = asarray(lz)
          self.lz = lz
          if not len(shape(lage)): lage = array([lage])
          if not len(shape(lz)): lz = array([lz])
          self.lam = self.waves
          tkeys = "tx", "ty", "c", "kx", "ky"
          aw = range(self.lam.shape[0])
          if not self.tcks:
              self.tcks = [[(k == "c" and self.generic[k][w] or self.generic[k]) for k in tkeys] for w in aw]
          self.flam0 = array([bisplev(lage,lz,tck) for tck in self.tcks]).flat
          if self.maraston:
             self.flam0 = exp10(self.flam0)
          self.flam0 = self.flam0 * LSUN / (4*pi*pow(10*CMPPC,2))
       self.flam = 1.0*self.flam0
       self.AlAv = None
    def extinct(self,Av,Law=Cardelli):
        if not self.AlAv:
           self.flam0 = self.flam
           self.AlAv = Law(self.lam)
        self.atten = exp10(-0.4*(Av*self.AlAv))
        self.flam = self.flam0*self.atten
    def tau(self,lage,lend,ltau,Z,ninc=16,emin=5):
        self.info["lage"] = lage
        self.info["lend"] = lend
        self.info["ltau"] = ltau
        self.info["Z"] = Z
        self.info["av"] = 0.0
        self.AlAv = None
        start = exp10(lage)
        finish = max([ exp10(lend),exp10(emin)])
        tau = exp10(ltau)
        if type(Z) is StringType: Zu = lambda t: eval(Z)
        elif type(Z) is FloatType: Zu = lambda t: Z
        else: Zu = Z
        if finish<start and tau:

           if 1:
#              if (start-finish) > 5*tau: finish = start - 5*tau
#              dt = min([tau/ninc,(start-finish)/ninc])
#              ninc2 = int((start-finish)/dt+1)
              dt = (log10(start)-log10(finish))/(ninc)
              times = exp10(arange(log10(finish),log10(start)+dt/100,dt,Float32))
              ltimes = log(times)
              dt = (times[2:]-times[:-2])/2.
              dt = concatenate([dt[:1],dt,dt[-1:]])
              dlt = (ltimes[2:]-ltimes[:-2])/2.
              dlt = concatenate([dlt[:1],dlt,dlt[-1:]])
              dt = times*dlt
              SF = lambda t: exp(-(start-t)/tau)
              SFs = array([SF(moment) for moment in times])
              SFtot = sum(SFs*dt)
              Masses = SFs*dt/SFtot
     
              flam = zeros(self.waves.shape,Float64)
              [(self.lagelz(log10(moment),Zu(moment)), add(flam,mass*self.flam,flam)) \
                                                      for mass,moment in zip(Masses,times)]
              self.flam = 1.0*flam

           if 0:
              if (start-finish) > 5*tau: finish = start - 5*tau
              dt = min([tau/ninc,(start-finish)/ninc])
              ninc2 = int((start-finish)/dt+1)
              dt = (start-finish)/ninc2
              times = arange(start,finish,-dt,Float32)
              flam = zeros(self.waves.shape,Float64)
              SFtot = 0.0
              SF = lambda t: exp(-(start-t)/tau)
              SFs = array([SF(moment) for moment in times])
              SFtot = sum(SFs)
              SFs = SFs/SFtot
              [[self.lagelz(log10(moment),Zu(moment)),add(flam,SF(moment)*self.flam,flam)] for moment in times]
              self.flam = flam/SFtot

        else:
           self.lagelz(log10(start),Zu(start))
    def __repr__(self):
       if self.age: return "%s age=%.5g Gyr" % (self.name,self.age/1e9)
       else: return "undetermined "+self.name

class Vega:
    def __init__(self):
       fl = "%s/A0V_KUR_BB.SED" % (FDIR)
       self.name = fl.split("/")[-1]
       self.lam,self.flam = readtable(fl,start=1)
       self.lam = self.lam
       tck = splrep(self.lam,self.flam)
       norm = 3.3992e-9 / splev(5556,tck)
       self.flam = self.flam * norm
    def __repr__(self):
       return self.name

VEGA = Vega()
#VEGA = BPZTemplate("Vega")

class Detector:
    def __init__(self,detector):
       self.name = detector
       if "imacs" in detector:
          fl = "%s/%s" % (LFDIR,detector)
          dwwav,dewav = readtable(fl)[:2]
          dwwav = dwwav*10
          self.dwwav,self.dewav = take([dwwav,dewav],argsort(dwwav),1)
       else:
          fd = GetFilter(detector,ab=1)
          self.dwwav,self.dewav = fd.wwav,fd.ewav

class Filter:
    def __init__(self,name,points=0,num=0,detector=None,ab=0,verb=0):
       self.name = name
       self.ab = ab
       self.verb = verb
       if detector: self.detector = Detector(detector)
       else: self.detector = None
       local = [key in name for key in localfilters]
       got = False
       if sometrue(local):
           try: 
             self.getfilt()
             self.num = -1
             self.props()
             got = True
           except:
             got = False
       if not got:
           self.points = points
           self.num = num
           self.getdat()
           self.getres()
           self.props()
    def __repr__(self):
       return "%3d %s" % (self.num,self.name)
    def getfilt(self):
       if "ACS" in self.name or "GROUND" in self.name or self.name in os.listdir(LFDIR):
          fl = "%s/%s" % (LFDIR,self.name)
          if not os.path.exists(fl):
             fl = "%s/THROUGHPUTS/ACS_WFC/%s" % (LFDIR,self.name)
          if not os.path.exists(fl):
             fl = "%s/GROUND/%s" % (LFDIR,self.name)
       else:
          fl = "%s/%s" % (FDIR,self.name)
       self.wwav,self.ewav = readtable(fl)[:2]
       if ("ACS" not in self.name) and ("irac" not in self.name):
           self.wwav = self.wwav*10
       if ("irac" in self.name):
           self.wwav = self.wwav*1e4
       self.jwav = arange(len(self.wwav))
    def props(self):
       f = 1.0/self.wwav
       f1,f2 = extrema(f)
       f,w,e = take([f,self.wwav,self.ewav],argsort(f),1)
       tck1 = splrep(f,e,k=3,nest=len(f),s=0)
       denom = splint(f1,f2,tck1)
       tck0 = splrep(f,e*w,k=3,nest=len(f),s=0)
       numer = splint(f1,f2,tck0)
       self.wleff = numer/denom
       tck0 = splrep(f,e*(w-self.wleff)**2,k=3,nest=len(f),s=0)
       numer = splint(f1,f2,tck0)
       self.bandpass1 = sqrt(numer/denom)
       hp = max(self.ewav)*0.5
       tck2 = splrep(self.wwav,self.ewav-hp,k=3,nest=len(f),s=0)
       self.bandpass1 = diff(extrema(sproot(tck2,mest=len(f))))[0]
       self.surface = 0.0
       self.conv_AB = 0.0
       if not self.ab:
          self.conv_AB = self.integrate(VEGA,0.0,0)
       if self.verb:
          print "%3d %s" % (self.num,self.name),
          print "%.3f" % (self.conv_AB)
    def getdat(self):
       fl = "%s/%s" % (FDIR,"filters.dat")
       f = open(fl,"r")
       fld = f.readlines()[3:]
       f.close()
       j,self.wleff,self.surface,self.bandpass1,self.conv_AB = [float(s) for s in fld[self.num-1].split()]
    def getres(self):
       fl = "%s/%s" % (FDIR,"FILTER.RES")
       f = open(fl,"r")
       fd = f.readlines()
       f.close()
       mall = nonzero([self.name in l for l in fd])
       m = mall[0]
       ud = fd[m+1:m+self.points+1]
       self.jwav,self.wwav,self.ewav = transpose([[float(w) for w in l.split()] for l in ud])
    def integrate(self,template,z=0.0,verb=0,ks=1,redfilt=0,logmult=1,straight=0):
       if verb: print "Integrating %s at z=%.3f with %s." % (template,z,self.name)
       zlam = template.lam
       zflam = template.flam
#       if template == VEGA and not straight:
#          ftck = splrep(zlam,zflam,k=ks,nest=len(template.lam),s=0)
#          zflam = zflam/splev(5556.0,ftck)
#          del ftck
       zlam = zlam*(1+z)
       zflam = zflam/(1+z)
       if redfilt: f = 1./(self.wwav*(1+z))
       else: f = 1./self.wwav
       zf = 1./zlam
       f1,f2 = extrema(f)
       zf,zlam,zflam = compress(between(zf,f1,f2,inc=1),[zf,zlam,zflam],1)[::,::-1]
       if len(zf) < 2*ks+2: return 99.0
       ftck = splrep(f[::-1],self.ewav[::-1],k=ks,nest=len(f),s=0)
       feff = splev(zf,ftck)
       del ftck
       if self.detector:
          df = 1./self.detector.dwwav[::-1]
          dd = self.detector.dewav[::-1]
          dtck = splrep(df,dd,k=ks,nest=len(f),s=0)
          feff = feff*splev(zf,dtck)
          del dtck
       zf = CLIGHT*1e8/zlam
       f = CLIGHT*1e8*f
       if logmult:
          zfnu = exp10(log10(zflam) + log10(zlam) - log10(zf))
       else:
          zfnu = zflam * zlam / zf
       f1,f2 = extrema(f)
       tck0 = splrep(zf,feff*zfnu/zf,k=ks,nest=len(zf),s=0)
       tck1 = splrep(zf,feff/zf,k=ks,nest=len(zf),s=0)
       numer = splint(f1,f2,tck0)
       denom = splint(f1,f2,tck1)
       del tck0,tck1
       integ = divz(numer,denom)
       if integ:
          mag = -2.5*log10(integ) - 48.60
       else:
          mag = -99.999
       if not straight:
          if not self.ab: mag = mag - self.conv_AB
       return mag

def GetFilter(name,detector=None,ab=0,verb=0):
    local = [key in name for key in localfilters]
    if sometrue(local):
        F = Filter(name,detector=detector,ab=ab,verb=verb)
    else:
        fl = "%s/%s" % (FDIR,"filters.log")
        f = open(fl,"r")
        fld = f.readlines()[2:-1]
        f.close()
        nums = [int(l.split()[0]) for l in fld]
        names = [l[7:52].strip().replace("\t","") for l in fld]
        points = [int(l.split()[-1]) for l in fld]
        f.close()
        for n in names:
           if name in n:
              p = points[names.index(n)]
              i = nums[names.index(n)]
              break
        F = Filter(n,p,i,detector=detector,ab=ab,verb=verb)
    return F

def Redshift1(z,t,filters):
    model = array([f.integrate(t,z) for f in filters])
    return model

def RedshiftAll(z,templates,filters):
    models = asarray([Redshift1(z,t,filters) for t in templates])
    return models

def AddModels(weights,models):
    fluxes = exp10(-0.4*(models-25.0))
    totals = add.reduce(weights[::,NewAxis]*fluxes,0)
    mags = -2.5*log10(totals) + 25.0
    return mags

def maxlage(z):
    t1 = tlook(z,0.7,[0.3,0.7])
    tf = tlook(MAXZFORM,0.7,[0.3,0.7])
    la1 = getlog10(tf-t1)+9.0
    return la1

def ta2la(ta,z):
    la = ta
    return la

def GetModelAgeMetal(tage,lz,lav,z,obs,err,w,template,filters,chi2=0,fit=1,verb=0,offsets=None):
    ml1 = maxlage(z)
    lage = clip(tage,min(template.lages),ml1)
    lage = ta2la(tage,z)
    template.lagelz(lage,lz)
    template.extinct(lav,Law=Cardelli)
    result = GetModel(z,obs,err,w,template,filters,chi2=chi2,fit=fit,offsets=offsets)
    if fit < 2:
       bad = (tage > ml1)
       if bad: result = power(100,tage-ml1)*result
       bad = (abs(lz) > 0.5)
       if bad: result = power(100,abs(lz)-0.5)*result
#       result = result / exp(-0.5*(lz/1.0)**2)
       if verb: print lage,lz,lav,z,sum(result**2)
    return result

def GetModelAge(tage,z,obs,err,w,template,filters,chi2=0,fit=1,verb=0,offsets=None):
    ml1 = maxlage(z)
    lage = clip(tage,min(template.lages),ml1)
    lage = ta2la(tage,z)
    template.lage(lage)
    lav = 1.0
    template.extinct(lav,Law=Cardelli)
    result = GetModel(z,obs,err,w,template,filters,chi2=chi2,fit=fit,offsets=offsets)
    if tage > ml1: result = power(100,tage-ml1)*result
    if verb: print lage,z,sum(result**2)
    return result

def GetOffsets(z,obs,offsets,lage=0):
    offmags = zeros(obs.shape,Float32)
    for offsets1 in offsets:
       colors = array([-diff(take(obs,offsets1["col"][o]))[0] for o in range(len(obs))])
       if "to" in offsets1:
         sol,co,zo,to = offsets1["sol"], offsets1["co"], offsets1["zo"], offsets1["to"]
       else:
         sol,co,zo = offsets1["sol"], offsets1["co"], offsets1["zo"]
         to = len(co)*[1]
       for jc,color in enumerate(colors):
          if zo[jc] > 0: zbases = polynom((1+array([min([z,1.5])])),zo[jc]-1)
          else: zbases = array([1.0])
          if co[jc] > 0: cbases = polynom(array([color]),co[jc]-1)
          else: cbases = array([1.0])
          if to[jc] > 0: tbases = polynom(array([lage]),to[jc]-1)
          else: tbases = array([1.0])
          ubases = transpose([zb*cb*tb for zb in zbases for cb in cbases for tb in tbases])
          if len(sol[jc]) > 0: offmags[jc] += dot(ubases,sol[jc])[0]
          else: offmags[jc] += ubases[0]*sol[jc][0]
    return offmags

def Rescale(obs,err,w,mags,wl,fbs,ab,zpts=None):
    u = less(obs,90)
    if 1 or not zpts: zpts = -ab
    if ab.shape == mags.shape: zpts2, ab2, wl2, fbs2 = zpts, ab, wl, fbs
    else: zpts2, ab2, wl2, fbs2 = zpts[NewAxis,::], ab[NewAxis,::], wl[NewAxis,::], fbs[NewAxis,::]
    lmfluxes = -0.4*(mags-zpts2-25.0)
    lofluxes = -0.4*(obs-zpts-25.0)
    mfluxes = exp10(lmfluxes)/wl2
    ofluxes = exp10(lofluxes)*u/wl
    ofluxes = where(u,ofluxes,0.0)
    ferr = where(u,(err/1.0857) * ofluxes, exp10(-0.4*(err-zpts-25.0))/wl)
    if mags.shape == obs.shape: mfluxes = asarray([mfluxes])
    uw = (u*w/ferr)**2
    scale = divz(add.reduce((uw*ofluxes)[NewAxis,::]*mfluxes,1), add.reduce(uw[NewAxis,::]*power(mfluxes,2),1))
    if mags.shape == obs.shape: scale = scale[0]
    scale = where(between(scale,-1e40,1e40),scale,1e40)
    scale = -2.5*log10(scale)
    return scale

def GetModel(z,obs,err,w,template,filters,chi2=0,fit=1,offsets=None):
    mags = Redshift1(z,template,filters)
    if offsets: mags = mags + GetOffsets(z,obs,offsets,lage=template.logage)
    wl = array([f.wleff for f in filters])
    ab = array([f.conv_AB for f in filters])
    fbs = array([f.bandpass1 for f in filters])
    zpt = Rescale(obs,err,w,mags,wl,fbs,ab)
    mags = mags + zpt
    fscale = -zpt
    u = less(obs,90)
    mfluxes = exp10(-0.4*(mags-25.0+ab))
    ofluxes = exp10(-0.4*(obs-25.0+ab))*u
    ofluxes = where(u,ofluxes,0.0)
    ferr = where(u,(err/1.0857) * ofluxes, exp10(-0.4*(err-25.0+ab)))
    residuals = mfluxes - ofluxes
    if chi2:
       chi2 = sum(compress(w,power(residuals/ferr,2)))
       return chi2
    elif fit:
       if fit == 2:
          return mags
       else:
          return compress(w,residuals/ferr)
    else:
       return (mfluxes,ofluxes,ferr,fscale)

def uspv(l, tck):
   d = splev(l,tck)
   return asarray([d],Float64)

def splvalley(tck, l):
   sol = leastsq(uspv, l, args=(tck,), factor=0.1, full_output=1,
        epsfcn=1e-4, ftol=1e-4, xtol=1e-4, gtol=0.0, maxfev=1000)
   return sol

def edgy(vtck,l):
   z = sproot(vtck)
   if sum(less(z,l)): z1 = sort(compress(less(z,l),z))[-1]
   else: z1 = 0.0
   if sum(greater(z,l)): z2 = sort(compress(less(l,z),z))[0]
   else: z2 = 99.0
   if z1 and z2: return z1,z2
   else: return l,l

def centery(s,m):
   f = edgy(s,m)
   c = mean(f)
   return c, m-f[0], f[1]-m

def GetRedshift(obs,err,w,filters,templates,plot=0,z1=0,z2=2.5,dz=0.05,obj=None,verb=0,offsets=None):
    if plot:
       time0 = time.time()
    za = exp(arange(log(z1+1),log(z2+1)+dz/2.,dz,Float32)) - 1
    if "undetermined" in repr(templates):
      nt = 0
      chi2 = zeros(za.shape,Float64)
      lages = zeros(za.shape,Float64)
      lav = zeros(za.shape,Float64)
      lmetals = zeros(za.shape,Float64)
      mages = zeros(za.shape,Float64)
      if templates.fixedz or templates.ty != None:
          thing = lambda la,z: GetModelAge(la,z,obs,err,w,templates,filters,fit=1,offsets=offsets)
          zthing = lambda z,guess: leastsq(thing,guess,args=(z,),factor=0.1,epsfcn=1e-4,ftol=1e-4,xtol=1e-4,maxfev=5000)
          guess = [maxlage(0.0)-0.3]
      else:
          thing = lambda laz,z: GetModelAgeMetal(laz[0],laz[1],laz[2],z,obs,err,w,templates,filters,fit=1,offsets=offsets)
          zthing = lambda z,guess: leastsq(thing,guess,args=(z,),diag=[1.0,10.0,1.0],factor=0.1,epsfcn=1e-4,gtol=1e-4,ftol=1e-4,xtol=1e-8,maxfev=5000)
          guess = [maxlage(0)-0.3,0.2,10.0]
      for jz in arange(len(za)):
          zsol = zthing(za[jz],guess)
          if templates.fixedz or templates.ty != None:
             lages[jz] = ta2la(zsol[0],za[jz])
             guess = [maxlage(za[jz])-0.3]
#             guess = [lages[jz]-0.1]
          else:
             lages[jz] = ta2la(zsol[0][0],za[jz])
             lmetals[jz] = zsol[0][1]
             lav[jz] = zsol[0][2]
             guess = [maxlage(za[jz])-0.3,0.2,-2.0]
#             guess = [lages[jz]-0.1,lmetals[jz]]
          mages[jz] = maxlage(za[jz])
          if templates.fixedz: templates.lage(lages[jz])
          else:
             templates.lagelz(lages[jz],lmetals[jz])
             ulav = (arctan(lav[jz])+pi/2.0)/pi
             templates.extinct(ulav,Law=Cardelli)
          chi2[jz] = GetModel(za[jz],obs,err,w,templates,filters,chi2=1,offsets=offsets)
          if verb: sys.stdout.write(".");sys.stdout.flush()
      if verb: sys.stdout.write("\n")
    else:
       thing = lambda z,t: GetModel(z,obs,err,w,t,filters,chi2=1,offsets=offsets)
       nt = len(templates)
       chi2s = transpose([[thing(z,t) for t in templates] for z in za])
       chi2 = minimum.reduce(chi2s,0)
       tchi2 = argmin(chi2s,0)
    if plot:
       opendev("/xs",nx=1,ny=2,land=0,asp=1)
       if nt: cmar = add(log10(extrema(ravel(chi2s))),[-0.3,0.3])
       else: cmar = add(log10(extrema(ravel(chi2))),[-0.3,0.3])
       newbox(extrema(za),cmar,ch=2.2,ylog=1,xlog=0,rax=nt,rb=0.85,ylab=1,ytop=0.85)
       if obj: title(obj,ch=2.2)
       xlabel("z",ch=2.2)
       ylabel("\gx\u2\d",ch=2.2,dist=2.5)
       [drawln(za,log10(chi2s[jt]),ls=1,ci=COLORS[jt]) for jt in range(nt)]
       drawln(za,log10(chi2),ls=4,ci=1)
       drawpt(za,log10(chi2),ps=4,ci=7)
    mz = za[argmin(chi2)]
    zp = 0.0
    ze1 = 0.0
    ze2 = 0.0
    mc = -9.0
    mt = -1
    mm = 0.0
    ns = 10.0
    if min(za) <= mz <= max(za):
       tck = splrep(za,chi2,s=0,k=3,nest=len(za))
       zp = splvalley(tck,mz)[0]
       if min(za) < zp < za[-2]:
           mc = max([splev(zp,tck),min(chi2)])
           tck = splrep(za,chi2-(mc+ns**2),s=0,k=3,nest=len(za))
           zm,ze1,ze2 = centery(tck,zp)
           ze1 = ze1/ns
           ze2 = ze2/ns
           mc = mc/(sum(w)-1)
           if nt:
              mt = tchi2[argmin(chi2)]
           else:
              mtz = zthing(zp,guess)
              mt = ta2la(mtz[0],zp)
              if templates.fixedz: mm = 0.0
              else: mm = mtz[1]
           if mc > 1:
              ze1 *= sqrt(mc)
              ze2 *= sqrt(mc)
    if plot:
       zu = [mz,zp][zp>0]
       time1 = time.time()
       drawln([zp-ze1,zp-ze1],cmar,ls=2,ci=1)
       drawln([zp,zp],cmar,ls=1,ci=1)
       drawln([zp+ze2,zp+ze2],cmar,ls=2,ci=1)
       if not nt:
          newbox(extrema(za),[7.0,10.3],ch=2.2,ylog=0,xlog=0,lax=0,era=0,ylab=1,rb=0.85,ytop=0.85)
          drawln(za,lages,ls=1,ci=4)
          drawln(za,mages,ls=1,ci=2)
          if sum(lmetals):
            newbox(extrema(za),[-1,1],ch=2.2,ylog=0,xlog=0,lax=0,era=0,ylab=0,rb=0.85,ytop=0.85)
            rlabel("log(age) (yr) and [Z/H]",ch=2.2,asp=1,dist=2.5)
            print lmetals
            drawln(za,lmetals,ls=1,ci=3)
          else:
            rlabel("log(age) (yr)",ch=2.2,asp=1,dist=2.5)
       fws = array([f.wleff for f in filters])
       fbs = array([f.bandpass1 for f in filters])
       wran = diff(extrema(fws))[0]
       wmar = min(fws)-0.1*wran, max(fws)+0.1*wran
       if nt:
          mfluxes,ofluxes,ferr,fscale = GetModel(zu,obs,err,w,templates[mt],filters,chi2=0,fit=0,offsets=offsets)
          subtitle("--- minimum with %s" % (templates[mt]),ch=1.5)
       else:
          zsol = zthing(zu,guess)
          if templates.fixedz or templates.ty != None:
             mt = ta2la(zsol[0],zu)
             mm = 0.0
             templates.lage(mt)
             subtitle("log(Age) at minimum = %.3f" % (mt),ch=2.2)
          else:
             mt = ta2la(zsol[0][0],zu)
             mm = zsol[0][1]
             ma = zsol[0][2]
             templates.lagelz(mt,mm)
             uma = (arctan(ma)+pi/2.0)/pi
             templates.extinct(uma,Law=Cardelli)
             subtitle("log(Age),[Z/H] at minimum = %.3f,%.3f" % (mt,mm),ch=2.2)
          mfluxes,ofluxes,ferr,fscale = GetModel(zu,obs,err,w,templates,filters,chi2=0,fit=0,offsets=offsets)
       fran = diff(extrema(ofluxes))[0] + 3*max(ferr)
       fmar = 0.0-0.1*fran, max(mfluxes+3*ferr)+0.25*fran
       newbox(wmar,fmar,ch=2.2,xlog=0,ylog=0,rb=0.85,ylab=1,ytop=0.90)
       xlabel("\gl (\A)",ch=2.2)
       ylabel("10\u-0.4(m-25+c)\d/\gl",ch=2.2,dist=2.5)
       subtitle("\gx\u2\d minimum at z=%.3f" % (zu),ch=2.2)
       if nt:
         [drawpt(fws,GetModel(zu,obs,err,w,templates[jt],filters,chi2=0,fit=0,offsets=offsets)[0],ps=4,ci=COLORS[jt]) for jt in range(len(templates))]
         r = pow(10,0.4*fscale)
         tw,tf = templates[mt].lam, templates[mt].flam/r
         tf = tf * tw
       else:
         mfluxes, ofluxes, ferr, fscale = GetModel(zu,obs,err,w,templates,filters,chi2=0,fit=0,offsets=offsets)
         r = pow(10,0.4*fscale)
         tw,tf = templates.lam, templates.flam/r
         tf = tf * tw
       tck = splrep(tw,tf,k=1,nest=len(tw),s=0)
       tw = tw*(1+zu)
       tu = between( tw,min(fws-fbs),max(fws+fbs))
       tw,tf = compress(tu,[tw,tf],1)
       tck = splrep(tw,tf,k=1,s=0)
       u = less(obs,90)*w
       ufws,umfluxes,uofluxes,ufbs,uferr = compress(w,[ fws,mfluxes,ofluxes,fbs,ferr],1)
       umtf = splev(ufws,tck)
       nr = sum(ufws*umtf/umfluxes/uferr**2)/sum(ufws/uferr**2)
       drawpt(ufws,umfluxes,ps=17,ci=4,ch=6)
       drawpt(ufws,uofluxes,ps=17,ci=2,ch=4)
       drawpterr(ufws,ufbs/2.0,uofluxes,uferr,ch=2)
       drawln(tw,tf/nr,ci=7)
       closedev()
       print "time=",time1-time0
    return (zp,ze1,ze2,mc,mt,mm)

def elfnct(k,x,fx,gx,ndimk,ns,jflag,fmax,fnoise,funcs=()):
    obs, err, w, templates, filters, offsets, verb, models = funcs
    lage,lz,lav,z = x
    xl = "%r" % (x.tolist())
    if xl not in models:
       if lav:
          templates.lagelz(lage,lz)
          templates.extinct(lav,Law=Cardelli)
          mf,of,fe,fsc = GetModel(z,obs,err,w,templates,filters,chi2=0,fit=0,offsets=offsets)
       else:
          mf = zeros(obs.shape,Float32)
          of = exp10(-0.4*(obs-25))
          u = less(obs,90)
          ab = array([f.conv_AB for f in filters])
          fe = where(u, err/1.0857 * of, exp10(-0.4*(err-25.0+ab)))
       models[xl] = (mf,of,fe)
    (mf,of,fe) = models[xl]
    chi2 = sum(w*power((mf-of)/fe,2))
    ndeg = sum(w) - len(x)
    if verb and k==1: print xl,chi2,chi2/ndeg,len(models.keys())
    fx[0] = chi2
    fnoise[0] = 1e-16*fx[0]

def xranger(k,mode,w1,w2,ndimk,nsubk,ns):
    nsubk[0] = ndimk
    for j in range(ndimk): w2[j] = w1[j]

def BMRedshift(obs,err,w,filters,templates,z0=0.5,age0=7.0,m0=0.2,att0=0.9,verb=0,offsets=None,fixedz=0,difgrd=1e-1):
    lb = [ 7.5,-0.5, 0.1, 0.0]
    ub = [10.0, 0.5, 1.0, 1.5]
    xlower = lambda i: lb[i-1]
    xupper = lambda i: ub[i-1]
    models = {}
    import BoundedMinimization as bm
    BM = bm.BoundedMinimization()
    BM.funcs = (obs,err,w,templates,filters,offsets,verb,models,)
    BM.x = array([age0,m0,att0,z0],Float64)
    BM.elfnct = elfnct
    BM.range = xranger
    BM.xlower = xlower
    BM.xupper = xupper
    ns = 1
    n = 4
    BM.ns = ns
    BM.n = n
    BM.invar = array(ns*range(1,n+1,1),Int32)
    BM.istate = array(n*[1]+ns*[-1],Int32)
    if fixedz: BM.istate[3] = 0
    BM.flowbd = 0.0
    BM.nvar = n*arange(ns)+1
    BM.ipfreq = [-1,1][verb]
    BM.ipwhat = 5
    BM.epsil = 1e-2
    BM.stepl[0] = 0.1
    BM.stmax = 0.5
    BM.maxls = 15
    BM.relpr = 1e-8
    BM.difgrd = difgrd
    BM.setlwk()
    x0 = multiply(1.0,BM.x)
    t0 = time.time()
    if verb: print BM.x
    BM.run()
    t1 = time.time()
    print "time=",t1-t0
    if 1 or verb:
       print len(models.keys())
       print BM.iflag
       print BM.info
       print BM.stepl
       print x0
       print BM.x
    iflag = BM.iflag
    chi2 = sum(BM.fx)
    if fixedz: ndeg = (sum(w)-3)
    else: ndeg = (sum(w)-4)
    rchi2 = chi2/ndeg
    hess = BM.wk[ns+BM.nvar[ns]-1:ns+BM.nvar[ns]+BM.lwk[1]-1]
    if 1 or verb:
       print "reduced chi2",rchi2
       print hess
    fco = zeros((n,n),Float64)
    a,b = 0,1
    for j in range(n):
      fco[j,:j+1] = hess[a:b]
      fco[:j+1,j] = hess[a:b]
      a,b = b, b+j+2
    zmin = BM.x[3]
    if verb: print fco
    fco = fco/2.0
    errors = sqrt(diag(divz(1,fco)))*sqrt(rchi2)
    if verb: print errors
    zerr = errors[3]
    zage,zmetal,zav = BM.x[:3]
    print zmin,zerr,zerr,rchi2,zage,zmetal,zav,iflag
    zmin,zerr,zerr,rchi2,zage,zmetal,zav,iflag,fco = \
        LMRedshift(obs,err,w,filters,templates,z0=zmin,age0=zage,m0=zmetal,att0=zav,verb=verb,offsets=None,fixedz=fixedz,factor=0.01)
    print zmin,zerr,zerr,rchi2,zage,zmetal,zav,iflag
    return zmin,zerr,zerr,rchi2,zage,zmetal,zav,BM.iflag,fco

def BMTrain(z,obs,err,w,filters,templates,age0=9.0,m0=0.2,att0=0.9,verb=0,offsets=None):
    results = BMRedshift(obs,err,w,filters,templates,z0=z,age0=age0,m0=m0,att0=att0,verb=verb,offsets=offsets,fixedz=1)
    zmin,zerr,zerr,rchi2,zage,zmetal,zav,iflag,fco = results
    errors = sqrt(diag(divz(1,fco)))
    mags = GetModelAgeMetal(zage,zmetal,zav,z,obs,err,w,templates,filters,fit=2,offsets=offsets)
    return mags,obs-mags,zage,zmetal,zav,rchi2,errors

def LMTrain(z,obs,err,w,filters,templates,age0=9.0,m0=0.2,att0=0.9,verb=0,offsets=None):
    results = LMRedshift(obs,err,w,filters,templates,z0=z,age0=age0,m0=m0,att0=att0,verb=verb,offsets=offsets,fixedz=1)
    zmin,zerr,zerr,rchi2,zage,zmetal,zav,iflag,fco = results
    errors = sqrt(diag(divz(1,fco)))
    templates.lagelz(zage,zmetal)
    templates.extinct(zav,Law=Cardelli)
    mags = GetModel(z,obs,err,w,templates,filters,fit=2,offsets=offsets)
    return mags,obs-mags,zage,zmetal,zav,rchi2,errors

def lmfnct(x,extras):
    lz0, lav0, obs, err, w, templates, filters, offsets, verb, models = extras
    if lz0: z = pow(10.0,lz0)
    else: z = pow(10.0,x[-1])
    if lav0: lav = pow(10.0,lav0)
    else: lav = pow(10.0,x[-2])
    lage,lz = x[:2]
    templates.lagelz(lage,lz)
    templates.extinct(lav,Law=Cardelli)
    fx = GetModel(z,obs,err,w,templates,filters,chi2=0,fit=1,offsets=offsets)
    if verb: print z,lage,lz,lav,sum(fx**2)
    return fx

def LMRedshift(obs,err,w,filters,templates,z0=0.5,age0=7.0,m0=0.2,av0=0.9,verb=0,offsets=None,fixedz=0,fixedav=-1,factor=0.1,epsfcn=1e-5):
    models = {}
    lav = getlog10(av0)
    guess = [age0,m0]
    if fixedav == -1: guess += [lav]
    if fixedz == 0: guess += [log10(z0)]
    extras = ([0,log10(z0)][(fixedz>0)],[0,lav][(fixedav>-1)],obs,err,w,templates,filters,offsets,verb,models)
    guess = array(guess)
    solution = leastsq(lmfnct,guess,args=(extras,),full_output=1,factor=factor,epsfcn=epsfcn,gtol=1e-5,ftol=1e-5,xtol=1e-8,maxfev=5000)
    if verb: print solution
    sol = solution[0]
    fjac = transpose(take(solution[1]["fjac"],argsort(solution[1]["ipvt"])))
    fco = cov(fjac)
    if verb: print fco
    fvec = solution[1]["fvec"]
    chi2 = sum(fvec**2)
    rchi2 = chi2/ (sum(w)-len(sol)-1)
    errors = sqrt(diag(divz(1,fco)))
    if verb: print errors*sqrt(rchi2)
    zage = sol[0]
    zmetal = sol[1]
    if fixedav == -1: zav = 10.0**sol[2]
    else: zav = av0
    if fixedz:
      zmin = z0
      zerr = 0.0
    else:
      zmin = pow(10,sol[-1])
      zerr = 2.303*errors[-1]*zmin
    iflag = 1
    return zmin,zerr,zerr,rchi2,zage,zmetal,zav,iflag,fco

def MinRedshift(obs,err,w,filters,templates,z0=0.5,age0=7.0,verb=0,offsets=None):
    zmin = -99
    zage = -99
    zmetal = -99
    zav = -99
    zerr = -1
    chi2 = -1
    rchi2 = -1
    if "undetermined" in repr(templates):
      nt = 0
      t0 = time.time()
      print z0,age0
      guess = array([log10(z0),age0])
      if (templates.fixedz or templates.ty != None):
         thing = lambda params: GetModelAge(params[1],10**params[0],obs,err,w,templates,filters,fit=1,verb=verb,offsets=offsets)
         solution = leastsq(thing,guess,factor=0.01,epsfcn=1e-5,gtol=1e-5,ftol=1e-5,xtol=1e-5,maxfev=5000,full_output=1)
         if verb: print "sol1=",solution[0];print
      else:
         thing = lambda params: GetModelAgeMetal(params[1],params[2],params[3],10**params[0],obs,err,w,templates,filters,fit=1,verb=verb,offsets=offsets)
         guess = concatenate([guess,[0.1,1.0]])
         solution = leastsq(thing,guess,factor=0.01,epsfcn=1e-5,gtol=1e-5,ftol=1e-5,xtol=1e-5,maxfev=5000,full_output=1)
         if verb: print "sol2=",solution[0];print
      sol = solution[0]
      fjac = transpose(take(solution[1]["fjac"],argsort(solution[1]["ipvt"])))
      fco = cov(fjac)
      fvec = solution[1]["fvec"]
      chi2 = sum(fvec**2)
      rchi2 = chi2/ (sum(w)-1)
      errors = sqrt(diag(divz(1,fco)))
      zmin = pow(10,sol[0])
      zerr = 2.303*errors[0]*zmin*sqrt(rchi2)
      zage = sol[1]
      if templates.fixedz or templates.ty != None: zmetal,zav = 0.0,-99
      else: zmetal,zav = sol[2],sol[3]
      t1 = time.time()
      print "time=", t1-t0
    return zmin,zerr,zerr,rchi2,zage,zmetal,zav,fco

def Train(z,obs,err,w,filters,templates,verb=0,offsets=None):
    guess = [9.0]
    thing = lambda params: GetModelAgeMetal(params[0],0.0,10,z,obs,err,w,templates,filters,fit=1,verb=verb,offsets=offsets)
    solution = leastsq(thing,guess,factor=0.01,epsfcn=1e-5,gtol=1e-5,ftol=1e-5,xtol=1e-8,maxfev=5000,full_output=1)
#    if verb: print solution[0];print
    thing = lambda params: GetModelAgeMetal(params[0],params[1],params[2],z,obs,err,w,templates,filters,fit=1,verb=verb,offsets=offsets)
    guess = [solution[0],0.1,1.0]
    solution = leastsq(thing,guess,factor=0.01,epsfcn=1e-5,gtol=1e-5,ftol=1e-5,xtol=1e-8,maxfev=5000,full_output=1)
    sol = solution[0]
    fjac = transpose(take(solution[1]["fjac"],argsort(solution[1]["ipvt"])))
    fco = cov(fjac)
    errors = sqrt(diag(divz(1,fco)))
    fvec = solution[1]["fvec"]
    chi2 = sum(fvec**2)
    rchi2 = chi2/ (1.0*sum(greater(w,0))-len(sol))
    zage,zmetal,zav = sol
    mags = GetModelAgeMetal(zage,zmetal,zav,z,obs,err,w*0+1,templates,filters,fit=2,offsets=offsets)
    return mags,obs-mags,zage,zmetal,zav,rchi2,errors

class PhotoTransformClass:
    def __init__(self,z,fz,f1,f2,d1=None,d2=None,dz=None,template=None,verb=0,abz=0,ab1=0,ab2=0):
        Fz = GetFilter(fz,dz,ab=abz)
        F1 = GetFilter(f1,d1,ab=ab1)
        F2 = GetFilter(f2,d2,ab=ab2)
        Filters = [Fz,F1,F2]
        if template: T = [template]
        else: T = [Template(ty=t) for t in ["E","Sbc","Scd","Im"]]
        mags = asarray([[f.integrate(t,z,verb=0,redfilt=(f is Fz)) for t in T] for f in Filters])
        con = array([f.conv_AB for f in Filters])
        top = (mags[0] + con[0]) - (mags[2] + con[2])
        bot = (mags[1] + con[1]) - (mags[2] + con[2])
        self.alpha = divz(top,bot)
        self.beta = self.alpha*(con[1] - con[2]) - con[0] + con[2] + 2.5*log10(1+z)
        self.color = mags[1]-mags[2]
        if verb:
           print self.color
           print self.alpha
           print self.beta

def PhotoTransform(z,fz,f1,f2,d1=None,d2=None,template=None,verb=1,abz=0,ab1=0,ab2=0):
    Transform = PhotoTransformClass(z,fz,f1,f2,d1=d1,d2=d2,template=template,verb=verb,abz=abz,ab1=ab1,ab2=ab2)
    if template: return Transform.alpha[0],Transform.beta[0]
    else: return Transform.alpha,Transform.beta

def MakeGridBPZ(filters,z0):
   avs = [0,0.5,1]
   ts = [BPZTemplate(t) for t in ["SB2","SB3","Sbc","Scd","Im"]]
#   ts += [AGNTemplate(1),AGNTemplate(2)]
   tn = [str(t) for t in ts]
   tr = [BCTemplate(ty=ty,age=age) for ty in ["52","62","72"] \
        for age in [5e9,3e9,2e9,1.5e9,1.2e9,1e9,0.5e9] for av in avs]
   for j,t in enumerate(tr):
      av = avs[j%3]
      t.extinct(av,Law=Cardelli)
      tn += ["%s,av=%.1f" % (t,av)]
   ts += tr
   def Go(z,template,tn1):
      mags = Redshift1(z,template,filters)
      print z,template,tn1
      return mags
   z1 = array([z for z in z0 for t in ts])
   ty1 = [t for z in z0 for t in tn]
   mags1 = asarray([Go(z,ts[jt],tn[jt]) for z in z0 for jt in range(len(ts))])
   return {"z":z1,"mags":mags1,"ty":ty1}

def MakeGrid(filters,template,z0,age0,m0,av0):
   def Go(zs,age,m,av):
      print age[0],m[0],av[0]
      template.lagelz(age[0],m[0])
      template.extinct(av[0],Law=Cardelli)
      mags = asarray([Redshift1(z,template,filters) for z in zs])
      return mags
   z1 = array([z0 for age in age0 for m in m0 for av in av0])
   age1 = array([age*ones(z0.shape) for age in age0 for m in m0 for av in av0])
   m1 = array([m*ones(z0.shape) for age in age0 for m in m0 for av in av0])
   av1 = array([av*ones(z0.shape) for age in age0 for m in m0 for av in av0])
   thing = lambda i: Go(z0,age1[i],m1[i],av1[i])
   N = len(age0)*len(m0)*len(av0)
   mags1 = asarray([thing(i) for i in range(N)])
   D = {"z":z1,"age":age1,"m":m1,"av":av1,"mags":mags1}
   for key in D:
      print D[key].shape,
      if key == "mag": D[key] = concatenate([D[key][::,::,i] for i in range(len(z0))],0)
      else: D[key] = concatenate([D[key][::,i] for i in range(len(z0))],0)
      print D[key].shape
   return D

def MakeGridTau(filters,template,z0,age0,m0,tau0,av0):
   def Go(zs,age,m,tau,av,finish=5):
      if type(finish) not in [ArrayType,ListType,TupleType]: finish = [finish]
      if "lage" not in template.info or ( \
              (template.info["lage"] != age[0]) or \
              (template.info["lend"] != finish[0]) or \
              (template.info["ltau"] != tau[0]) or \
              (template.info["Z"] != m[0])):
                      print age[0],m[0],tau[0]
                      template.tau(age[0],finish[0],tau[0],m[0],ninc=64,emin=7.5)
      if av[0] or template.AlAv:
         print age[0],m[0],tau[0],av[0]
         template.extinct(av[0],Law=Cardelli)
      mags = asarray([Redshift1(z,template,filters) for z in zs])
      return mags
   o0 = ones(z0.shape)
   z1 = array([z0 for age in age0 for m in m0 for tau in tau0 for av in av0])
   age1 = array([age*o0 for age in age0 for m in m0 for tau in tau0 for av in av0])
   m1 = array([m*o0 for age in age0 for m in m0 for tau in tau0 for av in av0])
   tau1 = array([tau*o0 for age in age0 for m in m0 for tau in tau0 for av in av0])
   av1 = array([av*o0 for age in age0 for m in m0 for tau in tau0 for av in av0])
   thing = lambda i: Go(z0,age1[i],m1[i],tau1[i],avi1[i])
   N = len(age0)*len(m0)*len(tau0)*len(av0)
   mags1 = asarray(map(Go,z1,age1,m1,tau1,av1))
#   mags1 = asarray([thing(i) for i in range(N)])
   D = {"z":z1,"age":age1,"m":m1,"av":av1,"tau":tau1,"mags":mags1}
   for key in D:
      print key,D[key].shape,
      if key == "mag":
##         D[key] = concatenate([D[key][::,::,i] for i in range(len(z0))],0)
#         nshape = (len(D[key]),len(D[key])/len(filters))
#         D[key] = reshape(D[key],nshape)
           D[key] = asarray([concatenate(D[key][::,::,f],0) for f in range(len(filters))])
      else:
##         D[key] = concatenate([D[key][::,i] for i in range(len(z0))],0)
##         D[key] = D[key].flat
           D[key] = concatenate(D[key],0)
      print D[key].shape
   return D

def FindGrid(grid,obs,err,w,zpts,filters,offsets=None,fixedz=0,nsol=3,disp=0,train=0,easy=0,fixedm=-99,fixedav=-1,fixedt=None,lmfit=None,keep=None,deltaz=0.05):

    begin = time.time()
    if "age" in grid:
        bc = True
        [z,age,m,av,mags] = [grid[key] for key in ["z","age","m","av","mags"]]
        if keep:
           keep = eval(keep)
           [z,age,m,av] = [compress(keep, x) for x in [z,age,m,av]]
           mags = transpose([compress(keep, x) for x in transpose(mags)])
        if fixedz>0:
           keep = less(abs(z-fixedz),0.03)
           [z,age,m,av] = [compress(keep, x) for x in [z,age,m,av]]
           mags = transpose([compress(keep, x) for x in transpose(mags)])
        if fixedm>-99:
           keep = equal(m,m[argmin(abs(m-fixedm))])
           [z,age,m,av] = [compress(keep, x) for x in [z,age,m,av]]
           mags = transpose([compress(keep, x) for x in transpose(mags)])
        if fixedav>=0:
           keep = equal(av,av[argmin(abs(av-fixedav))])
           [z,age,m,av] = [compress(keep, x) for x in [z,age,m,av]]
           mags = transpose([compress(keep, x) for x in transpose(mags)])
    else:
        bc = False
        [z,ty,mags] = [grid[key] for key in ["z","ty","mags"]]
        tys = Set(ty)
        nty = len(tys)
        tys = ty[:nty]
        if fixedt:
           keep = ones(z.shape,Float32)
           for fixedt1 in fixedt:
              if fixedt1[0] == "-":
                 keep = array([(fixedt1[1:] not in t) for t in ty])
              elif fixedt1[0] == "+":
                 keep = logical_or(keep,array([(fixedt1[1:] in t) for t in ty]))
              else:
                 keep = array([(fixedt1 in t) for t in ty])
              print len(keep), sum(1.0*keep)
           z = compress(keep,z)
           mags = transpose([compress(keep, x) for x in transpose(mags)])
           ty = [ty[j] for j in nonzero(keep)]
        age = 0.0*z
        m = 0.0*z
        av = 0.0*z
        fixedm = 0
        fixedav = 0

    zdict = {}; zdict = zdict.fromkeys(z)
    zset = sort(zdict.keys())
    tdict = {}; tdict = tdict.fromkeys(age)
    tset = sort(tdict.keys())
    if offsets:
      ot1 = time.time()
      print "Computing offsets."
      off = {}
      for zs in zset: off[zs] = GetOffsets(zs,obs,offsets,lage=tset[-1])
      [add(mags[j],off[z[j]],mags[j]) for j,mags1 in enumerate(mags)]
      ot2 = time.time()
      print "Time to compute offsets=",ot2-ot1

    wl = array([f.wleff for f in filters])
    ab = array([f.conv_AB for f in filters])
    fbs = array([f.bandpass1 for f in filters])
    u = less(obs,90)
    print "Rescaling fluxes."
    rt1 = time.time()
    scale = Rescale(obs,err,w,mags,wl,fbs,ab,zpts)
    rt2 = time.time()
    print "Time to compute scales=",rt2-rt1
    mfluxes = exp10(-0.4*(mags+scale[::,NewAxis]-25.0-ab[NewAxis,::]))
    ofluxes = exp10(-0.4*(obs-25.0-ab))*u
    ofluxes = where(u,ofluxes,0.0)
    ferr = where(u,(err/1.0857) * ofluxes, exp10(-0.4*(err-25.0-ab)))
    residuals = mfluxes - ofluxes[NewAxis,::]
    residuals = where(between(residuals,-1e24,1e24),residuals,1e24)
    print "Computing chi^2."
    chi2 = add.reduce(w*power(residuals/ferr[NewAxis,::],2),-1)
    if not bc:
       minty = 1+tys.index(ty[argmin(chi2)])

    minchi2 = min(chi2)
#    chi2 = chi2 - minchi2

    params = {}

    if bc:
       xvars = [age]
       vars = ["log(t)"]
    else:
       xvars = []
       vars = []
       params["m"] = (0.0,0.0,0.0,0.0,0.0)
       params["age"] = (minty,0.0,0.0,0.0,0.0)
       params["av"] = (0.0,0.0,0.0,0.0,0.0)

    if not (fixedz>0) and not train:
       xvars = [z] + xvars
       vars = ["z"] + vars
    else:
       params["z"] = (z,0.0,0.0,0.0,0.0)

    if not (fixedm>-99):
       xvars.append(m)
       vars.append("[Z/H]")
    else:
       params["m"] = (fixedm,0.0,0.0,0.0,0.0)
    if not (fixedav>=0):
       xvars.append(av)
       vars.append("A\dV\u")
    else:
       params["av"] = (fixedav,0.0,0.0,0.0,0.0)

    print vars
    nvar = len(vars)
    if not bc: nvar += 1
    ndeg = sum(w) - nvar

    if disp:
       if train: opendev("/xs",nx=3,ny=3,size=9.0,land=0,asp=0.66)
       else: opendev("/xs",nx=3,ny=4,size=9.0,land=0,asp=1.0)
       print "Doing",disp

    vals = []
    for jx,xvar in enumerate(xvars):
       xdict = {}; xdict = xdict.fromkeys(xvar)
       xs = sort(xdict.keys())
       ys = array([min(compress(equal(xvar,x1),chi2)) for x1 in xs])
       ys = where(between(ys,-1e24,1e24),ys,1e24)
       dx0 = mean(diff(xs))
       xs1,xs2 = extrema(xs)
       lo,hi = extrema(ys)
       var = vars[jx]
       tck = splrep(xs,ys,s=0,k=3)
       ltck = splrep(xs,getlog10(ys),s=0,k=3)
       xmin = xs[argmin(ys)]
       peak = xmin
       minx,maxx = extrema(xs)
       rchi2 = (lo+0*minchi2)/ndeg
       params["chi2"] = lo+0*minchi2

       dx = 0.002
       xsa = arange(minx,maxx,dx,Float32)
       csa = splev(xsa,ltck)
       ysa = exp10(csa)

       zs = chdtrc(ndeg,ys)
       probs = chdtrc(ndeg,exp10(csa))
       ptck = splrep(xsa,probs,s=0,k=3)
       pnorm = splint(minx,maxx,ptck)

       if pnorm:
          probs = probs/pnorm
          zs = zs/pnorm

       ptck = splrep(xsa,probs,s=0,k=5)
       cint = array([splint(xsa[0],x,ptck) for x in xsa])

       sc1, sc2 = 0.5 + 0.3427*array([-1,1])
       pc1, pc2 = 0.5 + 0.475*array([-1,1])

       xsa2,cint2 = take([xsa,cint],argsort(cint),1)
       guess = clip(searchsorted(cint2,0.5),0,len(xsa2)-1)
       midp = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-0.5)**2,[xsa2[guess]]),minx,maxx)

       guess = clip(searchsorted(cint2,sc1),0,len(xsa2)-1)
       sigm1 = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-sc1)**2,[xsa2[guess]]),minx,maxx)
       guess = clip(searchsorted(cint2,sc2),0,len(xsa2)-1)
       sigm2 = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-sc2)**2,[xsa2[guess]]),minx,maxx)

       guess = clip(searchsorted(cint2,pc1),0,len(xsa2)-1)
       conf1 = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-pc1)**2,[xsa2[guess]]),minx,maxx)
       guess = clip(searchsorted(cint2,pc2),0,len(xsa2)-1)
       conf2 = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-pc2)**2,[xsa2[guess]]),minx,maxx)

       if conf2 == conf1: xsa2,probs2 = [xsa,probs]
       else: xsa2,probs2 = compress(between(xsa,conf1,conf2),[xsa,probs],1)
       guess = argmax(probs2)

       peak = clip(fsolve(lambda x: splev(x,ptck,1)**2, xsa2[guess]),minx,maxx)
       if abs(peak-xsa2[guess]) > dx0: peak = xsa2[guess]

       err1 = peak-sigm1
       err2 = sigm2-peak
       con1 = peak-conf1
       con2 = conf2-peak
       print "%s=%9.3f(-%.3f+%.3f)(-%.3f+%.3f)" % (var,peak,err1,err2,con1,con2)
       print "%s=%9.3f(-%.3f+%.3f)(-%.3f+%.3f)" % (var,peak,err1,err2,con1,con2)
       print
       vals += [peak,err1,err2,con1,con2]

       if var == "z": params["z"] = (peak,err1,err2,con1,con2)
       if var == "log(t)": params["age"] = (peak,err1,err2,con1,con2)
       if var == "[Z/H]": params["m"] = (peak,err1,err2,con1,con2)
       if var == "A\dV\u": params["av"] = (peak,err1,err2,con1,con2)

       if var == "z" and fixedz < 0.0:
          z1 = abs(fixedz)-deltaz
          z2 = abs(fixedz)+deltaz
#          zprobt = splint(z1-2*deltaz,z2+deltaz,ptck)
#          zprob0 = splint(z1-2*deltaz,z1-deltaz,ptck)/zprobt
#          zprob1 = splint(z2+deltaz,z2+2*deltaz,ptck)/zprobt
          zprob = splint(z1,z2,ptck)
          print "Prob=",zprob
          zprob = zprob
          params["zprob"] = getlog10(zprob)
       else: zprob = -1.0

       if disp:
          newbox([minx,maxx],[0.0,lo+20*ndeg])
          xlabel(var,ch=2.2)
          ylabel("\gx\u2\d",ch=2.2)
          drawpt(xs,ys,ps=17,ch=3)
          drawln(xsa,ysa,ls=1,ci=1,lw=3)
          drawln([conf1,conf1],[0.0,lo+20*ndeg],ls=4,ci=1)
          drawln([peak,peak],[0.0,lo+20*ndeg],ls=2,ci=1)
          drawln([conf2,conf2],[0.0,lo+20*ndeg],ls=4,ci=1)

          newbox([minx,maxx],[0.0,1.05*max(probs*dx)])
          xlabel(var,ch=2.2)
          ylabel("P(\gx\u2\d|\gn)d%s" % (var),ch=2.2)
          drawpt(xs,zs*dx,ps=17,ch=3)
          drawln(xsa,probs*dx,ls=1,ci=1,lw=3)
          drawln([conf1,conf1],[0.0,1.1*max(probs*dx)],ls=4,ci=1)
          drawln([peak,peak],[0.0,1.1*max(probs*dx)],ls=2,ci=1)
          drawln([conf2,conf2],[0.0,1.1*max(probs*dx)],ls=4,ci=1)
          if var == vars[0]: subtitle("#%s" % (disp),ch=3.0)

          newbox([minx,maxx],[0,1.0])
          xlabel(var,ch=2.2)
          ylabel("\(2268)P(\gx\u2\d|\gn)d%s" % (var),ch=2.2)
          drawln(xsa,cint,ls=1,ci=1,lw=3)

          drawln([conf1,conf1],[0.0,1.1],ls=4,ci=1)
          drawln([peak,peak],[0.0,1.1],ls=2,ci=1)
          drawln([conf2,conf2],[0.0,1.1],ls=4,ci=1)
          drawln([minx,maxx],[0.5,0.5],ls=2,ci=1)
          drawln([minx,maxx],[pc1,pc1],ls=4,ci=1)
          drawln([minx,maxx],[pc2,pc2],ls=4,ci=1)

    if disp: closedev()
    if lmfit and sum(probs):
        print params
        if fixedav == -1 and params["av"] == 0.0: params["av"] += 0.001
        lmresults = LMRedshift(obs,err,w,filters,lmfit,z0=params["z"],age0=params["age"],m0=params["m"],av0=params["av"],\
                verb=0,offsets=offsets,fixedz=fixedz,fixedav=fixedav,factor=0.01)
        rchi2 = lmresults[3]
        if fixedz == 0:
           vals[0] = lmresults[0]
           vals[3] = lmresults[4]
           vals[6] = lmresults[5]
           if fixedav == -1: vals[9] = lmresults[6]
        else:
           vals[0] = lmresults[4]
           vals[3] = lmresults[5]
           if fixedav == -1: vals[6] = lmresults[6]

    end = time.time() 
    print "Elapsed time=",end-begin

    return params,rchi2

def FullGrid(grid,obs,err,w,zpts,filters,offsets=None,fixedz=0,nsol=3,disp=0,train=0,easy=0,fixedm=-99,fixedav=-1,fixedtau=-1,keep=None,priors=None,deltaz=0.05,massp=1,mkp=0):
    begin = time.time()

    if "age" in grid:
       bc = 1
       [z,age,m,av,tau,mags] = [grid[key] for key in ["z","age","m","av","tau","mags"]]
    else:
       bc = 0
       [z,ty,mags] = [grid[key] for key in ["z","ty","mags"]]
       fixedt = 9
       fixedm = 0
       fixedav = 0
       fixedtau = 0

    if keep:
       keep = eval(keep)
       if bc: [z,age,m,av,tau] = [compress(keep, x) for x in [z,age,m,av,tau]]
       else: [z,ty,av] = [compress(keep, x) for x in [z,ty,av]]
       mags = transpose([compress(keep, x) for x in transpose(mags)])
       priors = compress(keep,priors)

    if fixedz>0:
       keep = less(abs(z-fixedz),0.05)
       if bc: [z,age,m,av,tau] = [compress(keep, x) for x in [z,age,m,av,tau]]
       else: [z,ty,av] = [compress(keep, x) for x in [z,ty,av]]
       mags = transpose([compress(keep, x) for x in transpose(mags)])
       priors = compress(keep,priors)

    if bc and fixedm>-99:
       keep = equal(m,m[argmin(abs(m-fixedm))])
       [z,age,m,av,tau] = [compress(keep, x) for x in [z,age,m,av,tau]]
       mags = transpose([compress(keep, x) for x in transpose(mags)])
       priors = compress(keep,priors)

    if bc and fixedav>=0:
       keep = equal(av,av[argmin(abs(av-fixedav))])
       [z,age,m,av,tau] = [compress(keep, x) for x in [z,age,m,av,tau]]
       mags = transpose([compress(keep, x) for x in transpose(mags)])
       priors = compress(keep,priors)

    if bc and fixedtau>=0:
       keep = equal(tau,tau[argmin(abs(tau-fixedtau))])
       [z,age,m,av,tau] = [compress(keep, x) for x in [z,age,m,av,tau]]
       mags = transpose([compress(keep, x) for x in transpose(mags)])
       priors = compress(keep,priors)

    agemax = zeros(z.shape,Float64)
    zs = Set(grid["z"])
    [put(agemax,nonzero(equal(z,zs1)),\
        maxlage(zs1)*ones((sum(equal(z,zs1)),),Float64)) for zs1 in zs]

    xdict = {}
    xdict = xdict.fromkeys(z)
    zset = sort(xdict.keys())
    if bc:
      xdict = xdict.fromkeys(age)
      tset = sort(xdict.keys())
      xdict = xdict.fromkeys(m)
      mset = sort(xdict.keys())
      xdict = xdict.fromkeys(av)
      avset = sort(xdict.keys())
      xdict = xdict.fromkeys(tau)
      tauset = sort(xdict.keys())
    else:
      xdict = xdict.fromkeys(ty)
      tset = xdict.keys()
      tset.sort()

    if len(zset) > 1: dzset = diff(zset)[0]
    else: dzset = 1.0

    if bc:
      if len(tset) > 1: dtset = diff(tset)[0]
      else: dtset = 1.0
      if len(mset) > 1: dmset = diff(mset)[0]
      else: dmset = 1.0
      if len(avset) > 1: davset = diff(avset)[0]
      else: davset = 1.0
      if len(tauset) > 1: dtauset = diff(tauset)[0]
      else: dtauset = 1.0

      totau = age - tau
      totau = sign(totau)*(abs(totau)*100+0.5).astype(Int32)/100.0
      xdict = xdict.fromkeys(totau.flat)
      totauset = sort(xdict.keys())
      dtotauset = diff(totauset)[0]

    if offsets:
      ot1 = time.time()
      print "Computing offsets."
      off = {}
      for zs in zset: off[zs] = GetOffsets(zs,obs,offsets,lage=tset[-1])
      [add(mags[j],off[z[j]],mags[j]) for j,mags1 in enumerate(mags)]
      ot2 = time.time()
      print "Time to compute offsets=",ot2-ot1

    wl = array([f.wleff for f in filters])
    ab = array([f.conv_AB for f in filters])
    fbs = array([f.bandpass1 for f in filters])
    u = less(obs,90)
    print "Rescaling fluxes."
    rt1 = time.time()
    scale = Rescale(obs,err,w,mags,wl,fbs,ab,zpts)
    rt2 = time.time()
    print "Time to compute scales=",rt2-rt1
    mfluxes = exp10(-0.4*(mags+scale[::,NewAxis]-25.0-ab[NewAxis,::]))
    ofluxes = exp10(-0.4*(obs-25.0-ab))*u
    ofluxes = where(u,ofluxes,0.0)
    ferr = where(u,(err/1.0857) * ofluxes, exp10(-0.4*(err-25.0-ab)))
    residuals = mfluxes - ofluxes[NewAxis,::]
    residuals = where(between(residuals,-1e24,1e24),residuals,1e24)
    print "Computing chi^2."
    chi2 = add.reduce(w*power(residuals/ferr[NewAxis,::],2),-1)
    chi2 = where(equal(chi2,0),1e24,chi2)

    minchi2 = min(chi2)
    print "Minchi2=",minchi2

    params = {}

    if bc:
      xvars = [age]
      vars = ["log(t)"]
    else:
      xvars = []
      vars = []
      params["age"] = (0.0,0.0,0.0,0.0,0.0)

    if not (fixedz>0) and not train:
       xvars = [z] + xvars
       vars = ["z"] + vars
    else:
       params["z"] = (z,0.0,0.0,0.0,0.0)

    if not (fixedm>-99):
       xvars.append(m)
       vars.append("[Z/H]")
    else:
       params["m"] = (fixedm,0.0,0.0,0.0,0.0)

    if fixedav>=0:
       params["av"] = (fixedav,0.0,0.0,0.0,0.0)
    else:
       xvars.append(av)
       vars.append("A\dV\u")

    if fixedtau>=0:
       params["tau"] = (fixedtau,0.0,0.0,0.0,0.0)
       params["totau"] = (0.0,0.0,0.0,0.0,0.0)
    else:
       xvars.append(tau)
       vars.append("log(tau)")
       xvars.append(totau)
       vars.append("log(t/tau)")

    print vars
    nvar = len(vars)
    ndeg = sum(w) - nvar
    probs = chdtrc(ndeg,chi2)

    masses = -0.4*scale
    if massp:
       masse = -0.5*power(masses-12.0,2)/power(0.5,2)
       massg = where(between(masses,0,40),exp(masse),0.0)
       priors = priors * massg
#       priors = priors * between(masses,9.5,13.5,inc=1)

    if priors:
       probs = priors * probs
#       chi = where(priors,chi,1e24)
       chi = where(probs,chdtri(ndeg,probs),1e24)

    if bc:
      rshape = (len(tset),len(mset),len(tauset),len(avset),len(zset))
    else:
      rshape = (len(tset),len(zset))

    allz = reshape(z,rshape)
    allchi = reshape(chi2,rshape)
    allscale = reshape(scale,rshape)
    allmasses = reshape(masses,rshape)
    allprobs = reshape(probs,rshape)
    allpriors = reshape(priors,rshape)

    if bc:
       alltl = reshape(age,rshape)
       allmtl = reshape(agemax,rshape)
       allm = reshape(m,rshape)
       allav = reshape(av,rshape)
       alltau = reshape(tau,rshape)
       alltotau = reshape(totau,rshape)

    sc1, sc2 = 0.5 + 0.3427*array([-1,1])
    pc1, pc2 = 0.5 + 0.475*array([-1,1])

    LogMass=allmasses.flat[argmax(allprobs.flat)]
    print quartiles(allmasses.flat)
    print extrema(allmasses.flat)
    print "LogMass=",LogMass
    LogMass1, LogMass2 = 0.0, 0.0

    sortmasses,sortprobs = compress(allprobs.flat, [allmasses.flat,allprobs.flat], 1)
    if len(sortprobs):
       sortmasses,sortprobs = take([sortmasses,sortprobs],argsort(sortmasses),1)
       sortprobs = sortprobs/sum(sortprobs)
       useprobs = greater(sortprobs,max(sortprobs)/1e10)
       if useprobs:
          sortmasses,sortprobs = compress(useprobs,[sortmasses,sortprobs],1)
          intmprobs = add.accumulate(sortprobs)
          mtck = splrep(intmprobs,sortmasses,k=1,s=0)
          LogMass1, LogMass2 = splev([sc1,sc2],mtck)
          print LogMass1,LogMass2
          LogMass1, LogMass2 = abs(LogMass-array([LogMass1,LogMass2]))
          print LogMass1,LogMass2

    if bc:
       minchi_z = minimum.reduce(minimum.reduce(minimum.reduce(minimum.reduce(allchi,3),2),1),0)
       minchi_t = minimum.reduce(minimum.reduce(minimum.reduce(minimum.reduce(allchi,4),3),2),1)
       minchi_m = minimum.reduce(minimum.reduce(minimum.reduce(minimum.reduce(allchi,4),3),2),0)
       minchi_tau = minimum.reduce(minimum.reduce(minimum.reduce(minimum.reduce(allchi,4),3),1),0)
       minchi_av = minimum.reduce(minimum.reduce(minimum.reduce(minimum.reduce(allchi,4),2),1),0)
       minchi_totau = array([min(compress(equal(alltotau.flat,tot),allchi.flat)) for tot in totauset])
    else:
       minchi_z = minimum.reduce(allchi,1)

    if bc:
       allprobs_z  = add.reduce(add.reduce(add.reduce(add.reduce(allprobs,3),2),1),0)*dtset*dmset*davset*dtauset
       allprobs_t  = add.reduce(add.reduce(add.reduce(add.reduce(allprobs,4),3),2),1)*dzset*dmset*davset*dtauset
       allprobs_m  = add.reduce(add.reduce(add.reduce(add.reduce(allprobs,4),3),2),0)*dzset*dtset*davset*dtauset
       allprobs_tau = add.reduce(add.reduce(add.reduce(add.reduce(allprobs,4),3),1),0)*dzset*dtset*davset
       allprobs_av = add.reduce(add.reduce(add.reduce(add.reduce(allprobs,4),2),1),0)*dzset*dtset*dmset*dtauset
       allprobs_tm  = 1*transpose(add.reduce(add.reduce(add.reduce(allprobs,4),3),2)*dzset*davset*dtauset)
       allprobs_tm = divz(allprobs_tm,add.reduce(add.reduce(allprobs_tm,1),0)*dtset*dmset)
       allprobs_ttau  = 1*transpose(add.reduce(add.reduce(add.reduce(allprobs,4),3),1)*dzset*davset*dmset)
       allprobs_ttau = divz(allprobs_ttau,add.reduce(add.reduce(allprobs_ttau,1),0)*dtset*dtauset)
       allprobs_tav  = 1*transpose(add.reduce(add.reduce(add.reduce(allprobs,4),2),1)*dzset*dtauset*dmset)
       allprobs_tav = divz(allprobs_tav,add.reduce(add.reduce(allprobs_tav,1),0)*dtset*davset)
       allprobs_totau = array([sum(compress(equal(alltotau.flat,tot),allprobs.flat*dzset*dmset*davset)) for tot in totauset])
       allprobs_totau = allprobs_totau / sum(allprobs_totau*dtotauset)
    else:
       allprobs_z  = add.reduce(allprobs ,1)

    if mkp:
       return getlog10(allprobs_z)

    lo = minchi_z[argmax(allprobs_z)]
    rchi2 = lo/ndeg
    params["chi2"] = lo

    if disp:
       if train:
         nx = 3
         ny = 3
       else: 
         nx = 6
         ny = 3
       opendev("/xs",nx=nx,ny=ny,size=3.0*nx,land=0,asp=1.*ny/nx)
       print "Doing",disp

       newbox([min(tset),max(tset)],[min(mset),max(mset)],ch=2.3)
       xlabel("log t",ch=2.3)
       ylabel("[Fe/H]",ch=2.3)
       xfm = array([min(tset)-dtset/2.,dtset,0.0,min(mset)-dmset/2.,0.0,dmset])
       ydim,xdim = allprobs_tm.shape
       cpggray(allprobs_tm,xdim,ydim,1,xdim,1,ydim,0.5*max(allprobs_tm.flat),0,xfm)

       newbox([min(tset),max(tset)],[min(tauset),max(tauset)],ch=2.3)
       xlabel("log t",ch=2.3)
       ylabel("log \gt",ch=2.3)
       xfm = array([min(tset)-dtset/2.,dtset,0.0,min(tauset)-dtauset/2.,0.0,dtauset])
       ydim,xdim = allprobs_ttau.shape
       cpggray(allprobs_ttau,xdim,ydim,1,xdim,1,ydim,0.5*max(allprobs_ttau.flat),0,xfm)

       if len(avset) == 1:
          cpgpage()
       else:
          newbox([min(tset),max(tset)],[min(avset),max(avset)],ch=2.3)
          xlabel("log t",ch=2.3)
          ylabel("A\dV\u",ch=2.3)
          xfm = array([min(tset)-dtset/2.,dtset,0.0,min(avset)-davset/2.,0.0,davset])
          ydim,xdim = allprobs_tav.shape
          cpggray(allprobs_tav,xdim,ydim,1,xdim,1,ydim,0.5*max(allprobs_tav.flat),0,xfm)

    stuff = {}
    for jx,xvar in enumerate(xvars):
       xdict = {}; xdict = xdict.fromkeys(xvar)
       xs = sort(xdict.keys())
#       print vars[jx],xs
       dxs = diff(xs)[0]
       stuff[vars[jx]] = {"xs":xs,"dxs":dxs}

    if "z" in stuff:
       stuff["z"]["P"] = allprobs_z
       stuff["z"]["chi2"] = minchi_z

    if "log(t)" in stuff:
       stuff["log(t)"]["P"] = allprobs_t
       stuff["log(t)"]["chi2"] = minchi_t

    if "[Z/H]" in stuff:
       stuff["[Z/H]"]["P"] = allprobs_m
       stuff["[Z/H]"]["chi2"] = minchi_m

    if "A\dV\u" in stuff:
       stuff["A\dV\u"]["P"] = allprobs_av
       stuff["A\dV\u"]["chi2"] = minchi_av

    if "log(tau)" in stuff:
       stuff["log(tau)"]["P"] = allprobs_tau
       stuff["log(tau)"]["chi2"] = minchi_tau

    if "log(t/tau)" in stuff:
       stuff["log(t/tau)"]["P"] = allprobs_totau
       stuff["log(t/tau)"]["chi2"] = minchi_totau


    vals = []
    for jx,xvar in enumerate(xvars):
       xs = stuff[vars[jx]]["xs"]
       dxs = stuff[vars[jx]]["dxs"]
       P = stuff[vars[jx]]["P"]
       ys = stuff[vars[jx]]["chi2"]
       cP = compress(P,P)
       cys = compress(ys,ys)
       if cP: mcP = min(getlog10(cP))
       else: mcP = -99
       if cys: mcys = min(getlog10(cys))
       else: mcys = -99
       lP = getlog10(P,rep=mcP)
       lys = getlog10(ys,rep=mcys)

       xs1,xs2 = extrema(xs)
       lo,hi = extrema(ys)

       var = vars[jx]
       tck = splrep(xs,ys,s=0,k=3)
       aptck = splrep(xs,lP,s=0,k=3)
       ltck = splrep(xs,lys,s=0,k=3)

       xmin = xs[argmin(ys)]
       peak = xmin
       minx,maxx = extrema(xs)

#       dx = 0.002
       dx = dxs/5
       xsa = arange(minx,maxx,dx,Float32)
       schi2 = exp10(splev(xsa,ltck))
       probs = exp10(splev(xsa,aptck))
       ptck = splrep(xsa,probs,s=0,k=3)
       pnorm = splint(minx,maxx,ptck)
       if pnorm:
          probs = divz(probs,pnorm)

       ptck = splrep(xsa,probs,s=0,k=3)
       cint = array([splint(xsa[0],x,ptck) for x in xsa])

       xsa2,cint2 = take([xsa,cint],argsort(cint),1)
       guess = clip(searchsorted(cint2,0.5),0,len(xsa2)-1)
       midp = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-0.5)**2,[xsa2[guess]]),minx,maxx)

       guess = clip(searchsorted(cint2,sc1),0,len(xsa2)-1)
       sigm1 = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-sc1)**2,[xsa2[guess]]),minx,maxx)
       guess = clip(searchsorted(cint2,sc2),0,len(xsa2)-1)
       sigm2 = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-sc2)**2,[xsa2[guess]]),minx,maxx)

       guess = clip(searchsorted(cint2,pc1),0,len(xsa2)-1)
       conf1 = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-pc1)**2,[xsa2[guess]]),minx,maxx)
       guess = clip(searchsorted(cint2,pc2),0,len(xsa2)-1)
       conf2 = clip(fsolve(lambda x: (splint(xsa[0],x[0],ptck)-pc2)**2,[xsa2[guess]]),minx,maxx)

       if conf2 == conf1: xsa2,probs2 = [xsa,probs]
       else: xsa2,probs2 = compress(between(xsa,conf1,conf2),[xsa,probs],1)

       if len(probs2):
          guess = argmax(probs2)
       else:
          guess = 0
          xsa2 = [0]

       peak = clip(fsolve(lambda x: splev(x,ptck,1)**2, xsa2[guess]),minx,maxx)
       if abs(peak-xsa2[guess]) > dxs: peak = xsa2[guess]

       err1 = abs(peak-sigm1)
       err2 = abs(sigm2-peak)
       con1 = peak-conf1
       con2 = conf2-peak
       print "%s=%9.3f(-%.3f+%.3f)(-%.3f+%.3f)" % (var,peak,err1,err2,con1,con2)
       print "%s=%9.3f(-%.3f+%.3f)(-%.3f+%.3f)" % (var,peak,err1,err2,con1,con2)
       print
       vals += [peak,err1,err2,con1,con2]

       if var == "z": params["z"] = (peak,err1,err2,con1,con2)
       if var == "log(t)": params["age"] = (peak,err1,err2,con1,con2)
       if var == "[Z/H]": params["m"] = (peak,err1,err2,con1,con2)
       if var == "A\dV\u": params["av"] = (peak,err1,err2,con1,con2)
       if var == "log(tau)": params["tau"] = (peak,err1,err2,con1,con2)
       if var == "log(t/tau)": params["totau"] = (peak,err1,err2,con1,con2)

       if var == "z" and fixedz < 0.0:
          z1 = abs(fixedz)-deltaz
          z2 = abs(fixedz)+deltaz
          zprob = splint(z1,z2,ptck)
          print "Prob=",zprob
          zprob = zprob
          params["zprob"] = getlog10(zprob)
          if not bc:
             tyz,probsz = compress(between(z,peak-err1,peak+err2),[ty,probs],1)
             params["age"][0] = tyz[argmax(probsz)]
       else:
          zprob = -999.99
          params["zprob"] = zprob

       if disp:
          xmar = [minx,maxx]
          newbox(xmar,[0.0,lo+20*ndeg],ch=2.3)
          xlab = var.replace("tau","\gt")
          xlabel(xlab,ch=2.3)
          ylabel("\gx\u2\d",ch=2.3)
          drawln(xsa,schi2,ls=1,ci=1,lw=3)
          drawln([conf1,conf1],[0.0,lo+20*ndeg],ls=4,ci=1)
          drawln([peak,peak],[0.0,lo+20*ndeg],ls=2,ci=1)
          drawln([conf2,conf2],[0.0,lo+20*ndeg],ls=4,ci=1)

          newbox(xmar,[log10(1.05*max(probs*dx))-5,log10(1.05*max(probs*dx))],ch=2.3,ylog=1)
          xlabel(xlab,ch=2.3)
          ylabel("P(\gx\u2\d|\gn)d%s" % (xlab),ch=2.3)
          drawln(xsa,getlog10(probs*dx),ls=1,ci=1,lw=3)

          drawln([conf1,conf1],getlog10([0.0,1.1*max(probs*dx)]),ls=4,ci=1)
          drawln([peak,peak],getlog10([0.0,1.1*max(probs*dx)]),ls=2,ci=1)
          drawln([conf2,conf2],getlog10([0.0,1.1*max(probs*dx)]),ls=4,ci=1)

          if var == vars[0]: subtitle("#%s" % (disp),ch=3.0)

          newbox(xmar,[0,1.0],ch=2.3)
          xlabel(xlab,ch=2.3)
          ylabel("\(2268)P(\gx\u2\d|\gn)d%s" % (xlab),ch=2.3)
          drawln(xsa,cint,ls=1,ci=1,lw=3)

          drawln([conf1,conf1],[0.0,1.1],ls=4,ci=1)
          drawln([peak,peak],[0.0,1.1],ls=2,ci=1)
          drawln([conf2,conf2],[0.0,1.1],ls=4,ci=1)
          drawln([minx,maxx],[0.5,0.5],ls=2,ci=1)
          drawln([minx,maxx],[pc1,pc1],ls=4,ci=1)
          drawln([minx,maxx],[pc2,pc2],ls=4,ci=1)

    if disp: closedev()

    end = time.time() 
    print "Elapsed time=",end-begin

    return params,rchi2,(LogMass,LogMass1,LogMass2)


class FauxTemplate:
   def __init__(self,lam,flam):
      self.lam = lam
      self.flam = flam




