#!/usr/bin/env python

from sys import stdout
from types import *
from math import atan2
from VTKHelperFunctions import *
from masks import *
from ephem import *
from basics import *

class frame:
    def __init__(self,n):
       self.image = n
       self.catalog = replace(self.image,".fits",".ast")
    def sex(self,sexdir="./sex/",thresh=5.0,saturate=30000,scale=0.2):
       self.scale = scale
       self.sexcom = ["sex",self.image]
       self.sexcom += ["-c "+sexdir+"default.sex"]
       self.sexcom += ["-PARAMETERS_NAME "+sexdir+"default.param"]
       self.sexcom += ["-DETECT_MINAREA 5"]
       self.sexcom += ["-DETECT_THRESH "+str(thresh)]
       self.sexcom += ["-ANALYSIS_THRESH "+str(thresh)]
       self.sexcom += ["-PIXEL_SCALE "+str(self.scale)]
       self.sexcom += ["-SEEING_FWHM "+str(0.8/self.scale)]
       self.sexcom += ["-CATALOG_NAME",self.catalog]
#        self.sex += ["-FILTER_NAME "+sexdir+"default.conv"]
       self.sexcom += ["-FILTER_NAME "+sexdir+"gauss_3.0_5x5.conv"]
       self.sexcom += ["-STARNNW_NAME "+sexdir+"default.nnw"]
       self.sexcom += ["-SATUR_LEVEL "+str(saturate)]
       self.sexcom = join(self.sexcom)
       verbex(self.sexcom)
    def readcat(self,refcat=None,nuse=None,clean=0):
       self.nuse = nuse
       self.dat = readtable(self.catalog,icols=[0,7],start=9)
       self.imx = self.dat[1]; self.imy = self.dat[2]
       self.catdat = self.dat
       self.catmatch = matching(self.catdat,nuse=self.nuse,cull=clean,errlim=0.2)
       self.refdat = readtable(refcat,icols=[0,7],start=9)
       self.refmatch = matching(self.refdat,nuse=self.nuse,cull=clean,errlim=0.2)
    def matchup(self,verb=0,mtol=1,ttol=1,nterms=1):
       self.bx,self.by,self.bt= self.catmatch.match(1,self.refmatch,verb=verb,im=self.image,mtol=mtol,ttol=ttol,simple=1)
       print "Found (bx,sx):",self.bx
       print "Found (by,sy):",self.by
       f = FITS(self.image,readonly=0)
       f.modkey("xtrans",self.bx[0])
       f.modkey("ytrans",self.by[0])
       f.close()

class rotsol:
   def __init__(self,fixsol,fac=1.0):
      self.rot = (pi/2 - apply(atan2,fixsol[2:]))
      self.xsh,self.ysh = fixsol[:2]
      self.cos = cos(self.rot)
      self.sin = sin(self.rot)
      self.rescale = 1.0/sum(fixsol[2:]**2)
      print "Shifts, dPA, Scale = (%.2f,%.2f) (arcsec), %.3f (deg), %3.3f " % (self.xsh,self.ysh,raddeg(self.rot),self.rescale)

class matching:
   def __init__(self,d,nuse=50,errlim=0.1,cull=1,sky=0,verb=0):
       n, x, y, m, e, ma, ea, fl = d[:8]
       n = arange(len(x))
       u = arange(len(n))

       if cull:
          u = less(ea,errlim)*between(ma-m,1.0,4.0)
#          u = u*not_equal(fl,4)*less(fl,30)
          u = u*less(fl,30)
          u = nonzero(u)
          n = take(n,u)
          x,y,m,e,ma,ea,fl = take([x,y,m,e,ma,ea,fl],u,1)
          un = n; ux,uy,um,ue = x,y,m,e
#          un = take(n,u); ux,uy,um,ue = take([x,y,m,e],u,1)
          print "Culling",len(un),
          if nuse:
              a = argsort(m)[:nuse]
              n = take(n,a)
              x,y,m,e,ma,ea,fl = take([x,y,m,e,ma,ea,fl],a,1)
              u = arange(len(x))
              un = n; ux,uy,um,ue = x,y,m,e
              print "and using",len(un)
          else: print
       else:
          a = argsort(m)
          if nuse: a = a[:nuse]
          n = take(n,a)
          x,y,m,e, = take([x,y,m,e],a,1)
          un = n; ux,uy,um,ue = x,y,m,e
          u = arange(len(n))
       if nuse:             
           if verb: print "There are",len(n),"reference objects."

           id = []
           n1 = []; x1 = []; y1 = []; m1 = []; e1 = []; s1 = []; t1=[]
           for j in u:
              n2 = []; x2 = []; y2 = []; m2 = []; e2 = []; s2 = []; t2=[]
              for k in u:
                 dx = x[j] - x[k]
                 dy = y[j] - y[k]
                 ds = sqrt(dx**2+dy**2)
                 dt = raddeg(atan2(dy,dx)-pi/2)
                 n2.append( n[k]); x2.append( x[k]); y2.append( y[k])
                 m2.append( m[k]); e2.append( e[k])
                 s2.append( ds); t2.append( dt)
              n2 = asarray(n2,Int32)
              x2,y2,m2,e2,s2,t2 = asarray([x2,y2,m2,e2,s2,t2])
              o2 = argsort(s2)
              n2 = take(n2,o2)
              x2,y2,m2,e2,s2,t2 = take([x2,y2,m2,e2,s2,t2],o2,1)
              id.append(n[j])
              n1.append(n2); x1.append(x2); y1.append(y2)
              m1.append(m2); e1.append(e2)
              s1.append(s2); t1.append(t2)
           id = asarray(id,Int32)
           n1 = asarray(n1,Int32)
           x1,y1,m1,e1,s1,t1 = asarray([x1,y1,m1,e1,s1,t1])
           self.id = id; self.n1 = n1; self.x1 = x1; self.y1 = y1
           self.m1 = m1; self.e1 = e1; self.s1 = s1; self.t1 = t1
       self.n = un; self.x = ux; self.y = uy
       self.m = um; self.e = ue

   def match(self,Jim,other,verb=0,im=None,mtol=1.0,ttol=1.0,simple=0):
       sys.stdout.flush()
       print "Entering matching algorithm:",verb
       J1,J2=len(self.n1),len(other.n1)
       b = []; n = []; x = []; y = []
       for j1 in range(J1):
          if len(n) == J2: break
          print j1,
          if j1 and not j1%25: print
          sys.stdout.flush()
          possx = []
          possy = []
          possn = []
          possb = []
          possm = []
          for j2 in range(J2):
            if j2 not in b:
               ts = (self.s1[j1][::,NewAxis]-other.s1[j2][NewAxis,::])
               ss = less(abs(ts),mtol)
               ss[0,0] = 0
               if verb == 3:
                  print
                  print j1,j2
                  for temp in ts: print temp.astype(Int32)
                  for temp in ss: print temp.astype(Int32)
               if sum(ss.flat) > 0:
                   tt = self.t1[j1][::,NewAxis]-other.t1[j2][NewAxis,::]
                   st = less(abs(tt),ttol)
                   st[0,0] = 0
                   mat = (ss*st)
                   tmat = sum(mat.flat)
                   if verb == 3:
                      for temp in tt: print temp.astype(Int32)
                      for temp in st: print temp.astype(Int32)
                      print
                      for temp in mat: print temp
                      print tmat
                   if tmat > 3:
                      if verb > 1:
                          print j1,j2,tmat
                          spaces = compress(mat.flat,ts.flat)
                          thetas = compress(mat.flat,tt.flat)
                          print spaces
                          print thetas
                          print
                      possx.append(self.x[j1])
                      possy.append(self.y[j1])
                      possn.append(self.n[j1])
                      possb.append(j2)
                      possm.append(tmat)
                      break
          if len(possm):
            if verb > 1:
               print j1, possb
            use = argmax(possm)
            x.append(possx[use])
            y.append(possy[use])
            b.append(possb[use])
            n.append(possn[use])
       x,y = asarray([x,y])
       print
       noth = take(other.n,b)
       xo = take(other.x,b)
       yo = take(other.y,b)
       if verb: print xo-x
       if verb: print yo-y
       bx,sx = bwt(x-xo)
       by,sy = bwt(y-yo)
       if verb: len(xo)
       if verb: print bx,sx
       if verb: print by,sy
       if bx == sx == 0.0: sx = 10
       if by == sy == 0.0: sy = 10
       usex = between(x-xo,bx-4*sx,bx+4*sy)
       usey = between(y-yo,by-4*sy,by+4*sy)
       if verb: print usex*usey
       xo,yo,x,y = compress(usex*usey,[xo,yo,x,y],1)
       noth,n = compress(usex*usey,[noth,n],1)
       if verb: print xo-x
       if verb: print yo-y
       print "Matched:",len(xo)
       print "Extrema (x):",extrema(xo)
       print "Extrema (y):",extrema(yo)
       if simple:
          bx = bwt(x-xo)
          by = bwt(y-yo)
          rt = array([(atan2(y[j]-mean(y),x[j]-mean(x))-atan2(yo[j]-mean(yo),xo[j]-mean(xo))) for j in range(len(x))])
          bt = bwt(rt)
          return bx,by,bt
       else:
          basis = abasis(0,xo,yo,rot=1)
          dat = concatenate([x,y])
          sol = svdfit(basis,dat)
          print sol
          f = add.reduce(sol[NewAxis,::]*basis,1)
          fx = f[:len(x)]; fy = f[len(x):]
          return sol,[n,noth]

   def xfm(self,wirc,tran=[0,0,1,0],verb=0,mtol=3.0,nord=1,arcsec=0):
      other = wirc.catmatch
      if tran: tran = array(tran)
      else: tran = wirc.tran
      if arcsec:
         fac = 1.0; xc = 0.0; yc = 0.0
         aix = self.x; aiy = self.y
         arx = other.x; ary = other.y
         print len(aix), len(arx)
      else:
         fac = (3600*180/pi)
         xc = wirc.xc; yc = wirc.yc
         aix = wirc.imx - xc
         aiy = wirc.imy - yc
         arx = other.x; ary = other.y
      for j in range(3):
          rx = []; ry = []; ix = []; iy = []; fx = []; fy = []
          b = abasis(nord,aix,aiy,rot=1)
          f = add.reduce(tran[NewAxis,::]*b,1)*fac
          afx = f[:len(aix)]; afy = f[len(aix):]
          for k in range(len(aix)):
             dx = power(afx[k]-arx,2)
             dy = power(afy[k]-ary,2)
             ds = sqrt(dx+dy)
             l = argmin(ds)
             if ds[l] < mtol:
                ix += [aix[k]]; iy += [aiy[k]]
                fx += [afx[k]]; fy += [afy[k]]
                rx += [arx[l]]; ry += [ary[l]]
          ix,iy,rx,ry,fx,fy = asarray([ix,iy,rx,ry,fx,fy])
          mx,dx = bwt(fx-rx)
          my,dy = bwt(fy-ry)
          if verb: print mx,dx,mx,dy,len(fx)
          sel = between(fx-rx,mx-4*dx,mx+4*dx)*between(fy-ry,my-4*dy,my+4*dy)
          ix = compress(sel,ix,0)
          iy = compress(sel,iy,0)
          rx = compress(sel,rx,0)
          ry = compress(sel,ry,0)
          b = abasis(nord,ix,iy,rot=1)
          y = concatenate([rx,ry])
          tran = svdfit(b,y)
#          tran = planefit(b,y,tran,minfunc=mad)[0]
          tran = tran/fac
          f = add.reduce(tran[NewAxis,::]*b,1)*fac
      fx = f[:len(ix)]; fy = f[len(ix):]
      mx,dx = bwt(fx-rx)
      my,dy = bwt(fy-ry)
      print "Sigmas (x,y)=(%.4f,%.4f) [%d]" % (dx,dy,len(fx))
      return tran

def coadd(images,output,gain=3.0,usemask=None,sample=2,quick=0,bounds=None,minimages=None):

       addim = output
       sigim = replace(output,".fits","sig.fits")
       print "Making:",addim,sigim
       if not minimages: minimages = len(images)

       f0 = FITS(images[0],readonly=1)
       f0sh = f0.naxes()
       f0sh.reverse()
       f0.close()
       oy,ox = f0sh
       nx = sample * ox
       ny = sample * oy

       if bounds == None: bounds = [[0,nx],[0,ny]]
       else: bounds = multiply(bounds,sample)
       cnx = bounds[0][1] - bounds[0][0]
       cny = bounds[1][1] - bounds[1][0]

       f=FITS(addim,create=1,template=images[0])
       f.modkey("exptime",1.0)
       f.resize([cnx,cny])
       f.close()
       f=FITS(sigim,create=1,template=images[0])
       f.modkey("exptime",1.0)
       f.resize([cnx,cny])
       f.close()
       badval = 1e8

       if usemask:
           print "Reading mask:",usemask
           fmask = FITS(usemask,readonly=1)
           dmask = logical_not(fmask.data())
           fmask.close()
       fims = [FITS(n,readonly=1) for n in images]
       nf = len(images)
       rnf = range(nf)
       times = array([float(f["exptime"]) for f in fims])
       backs = array([float(f["backgrnd"]) for f in fims])
       xtrans = sample*array([float(f["xtrans"]) for f in fims])
       ytrans = sample*array([float(f["ytrans"]) for f in fims])
       minx = min(xtrans)
       miny = min(ytrans)
       xtrans = xtrans - minx
       ytrans = ytrans - miny
       xtran2 = (xtrans).astype(Int32) + int(minx)
       ytran2 = (ytrans).astype(Int32) + int(miny)
       print "dX=",xtran2
       print "dY=",ytran2
       ay2 = arange(sample*oy)

       ry = range(bounds[1][0],bounds[1][1])

       t0 = -999; jt0 = 0
       fadd=FITS(addim,readonly=0)
       fsig=FITS(sigim,readonly=0)
       for jrow in ry:
           ja = oy/2.0 + (jrow - (ny/2.0))/(1.*sample)
           ja2 = sample*ja
           uj2 = (ja2 + ytran2)/(1.*sample)
           nj2 = uj2 + not_equal(uj2,uj2.astype(Int32))/(1.*sample)
           nj2 = nj2.astype(Int32)
           znf = between(nj2,0,oy-1,inc=1)
           oydat = zeros((cnx,),Float32)
           oysig = zeros((cnx,),Float32)
           if sum(znf) >= minimages:
               if t0 < -100: t0 = time.time(); jt0 = jrow
               unf = nonzero(znf)
               ydat = asarray([repeat(fims[l].row(nj2[l])/times[l],sample*ones((ox,))) for l in unf])
               vdat = asarray([(backs[l]/gain/times[l])*ones((sample*ox,)) for l in unf])
               ydat = ydat / pow(sample,2); vdat = vdat / pow(sample,2)
               if usemask:
                  mdat = asarray([repeat(dmask[nj2[l]],sample*ones((ox,))) for l in unf])
                  vdat = (vdat*logical_not(mdat)).astype(Float32)
                  ydat = (ydat + 1e24*mdat).astype(Float32)
               xdat = asarray([xtran2[l] for l in unf])
               lnf = len(unf)
               bnf = range(lnf)
               bdat = zeros((lnf,sample*ox),Float32)
               cdat = zeros((lnf,sample*ox),Float32)
               bdat[::,nx/2-sample*ox/2:nx/2+sample*ox/2] = ydat.astype(Float32)
               cdat[::,nx/2-sample*ox/2:nx/2+sample*ox/2] = vdat.astype(Float32)
               ydat = asarray([shiftim(bdat[l],xdat[l],0) for l in bnf])[::,bounds[0][0]:bounds[0][1]]
               vdat = asarray([shiftim(cdat[l],xdat[l],0) for l in bnf])[::,bounds[0][0]:bounds[0][1]]
               ydat = (ydat + 1e24*logical_not(vdat)).astype(Float32)
               edat = greater(vdat,0)*less(ydat,1e18)
               if sometrue(edat):
                  ndat = sum(edat)
                  mn = max(ndat)
                  udat = greater(ndat,0)
                  xc = nonzero(udat)
                  xc0 = xc[0]; xc1 = xc[-1]+1
                  qd = xc1-xc0
                  rqx = range(qd)
                  ydat = ydat[::,xc0:xc1]
                  vdat = vdat[::,xc0:xc1]
                  ndat = ndat[xc0:xc1]
                  aydat = [argsort(ydat[::,qx]) for qx in rqx]
                  ydat = asarray([take(ydat[::,qx],aydat[qx]) for qx in rqx])
                  vdat = asarray([take(vdat[::,qx],aydat[qx]) for qx in rqx])
                  if quick:
                    soydat = asarray([mean(ydat[qx,:ndat[qx]]) for qx in rqx])
#                    soysig = asarray([ndat[qx] for qx in rqx])
                    soysig = asarray([divz(rms(ydat[qx,:ndat[qx]]),sqrt(ndat[qx])) for qx in rqx])
                  else:
                    ymed = asarray([median(ydat[qx,:ndat[qx]],sorted=0) for qx in rqx])
                    ymad = 1.49*asarray([mad(ydat[qx,:ndat[qx]],ymed[qx]) for qx in rqx])
                    vq = less_equal(abs(ydat-ymed[::,NewAxis]),4*ymad[::,NewAxis])
                    vdat = vdat * vq
                    uvq = greater(vdat,0)
                    nvq = add.reduce(uvq,1)
                    gvq = greater(nvq,3)
                    if sometrue(gvq):
                       wt = where(vdat,1.0/(vdat+equal(vdat,0)),0.0)
                       numer = add.reduce(ydat*wt,1)
                       denom = add.reduce(wt,1)
                       soydat = where(denom,numer/(denom+equal(denom,0)),1e9) * gvq
                       numer = add.reduce(power(ydat-soydat[::,NewAxis],2)*wt,1)
                       soyvar = where(denom,numer/(denom+equal(denom,0)),1e9) * gvq
                       ovq = where(nvq-1,1.0/(nvq-1),1)
                       soysig = sqrt(soyvar*ovq)
                    else:
                       soydat = ymed
                       soysig = nvq
                  oydat[xc0:xc1] = soydat.astype(Float32)
                  oysig[xc0:xc1] = soysig.astype(Float32)
                  t1 = time.time()
                  fadd.writerow(oydat,jrow-ry[0])
                  fsig.writerow(oysig,jrow-ry[0])
                  dtime = (t1-t0)
                  print jrow,sum(oydat),xc0,xc1,len(unf),dtime,dtime/(jrow-jt0+1),(ry[-1]-jrow)*dtime/(jrow-jt0+1)
                  stdout.flush()
       fadd.close()
       fsig.close()
       [f.close() for f in fims]


