
import sys
import time 
from MLab import *
from VTKHelperFunctions import *
from Multipack import *
from types import *
from basics import *
from Numeric import *
from LinearAlgebra import *
from pgplot import *
from umath import *
import math
import copy
from string import *
from basics import readgrid
from ephem import *
import os

def getstat(f,x,default):
   if len(x): return f(x)
   else: return default

#def degrad(x): return (pi/180)*x
#def hrrad(x): return (15*pi/180)*x
#def raddeg(x): return (180/pi)*x
#def radhr(x): return (180/pi/15)*x
#def formatSexagesimal(x,f=4,d=360000):
#    a = int(x)
#    x = 60*(x-a)
#    b = int(x)
#    x = 60*(x-b)
#    c = int(x)
#    d = x-c
#    fmt = "%"+str(f)+"d:%2s:%2s"
#    s = fmt % (a,zfill(b,2),zfill(c,2))
#    p = "%4.4f" % (d)
#    s = s + p[1:]
#    return s

class slitmask:
    def __init__(self,instrument,box=None,verb=0,wran=None):
        self.instrument=instrument
        self.wran=wran
        if verb: print self.instrument

        if self.instrument=="lris":
           #TAKEN FROM THE UCSCLRIS PACKAGE DATED 97 DEC 09
           self.LAT=19.8265		# Keck latitude
           self.M_BEND=1.94		# Mask bend in degrees
           self.M_ANGLE=8.06		# Mask angle (tilt) in degrees
           #MASK_X0=107.95		# Half of Mask X size (215.9) (mm)
           self.MASK_Y0=167.8		# Half of Mask Y size (335.6) (mm)
           self.MASK_X0=132.3		# Half of Mask X size (215.9) (mm)
           #MASK_Y0=177.8		# Half of Mask Y size (335.6) (mm)
           self.MM_ARCS=0.7253		# mm per arcsec
           self.MM_PIX=0.15767		# mm per pixel
           self.FP_XCEN=305.		# Mask x-center (mm) in FP coord
           self.FP_YCEN=0.		# Mask y-center (mm) in FP coord
           self.CCDXCEN=1024.5		# CCD x-center
           self.CCDYCEN=1024.5		# CCD y-center
           self.XPIXOFF=0.		# x-offset(pix) (not in AUTOSLIT)
           self.YPIXOFF=80. 		# y-offset(pix)
           self.MIN_MY=3.5		# Minimum mask y-value
           self.MAX_MY=352.1		# Maximum mask y-value
           self.MAX_MY=332.		# Maximum mask y-value
           self.MIN_MX=3.5		# Minimum mask y-value
           self.MAX_MX=261.1		# Maximum mask y-value
           #MAX_MX=212.4		# Maximum mask y-value
           self.XFIELD = self.MAX_MX - self.MIN_MX
           self.YFIELD = self.MAX_MY - self.MIN_MY
           self.YSCLCOR=1.0175		# yscale_cor
           self.SCL_FIX=1.0029		# Scaling correction
           self.ACORR1=-2.95		# Astrom. correction term 1
           self.ACORR2=7.734e-3		# Astrom. correction term 2
           self.ACORR3=-3.212e-6	# Astrom. correction term 3
           self.ACORR4=-3.274e-10	# Astrom. correction term 4
           self.TV_XOFF=231.8 		# edge of TV CCD (mm, FPCS)
           self.TV_YOFF=31.5		# edge of TV CCD (mm, FPCS)
           self.TV_MMPIX=0.1835		# pixel scale: mm (FP) / pix (TV)
           self.SZ_INLN=128		# Maximum length of input line
           self.SZ_ID=32		# Maximum length of ID string
           self.CODE_GS=-1		# Priority code for guide star		XXX
           self.CODE_AS=-2		# Priority code for alignment star	XXX
           self.SLIT_SPACE=0.35*self.MM_ARCS
           self.GUID_HOL=0.5
           self.STARBOX=4.0
           self.PRESSURE = 615
           self.TEMPERATURE = 0.0
           self.SPEC_OFF = 0.0

        if self.instrument=="lrisb":
           #TAKEN FROM THE UCSCLRIS PACKAGE DATED 97 DEC 09
           self.LAT=19.8265		# Keck latitude
           self.M_BEND=1.94		# Mask bend in degrees
           self.M_ANGLE=8.06		# Mask angle (tilt) in degrees
           #MASK_X0=107.95		# Half of Mask X size (215.9) (mm)
           self.MASK_Y0=167.8		# Half of Mask Y size (335.6) (mm)
           self.MASK_X0=132.3		# Half of Mask X size (215.9) (mm)
           #MASK_Y0=177.8		# Half of Mask Y size (335.6) (mm)
           self.MM_ARCS=0.7253		# mm per arcsec
           self.MM_PIX=0.0979155		# mm per pixel
           self.FP_XCEN=305.		# Mask x-center (mm) in FP coord
           self.FP_YCEN=0.		# Mask y-center (mm) in FP coord
           self.CCDXCEN=2048		# CCD x-center
           self.CCDYCEN=2048		# CCD y-center
           self.XPIXOFF=0.		# x-offset(pix) (not in AUTOSLIT)
           self.YPIXOFF=80. 		# y-offset(pix)
           self.MIN_MY=3.5		# Minimum mask y-value
           self.MAX_MY=352.1		# Maximum mask y-value
           self.MAX_MY=332.		# Maximum mask y-value
           self.MIN_MX=3.5		# Minimum mask y-value
           self.MAX_MX=261.1		# Maximum mask y-value
           #MAX_MX=212.4		# Maximum mask y-value
           self.XFIELD = self.MAX_MX - self.MIN_MX
           self.YFIELD = self.MAX_MY - self.MIN_MY
           self.YSCLCOR=1.0175		# yscale_cor
           self.SCL_FIX=1.0029		# Scaling correction
           self.ACORR1=-2.95		# Astrom. correction term 1
           self.ACORR2=7.734e-3		# Astrom. correction term 2
           self.ACORR3=-3.212e-6	# Astrom. correction term 3
           self.ACORR4=-3.274e-10	# Astrom. correction term 4
           self.TV_XOFF=231.8 		# edge of TV CCD (mm, FPCS)
           self.TV_YOFF=31.5		# edge of TV CCD (mm, FPCS)
           self.TV_MMPIX=0.1835		# pixel scale: mm (FP) / pix (TV)
           self.SZ_INLN=128		# Maximum length of input line
           self.SZ_ID=32		# Maximum length of ID string
           self.CODE_GS=-1		# Priority code for guide star		XXX
           self.CODE_AS=-2		# Priority code for alignment star	XXX
           self.SLIT_SPACE=0.35*self.MM_ARCS
           self.GUID_HOL=0.5
           self.STARBOX=4.0
           self.PRESSURE = 615
           self.TEMPERATURE = 0.0
           self.SPEC_OFF = 0.0

        if self.instrument=="deimos":
           #TAKEN FROM DREW PHILLIPS WEB INFO 20 DEC 2002
           self.LAT=19.8265		# Keck latitude
           self.MM_ARCS=0.7253		# mm per arcsec
           self.ARC_PIX=0.119
           self.MM_PIX = self.MM_ARCS * self.ARC_PIX
           self.YFIELD = (502.0 * 2) * self.MM_ARCS
           self.XFIELD = (479.6 - 186.8) * self.MM_ARCS
           self.MASK_X0 = 333.2 * self.MM_ARCS           # Mask x-center (mm) in FP coord
           self.MASK_Y0 = 0.0		# Mask y-center (mm) in FP coord
           self.XOFFSET = 85.0 * self.MM_ARCS
           self.YOFFSET = 0.0 * self.MM_ARCS
           self.MIN_MY = -502.0 * self.MM_ARCS
           self.MAX_MY =  502.0 * self.MM_ARCS
           self.MIN_MX =  186.8 * self.MM_ARCS
           self.MAX_MX =  479.6 * self.MM_ARCS

           self.CODE_GS=-1		# Priority code for guide star		XXX
           self.CODE_AS=-2		# Priority code for alignment star	XXX
           self.SLIT_SPACE=0.35*self.MM_ARCS
           self.GUID_HOL=0.5
           self.STARBOX=4.0
           self.PRESSURE = 615
           self.TEMPERATURE = 0.0
           self.PIXOFF = -150
           self.SPEC_OFF = 0.0

        if self.instrument=="ldss2":
           self.LAT=scanSexagesimal("-29:00:13")
           self.ARC_PIX=0.3767
           self.MM_ARCS=1/2.880
           self.MM_PIX=self.ARC_PIX * self.MM_ARCS
           self.XFIELD = 290 * self.MM_ARCS
           self.YFIELD = 400 * self.MM_ARCS
#           self.MASK_Y0=110/2.0
#           self.MASK_X0=110/2.0
           self.MASK_Y0=self.YFIELD/2.0
           self.MASK_X0=self.XFIELD/2.0
           self.MIN_MY=self.MASK_Y0 - self.YFIELD/2.0 + 5 * self.MM_ARCS
           self.MAX_MY=self.MASK_Y0 + self.YFIELD/2.0 - 5 * self.MM_ARCS
           self.MIN_MX=self.MASK_X0 - self.XFIELD/2.0 + 5 * self.MM_ARCS
           self.MAX_MX=self.MASK_X0 + self.XFIELD/2.0 - 5 * self.MM_ARCS
           self.CODE_GS=-1		# Priority code for guide star		XXX
           self.CODE_AS=-2		# Priority code for alignment star	XXX
#           self.SLIT_SPACE=0.5*self.MM_ARCS
           self.SLIT_SPACE=2.0*self.MM_PIX
           self.GUID_HOL=0
           self.STARBOX=5.7
           self.PRESSURE = 655
           self.TEMPERATURE = 10.0
           self.SPEC_OFF = 0.0

        if self.instrument=="ldss3":
           self.LAT=scanSexagesimal("-29:00:13")
           self.ARC_PIX=0.189
           self.MM_ARCS=1/2.880
           self.MM_PIX=self.ARC_PIX * self.MM_ARCS
           self.XFIELD0 = 498 * self.MM_ARCS
           self.YFIELD0 = 498 * self.MM_ARCS
           self.XFIELD = 498 * self.MM_ARCS
           self.YFIELD = 498 * self.MM_ARCS
           self.MASK_Y0=self.YFIELD/2.0
           self.MASK_X0=self.XFIELD/2.0
           self.MIN_MY=self.MASK_Y0 - self.YFIELD/2.0 + 3 * self.MM_ARCS
           self.MAX_MY=self.MASK_Y0 + self.YFIELD/2.0 - 3 * self.MM_ARCS
           self.MIN_MX=self.MASK_X0 - 1.5*self.XFIELD/2.0 + 1 * self.MM_ARCS
           self.MAX_MX=self.MASK_X0 + 1.5*self.XFIELD/2.0 - 1 * self.MM_ARCS
           self.CODE_GS=-1		# Priority code for guide star		XXX
           self.CODE_AS=-2		# Priority code for alignment star	XXX
           self.SLIT_SPACE=0.5*self.MM_ARCS
#           self.SLIT_SPACE=2.0*self.MM_PIX
           self.GUID_HOL=0
           self.STARBOX=5.0
           self.PRESSURE = 655
           self.TEMPERATURE = 10.0
           self.SPEC_OFF = 0.0

        if self.instrument=="imacs2":
           self.LAT=scanSexagesimal("-29:00:13")
           self.ARC_PIX=0.20
           self.MM_ARCS=1/2.880
           self.MM_PIX=self.ARC_PIX * self.MM_ARCS
           self.XFIELD0 = 24*60 * self.MM_ARCS
           self.YFIELD0 = 24*60 * self.MM_ARCS
           self.XFIELD = 30*60 * self.MM_ARCS
           self.YFIELD = 27*60 * self.MM_ARCS
           self.MASK_Y0 = self.YFIELD/2.0
           self.MASK_X0 = self.XFIELD/2.0
           self.MIN_MY=0.0
           self.MAX_MY=self.YFIELD - 0.0
           self.MIN_MX=72.0
           self.MAX_MX=self.XFIELD - 5.0
#           self.MIN_MX=self.MASK_X0 - self.XFIELD/2.0 + 15 * self.MM_ARCS
#           self.MAX_MX=self.MASK_X0 + self.XFIELD/2.0 - 15 * self.MM_ARCS
           self.CODE_GS=-1		# Priority code for guide star		XXX
           self.CODE_AS=-2		# Priority code for alignment star	XXX
           self.SLIT_SPACE=5.0*self.MM_PIX
           self.GUID_HOL=0
           self.STARBOX=5.0
           self.PRESSURE = 655
           self.TEMPERATURE = 10.0
           self.SPEC_OFF = -55.0

        if self.instrument=="imacs4":
           self.LAT=scanSexagesimal("-29:00:13")
           self.ARC_PIX=0.11
           self.MM_ARCS=1/2.880
           self.MM_PIX=self.ARC_PIX * self.MM_ARCS
           self.XFIELD = 924 * self.MM_ARCS
           self.YFIELD = 924 * self.MM_ARCS
           self.MASK_Y0=self.YFIELD/2.0
           self.MASK_X0=self.XFIELD/2.0
           self.MIN_MY=self.MASK_Y0 - self.YFIELD/2.0 + 5 * self.MM_ARCS
           self.MAX_MY=self.MASK_Y0 + self.YFIELD/2.0 - 5 * self.MM_ARCS
           self.MIN_MX=self.MASK_X0 - self.XFIELD/2.0 + 5 * self.MM_ARCS
           self.MAX_MX=self.MASK_X0 + self.XFIELD/2.0 - 5 * self.MM_ARCS
           self.CODE_GS=-1		# Priority code for guide star		XXX
           self.CODE_AS=-2		# Priority code for alignment star	XXX
           self.SLIT_SPACE=5.0*self.MM_PIX
           self.GUID_HOL=0
           self.STARBOX=5.0
           self.PRESSURE = 655
           self.TEMPERATURE = 10.0
           self.SPEC_OFF = 0.0

        if self.instrument=="vimos":
           self.ARC_PIX = 0.205
           self.MM_ARCS = 1/1.718
           self.MM_PIX = self.MM_ARCS * self.ARC_PIX
           self.XFIELD = 250.0
           self.YFIELD =  250.0
           self.MIN_MX = -125.0
           self.MAX_MX =  125.0
           self.MIN_MY = -125.0
           self.MAX_MY =  125.0
           self.STARBOX = 12.0

        if self.instrument=="wfccd":
           self.LAT=scanSexagesimal("-29:00:13")
           self.ARC_PIX = 0.77402
           self.MM_ARCS = 1/10.91585
           self.MM_PIX = self.ARC_PIX * self.MM_ARCS
           self.XFIELD = 1080 * self.MM_ARCS
           self.YFIELD = 1500 * self.MM_ARCS
           self.MASK_Y0 = self.YFIELD/2.0
           self.MASK_X0 = self.XFIELD/2.0
           self.MIN_MY = self.MASK_Y0 - self.YFIELD/2.0 + 5 * self.MM_ARCS
           self.MAX_MY = self.MASK_Y0 + self.YFIELD/2.0 - 5 * self.MM_ARCS
           self.MIN_MX = self.MASK_X0 - self.XFIELD/2.0 + 5 * self.MM_ARCS
           self.MAX_MX = self.MASK_X0 + self.XFIELD/2.0 - 5 * self.MM_ARCS
           self.CODE_GS = -1		# Priority code for guide star		XXX
           self.CODE_AS = -2		# Priority code for alignment star	XXX
           self.SLIT_SPACE = 2.0*self.MM_PIX
           self.GUID_HOL = 0
           self.STARBOX = 12.0
           self.PRESSURE = 655
           self.TEMPERATURE = 10.0
           self.SPEC_OFF = 0.0

        if box:
           if verb: print "Changing Star Box Width from",2.*self.STARBOX,"to",box
           self.STARBOX=box

        self.STARBOX=self.STARBOX/2.0

    def gmcode(self,root,tool,xs,xe,ys,ye,sw):
       import gm
       xs,xe,ys,ye,sw = asarray([ye,ys,xs,xe,sw])
       if tool > 0.011: o1,o2 = root+".lds","crapola"
       else: o1,o2 = root+".ld1", root+".ld2"
       gm.cgmcode(len(xs),xs,ys,xe,ye,sw,tool,self.YFIELD,root,o1,o2)

    def maskrefract(self,ra,dec,ha,wave):
       latd = self.LAT
       pres = self.PRESSURE
       temp = self.TEMPERATURE
       if wave: r,d = refract(ra,dec,ha,latd,pres,temp,wave)
       else: r,d = ra,dec
       return r,d

    def lris_coord(self,xfp, yfp, xref, yref):
        xx =  (xfp - self.FP_XCEN) / cos (degrad(self.M_ANGLE)) + self.FP_XCEN
        yy =  yfp / cos (degrad(self.M_BEND))	# Crude
        x = self.MASK_X0 + (self.FP_XCEN - xx) / self.SCL_FIX		# These in mm, and ...
        yccd = self.CCDYCEN + (self.FP_YCEN + yy) / self.MM_PIX	# ... these in pix
        x = x + (self.ACORR1 + yccd*(self.ACORR2 + yccd*(self.ACORR3 + yccd*self.ACORR4))) * self.MM_PIX
        y = self.MASK_Y0 - yy - self.FP_YCEN
        x = x - xref
        y = y - yref
        return x, y
    def maskpos(self,u,v):
      if self.instrument == "lris":
        xfp, yfp = u*self.MM_ARCS, v*self.MM_ARCS
        x0,y0=self.lris_coord(xfp[0],yfp[0],0.0,0.0)
        delx,dely=self.lris_coord(xfp,yfp,x0,y0)
        x=delx + self.MASK_X0
        y=dely + self.MASK_Y0
      elif self.instrument == "imacs4":
        delx, dely = -u * self.MM_ARCS, v * self.MM_ARCS
        x=delx + self.MASK_X0
        y=dely + self.MASK_Y0
      elif self.instrument == "imacs2":
        delx, dely = -u * self.MM_ARCS, v * self.MM_ARCS
        x=delx + self.MASK_X0
        y=dely + self.MASK_Y0
      elif self.instrument == "ldss2":
        delx, dely = -u * self.MM_ARCS, v * self.MM_ARCS
        x=delx + self.MASK_X0
        y=dely + self.MASK_Y0
      elif self.instrument == "ldss3":
        delx, dely = -u * self.MM_ARCS, v * self.MM_ARCS
        x=delx + self.MASK_X0
        y=dely + self.MASK_Y0
      elif self.instrument == "wfccd":
        delx, dely = -u * self.MM_ARCS, v * self.MM_ARCS
        x=delx + self.MASK_X0
        y=dely + self.MASK_Y0
      elif self.instrument == "deimos":
        delx, dely = u * self.MM_ARCS, v * self.MM_ARCS
        x=delx + self.MASK_X0
        y=dely + self.MASK_Y0
      return x,y

    def mapmask(self,id,ha,ra,dec,prior,mag,slitpa,bot,top,verb=0,wave=0.8):
        cent = id.index("CENTER")
        maskpa = slitpa[0]
        slitpa = subtract(slitpa,maskpa)
        if self.instrument == "deimos":
            xoff,yoff = self.XOFFSET*cos(maskpa) - self.YOFFSET*sin(maskpa), \
                    self.XOFFSET*sin(maskpa) + self.YOFFSET*cos(maskpa)
            xoff = -degrad(xoff/3600)
            yoff = degrad(yoff/3600)
            ra[cent],dec[cent] = equicoord(xoff,yoff,ra[cent],dec[cent])

        raf,decf = self.maskrefract(ra,dec,ha,wave)
        u,v = normalcoord(raf,decf,raf[cent],decf[cent],conv=3600*180/pi)
        cosa=cos(maskpa); sina=sin(maskpa)
        u,v = u*cosa - v*sina, u*sina + v*cosa
        xc,yc = self.maskpos( u[0],v[0])
        x,y = self.maskpos( u,v)
        if self.instrument in ["imacs4","imacs2","ldss2","ldss3","wfccd","deimos"]:
           y = 2*self.MASK_Y0 - y

	guid=nonzero(equal(prior,self.CODE_GS))
        if len(guid):
            i = guid[0]
            mirrid = id[i]; mx=x[i]; my=y[i]; mmag=mag[i]
            mra=ra[i]; mdec=dec[i]
        else:
            mirrid = None; mx=0.0; my=0.0; mmag=0.0
            mra=0.0; mdec=0.0

	yargs = argsort(y)
        keep = not_equal(yargs,cent)
        yargs = compress(keep,yargs)
        id = [id[j] for j in yargs]
        x,y,ra,dec,raf,decf,prior,mag,slitpa,bot,top = \
            take([x,y,ra,dec,raf,decf,prior,mag,slitpa,bot,top],yargs,1)

	yargs = argsort(y)
        keep = not_equal(prior,self.CODE_GS)
        yargs = compress(keep,yargs)
        id = [id[j] for j in yargs]
        x,y,ra,dec,raf,decf,prior,mag,slitpa,bot,top = \
            take([x,y,ra,dec,raf,decf,prior,mag,slitpa,bot,top],yargs,1)

	alig=equal(prior,self.CODE_AS)

        if not self.wran:
            bot = where(greater(bot,-998),bot,-999/self.MM_ARCS)
            top = where(greater(top,-998),top,-999/self.MM_ARCS)
            bot = where(alig,self.STARBOX,bot)*self.MM_ARCS
            top = where(alig,self.STARBOX,top)*self.MM_ARCS
            spy = (diff(y) - self.SLIT_SPACE)/2.0
            spy1 = concatenate([spy,[0]])
            spy2 = concatenate([[0],spy])
            bot = where(greater(bot,-998),bot,spy1)
            top = where(greater(top,-998),top,spy2)
        else:
            bot = where(greater(bot,-998),bot,-999/self.MM_ARCS)
            top = where(greater(top,-998),top,-999/self.MM_ARCS)
            bot = where(alig,self.STARBOX,bot)*self.MM_ARCS
            top = where(alig,self.STARBOX,top)*self.MM_ARCS
            bots = [[] for j in range(len(x))]
            tops = [[] for j in range(len(x))]
            wranm = self.wran * self.MM_PIX
            x0 = x + wranm/2.0
            x1 = x - wranm/2.0
            for j in range(len(x)):
              xbad = logical_and(less_equal(x1,x0[j]),greater_equal(x0,x1[j]))
              nx = nonzero(xbad)
              if alig[j]:
                 bots[j].append(bot[j])
                 tops[j].append(top[j])
              if len(nx)>1:
                 jx,jy,ja,jb,jt = take([x,y,alig,bot,top],nx,1)
                 jn = searchsorted(jy,y[j])
                 spy = diff(jy)/2.0 - self.SLIT_SPACE/2.0
                 spy1 = concatenate([spy,[0]])
                 spy2 = concatenate([[0],spy])
                 spy1 = where(jb<-900,where(ja,jb,spy1),jb)
                 spy2 = where(jt<-900,where(ja,jt,spy2),jt)
                 bots[j].append(spy1[jn])
                 tops[j].append(spy2[jn])
            bot = array([getstat(min,bots[j],0) for j in range(len(x))])
            top = array([getstat(min,tops[j],0) for j in range(len(x))])
            if verb==2:
               for j in range(len(y)):
                 print "%10s %9.2f %9.2f %9.2f %9.2f" % (id[j],x1[j],x[j],x0[j],y[j])

        if not alig[0]: top[0] = 5*bot[0]
        if not alig[-1]: bot[-1] = 5*top[-1]

        y0 = y - top
        y1 = y + bot

        prev = y1[:-1]
        next = y0[1:]

        if not self.wran:
           fix = (next-prev)/2.0 - self.SLIT_SPACE
           y0[1:] = next - fix*logical_not(alig[1:])
           y1[:-1] = prev + fix*logical_not(alig[:-1])

        if self.instrument in ["imacs2","ldss3"]:
           espace = abs(self.MAX_MY-(self.MASK_Y0+self.YFIELD/2.0))
           if prior[0] > 0: y0[0] = max([self.MIN_MY+espace+(self.MASK_Y0-sqrt((self.YFIELD/2.0)**2-(x[0]-self.MASK_X0)**2)),y0[0]])
           if prior[-1] > 0: y1[-1] = min([self.MAX_MY-espace-(self.MASK_Y0-sqrt((self.YFIELD/2.0)**2-(x[-1]-self.MASK_X0)**2)),y1[-1]])
        else:
           if prior[0] > 0: y0[0] = max([self.MIN_MY,y0[0]])
           if prior[-1] > 0: y1[-1] = min([self.MAX_MY,y1[-1]])

        if verb: 
           print "Number of non-alignment objects=",len(compress(greater(prior,0),y1-y0))
           print "Sum of priorities=",sum(compress(greater(prior,0),prior))
           print "Slit-length extrema=",
           print extrema(compress(greater(prior,0),y1-y0))/self.MM_ARCS
           print sort(compress(greater(prior,0),y1-y0))[:5]/self.MM_ARCS
           print sort(compress(greater(prior,0),y1-y0))[-5:]/self.MM_ARCS
           print "Slit-length quartiles=",
           print quartiles(compress(greater(prior,0),y1-y0))/self.MM_ARCS
           print "Slit-length mean=",
           print mean(compress(greater(prior,0),y1-y0))/self.MM_ARCS
           print


        xuse = x
        yuse = y
        pas = slitpa
        ruse = ra
        duse = dec
        puse = prior
        muse = mag

        if mirrid:
           xuse = concatenate([xuse,[mx]])
           yuse = concatenate([yuse,[my]])
           ruse = concatenate([ruse,[mra]])
           duse = concatenate([duse,[mdec]])
           puse = concatenate([puse,[self.CODE_GS]])
           muse = concatenate([muse,[mmag]])
           pas = concatenate([pas,[0.0]])
           y1 = concatenate([y1,[my+0.5*self.MM_ARCS]])
           y0 = concatenate([y0,[my-0.5*self.MM_ARCS]])

	dmy=self.GUID_HOL

	maskpar=[xuse,yuse,y1,y0,pas,ruse,duse,puse,id,\
		muse,mx,my,dmy,mirrid,mmag,mra,mdec]

	return maskpar

    def markmask(self,i1,i2,j1,j2,soln,swid,maskpa,maskpar,rac,decc,rotate=0,color=0):

        if rac > 0.0 and decc > -900:
            puse=maskpar[7]
            wid=swid*greater(puse,0) + self.STARBOX*less(puse,0)
            boxcolor = color
            colors=boxcolor*equal(puse,self.CODE_AS) + color*greater(puse,0)
            id=maskpar[8]
            ruse=maskpar[5]; duse=maskpar[6]
            xposu, yposu, scale, ipa = rd2xy(soln,ruse,duse)

            dpa2 = (-maskpa-ipa)
            dpa = (dpa2 - maskpar[4])

            x1 = (maskpar[1]-maskpar[3])/self.MM_ARCS/scale
            xpos1 = x1*sin(dpa) + xposu
            ypos1 = x1*cos(dpa) + yposu
            xpos1b = -wid/scale*cos(dpa2) + xpos1
            ypos1b = wid/scale*sin(dpa2) + ypos1
            xpos1c = wid/scale*cos(dpa2) + xpos1
            ypos1c = -wid/scale*sin(dpa2) + ypos1
            x0 = (maskpar[1]-maskpar[2])/self.MM_ARCS/scale
            xpos0 = x0*sin(dpa) + xposu
            ypos0 = x0*cos(dpa) + yposu
            xpos0b = -wid/scale*cos(dpa2) + xpos0
            ypos0b = wid/scale*sin(dpa2) + ypos0
            xpos0c = wid/scale*cos(dpa2) + xpos0
            ypos0c = -wid/scale*sin(dpa2) + ypos0
            for i in range(0,len(xposu)):
              x=array([xpos1b[i],xpos1c[i],xpos0c[i],xpos0b[i],xpos1b[i]],Float32)
              y=array([ypos1b[i],ypos1c[i],ypos0c[i],ypos0b[i],ypos1b[i]],Float32)
              if rotate: drawslit(y,x,colors[i])
              else: drawslit(x,y,colors[i])
            mra=array([maskpar[15]],Float32)
            mdec=array([maskpar[16]],Float32)
            xm, ym, scale, ipa = rd2xy(soln,mra,mdec)

            cra = array([rac],Float32)
            cdec = array([decc],Float32)
            xc, yc, scale, ipa = rd2xy(soln,cra,cdec)
            xn0=xc; yn0=yc
            if rotate: drawpt( yc,xc,3,1,2)
            else: drawpt( xc,yc,3,1,2)

            if self.instrument in ["imacs4","ldss2","wfccd","deimos"]:
              hh = 0.5*self.YFIELD/self.MM_ARCS/scale
              ww = 0.5*self.XFIELD/self.MM_ARCS/scale
              xn,yn=makebox( hh, ww, xc, yc, dpa2, color, rotate)
            elif self.instrument in ["imacs2","ldss3"]:
              hh = 0.5*self.YFIELD/self.MM_ARCS/scale
              ww = 0.5*self.XFIELD/self.MM_ARCS/scale
              cpgsfs(2)
              cpgsci(color)
              cpgcirc(xc[0],yc[0],hh)
              cpgsci(1)
            elif self.instrument == "lris":
              hh = self.MASK_Y0/self.MM_ARCS/scale
              ww = self.MASK_X0/self.MM_ARCS/scale
              xn,yn=makebox( hh, ww, xc, yc, dpa2, color, rotate)

              hh = 2.5/self.MM_ARCS/scale
              ww = self.MASK_X0/self.MM_ARCS/scale
              xn,yn=makebox( hh, ww, xc, yc, dpa2, color, rotate)

              mwid = 22.0
              mir_in_x=167.3
              mir_up_y=186.8 - 10

              dxm = (mir_in_x-self.MASK_X0+mwid/2)/self.MM_ARCS/scale
              dym = -(mir_up_y-self.MASK_Y0+mwid/2)/self.MM_ARCS/scale
              xc = array([xn0[0] + dxm*cos(dpa2)+dym*sin(dpa2)],Float32)
              yc = array([yn0[0] - dxm*sin(dpa2)+dym*cos(dpa2)],Float32)
              hh = mwid/self.MM_ARCS/scale/2.0
              ww = mwid/self.MM_ARCS/scale/2.0
              xn,yn=makebox( hh, ww, xc, yc, dpa2, color, rotate)

def readsim(input):
	t0=os.times()[0]
        if type(input) == type(sys.stdin):
          data = input.readlines()
        else:
          f=open(input,'r')
          data=[rl for rl in f.readlines() if rl[0] != "#"]
          f.close()

	grid=[]
	for i in range(8):
	  coldat=[]
	  for d in data:
	   v = -999
#           if not i:
#              print len(split(d)),data.index(d),len(data)
#              print d
	   if i < len(split(d)):
	     v = split(d)[i]
	     if i == 1 or i == 2 or i>4:
#		if v.upper() != "INDEF": v=float(v)
#		else: v=-999
		try: v=float(v)
		except: v=-999
	   else: v = -999
	   coldat.append(v)
	  grid.append(coldat)
	t1=os.times()[0]

	return (grid,t1-t0)

def sexlist(l,tran=degrad): return asarray(map(tran,map(scanSexagesimal,l)))

def makebox( hh, ww, xc, yc, dpa2, ci=2, rotate=0, draw=1):

	xm0 = hh*sin(dpa2) + xc
	ym0 = hh*cos(dpa2) + yc
	xm0b = -ww*cos(dpa2) + xm0
	ym0b = ww*sin(dpa2) + ym0
	xm0c = ww*cos(dpa2) + xm0
	ym0c = -ww*sin(dpa2) + ym0

	xm1 = -hh*sin(dpa2) + xc
	ym1 = -hh*cos(dpa2) + yc
	xm1b = -ww*cos(dpa2) + xm1
	ym1b = ww*sin(dpa2) + ym1
	xm1c = ww*cos(dpa2) + xm1
	ym1c = -ww*sin(dpa2) + ym1

	x=array([xm1b[0],xm1c[0],xm0c[0],xm0b[0],xm1b[0]],Float32)
	y=array([ym1b[0],ym1c[0],ym0c[0],ym0b[0],ym1b[0]],Float32)

        if draw:
	    if rotate: drawslit(y,x, ci)
	    else: drawslit(x,y, ci)

	return x,y

def dumpmask(maskpar):

	xuse=maskpar[0]; yuse=maskpar[1]
	y1=maskpar[2]; y0=maskpar[3]
	pas=maskpar[4]
	puse=maskpar[7]
	id=maskpar[8]
	muse=maskpar[9]
	mx=maskpar[10]
	my=maskpar[11]
	dmy=maskpar[12]
	mirrid=maskpar[13]
	mmag=maskpar[14]

	for g in range(0,len(xuse)):
	 print "%7.3f %7.3f %7.3f %7.3f" % (xuse[g], yuse[g], y1[g], y0[g]),
	 print "%6.2f %5d %5.2f" % (raddeg(pas[g]),puse[g],muse[g]),
	 print id[g]

	print "%7.3f %7.3f %7.3f %7.3f" % (mx, my, my+dmy, my-dmy),
	print "%6.2f %5d %5.2f" % (0.0,-1,mmag),
	print mirrid

def ccd_coord_predict(xsky,ysky):
    xarc = xsky * (180/pi) * 3600
    yarc = ysky * (180/pi) * 3600

    A = array([ 1024.0, -0.47113141E+01, 0.28437639E-02,\
    0.39969638E-04, -0.61005225E-04, -0.68275855E-05,\
    -0.50946248E-06,  0.45159715E-07, -0.54639716E-06,\
    0.56327896E-07])

    B = array([ 1024.0, 0.41723373E-02, 0.47246534E+01,\
    -0.57400376E-05, -0.43390770E-05, -0.10129826E-03,\
    0.25250128E-07, 0.54941795E-06, -0.59603120E-07,\
    0.42787942E-06])

    C = ravel([ [A[j],B[j]] for j in range(10)])
    xccd, yccd = mbasis(C,xarc,yarc)
    yccd = yccd + 5.0
    return xccd, yccd


def lris_coord(xfp, yfp, xref, yref):

	xx =  (xfp - FP_XCEN) / cos (degrad(M_ANGLE)) + FP_XCEN
	yy =  yfp / cos (degrad(M_BEND))	# Crude

	x = MASK_X0 + (FP_XCEN - xx) / SCL_FIX		# These in mm, and ...
	yccd = CCDYCEN + (FP_YCEN + yy) / MM_PIX	# ... these in pix

	x = x + (ACORR1 + yccd*(ACORR2 + yccd*(ACORR3 + yccd*ACORR4))) * MM_PIX
	y = MASK_Y0 - yy - FP_YCEN

	x = x - xref
	y = y - yref

	return x, y

def fp_coord(ra,dec,ra0, dec0, cosa, sina, MM_ARCS=0.7253):

	delra = ra - ra0
	sindr = sin (delra)
	cosdr = cos (delra)
	sind = sin (dec)
	cosd = cos (dec)
	sind0 = sin (dec0)
	cosd0 = cos (dec0)
	denom = (sind * sind0 + cosd * cosd0 * cosdr)

	eta = 180.0*(cosd * sindr / denom) * 3600.0/pi
        nu  = 180.0*((sind * cosd0 - cosd * sind0 * cosdr) / denom) * 3600.0/pi

	xx = ( cosa * eta + sina * nu) * MM_ARCS
	yy = (-sina * eta + cosa * nu) * MM_ARCS

	return xx, yy

def getsolncenter(soln):
	f=open(soln,'r')
	s=f.readlines()
	f.close()

	nterms=int(split(s[0])[0])
	xc = float(split(s[0])[2])
	yc = float(split(s[0])[3])
        if find(split(s[1])[0],":")+1:
	   rc = hrrad(scanSexagesimal(split(s[1])[0]))
	   dc = degrad(scanSexagesimal(split(s[1])[1]))
        else:
	   rc = float(split(s[1])[0])
	   dc = float(split(s[1])[1])

	return xc,yc,rc,dc

def readsoln(soln):
	f=open(soln,'r')
	s=f.readlines()
	f.close()

	nterms=int(split(s[0])[0])
	xc = float(split(s[0])[2])
	yc = float(split(s[0])[3])
        if find(split(s[1])[0],":")+1:
	   rc = hrrad(scanSexagesimal(split(s[1])[0]))
	   dc = degrad(scanSexagesimal(split(s[1])[1]))
        else:
	   rc = float(split(s[1])[0])
	   dc = float(split(s[1])[1])

	co=[]
        for n in range(0,nterms): co.append(float(strip(s[n+2])))
	co=array(co)
        return {"center":(xc,yc),"sky":(rc,dc),"c":co}


def xy2rd(soln,x0,y0,text=1):
        x,y = asarray([x0,y0])

	f=open(soln,'r')
	s=f.readlines()
	f.close()

	nterms=int(split(s[0])[0])
	xc = float(split(s[0])[2])
	yc = float(split(s[0])[3])
        if find(split(s[1])[0],":")+1:
	   rc = hrrad(scanSexagesimal(split(s[1])[0]))
	   dc = degrad(scanSexagesimal(split(s[1])[1]))
        else:
	   rc = float(split(s[1])[0])
	   dc = float(split(s[1])[1])

	co=[]
        for n in range(nterms): co.append(float(strip(s[n+2])))
	co=array(co)

        x= x-xc; y= y-yc
        if len(shape(x)) == 0:
          onlyone=1; x = array([x]); y = array([y])
        else: onlyone=0

        u,v = mbasis(co,x,y)
        ra,dec = equicoord(u,v,rc,dc)

        if text:
           if onlyone:
             ra = formatSexagesimal(radhr(ra[0]), 3, 3600000)
             dec = formatSexagesimal(raddeg(dec[0]), 4, 3600000)
           else:
             ra = [formatSexagesimal(radhr(ra1), 3, 3600000) for ra1 in ra]
             dec = [formatSexagesimal(raddeg(dec1), 4, 3600000) for dec1 in dec]
#        if onlyone: ra = ra[0]; dec=dec[0]
	return ra, dec

def abasis(nord,ix,iy,rot=0):
   I=ones(ix.shape)
   Z=zeros(ix.shape)
   ux = ix
   uy = iy
   if nord == 0:
    sb = [ I,  Z,  ux,  uy]
    if rot: eb = [ Z,  I,  uy,  ux]
    else: eb = [ Z,  I, -uy,  ux]
   if nord >= 1:
    sb = [ I,  Z, ux,   Z,   uy,  Z]
    eb = [ Z,  I,  Z,  ux,    Z, uy]
   if nord >= 2:
    ux = ix/1e3
    uy = iy/1e3
    ux2 = power(ux,2)
    uy2 = power(uy,2)
    sb2 = [ ux2, Z, ux*uy, Z, uy2, Z]
    eb2 = [Z, ux2, Z, ux*uy, Z, uy2]
    sb2 = [1e6*sb1 for sb1 in sb2]
    eb2 = [1e6*eb1 for eb1 in eb2]
    sb = sb + sb2
    eb = eb + eb2
   if nord >= 3:
    ux3 = ux2*ux
    uy3 = uy2*uy
    sb3 = [ ux3, Z, ux2*uy, Z, ux*uy2, Z, uy3, Z]
    eb3 = [Z, ux3, Z, ux2*uy, Z, ux*uy2, Z, uy3]
    sb3 = [1e9*sb1 for sb1 in sb3]
    eb3 = [1e9*eb1 for eb1 in eb3]
    sb = sb + sb3
    eb = eb + eb3
   if nord >= 4:
    ux4 = ux2*ux2
    uy4 = uy2*uy2
    sb4 = [ ux4, Z, ux3*uy, Z, ux2*uy2, Z, ux*uy3, Z, uy4, Z]
    eb4 = [Z, ux4, Z, ux3*uy, Z, ux2*uy2, Z, ux*uy3, Z, uy4]
    sb4 = [1e12*sb1 for sb1 in sb4]
    eb4 = [1e12*eb1 for eb1 in eb4]
    sb = sb + sb4
    eb = eb + eb4
   if nord >= 5:
    ux5 = ux3*ux2
    uy5 = uy3*uy2
    sb5 = [ ux5, Z, ux4*uy, Z, ux3*uy2, Z, ux2*uy3, Z, ux*uy4, Z, uy5, Z]
    eb5 = [Z, ux5, Z, ux4*uy, Z, ux3*uy2, Z, ux2*uy3, Z, ux*uy4, Z, uy5]
    sb5 = [1e15*sb1 for sb1 in sb5]
    eb5 = [1e15*eb1 for eb1 in eb5]
    sb = sb + sb5
    eb = eb + eb5
   if nord >= 6:
    ux6 = ux3*ux3
    uy6 = uy3*uy3
    sb6 = [ ux6, Z, ux5*uy,   Z, ux4*uy2, Z, ux3*uy3, Z, ux2*uy4, Z, ux*uy5, Z, uy6, Z]
    eb6 = [Z, ux6, Z, ux5*uy, Z, ux4*uy2, Z, ux3*uy3, Z, ux2*uy4, Z, ux*uy5, Z, uy6]
    sb6 = [1e18*sb1 for sb1 in sb6]
    eb6 = [1e18*eb1 for eb1 in eb6]
    sb = sb + sb6
    eb = eb + eb6
   sb = transpose(sb); eb = transpose(eb)
   return concatenate([sb,eb])

def mbasis(co,ux,uy,rot=0):
   if len(shape(ux)) == 0:
      ux = asarray([ux]); uy = asarray([uy]); oo = 1
   else: oo = 0
   I=ones(ux.shape,ux.typecode())
   Z=zeros(ux.shape,ux.typecode())
   nord = len(co)
   sb = zeros(ux.shape,ux.typecode())
   eb = zeros(ux.shape,ux.typecode())
   if nord == 4:
    sb = sb + co[0]; sb = sb + co[2]*ux; sb = sb + co[3]*uy
    eb = eb + co[1]; eb = eb - pow(-1,rot)*co[2]*uy; eb = eb + co[3]*ux
   if nord >= 6:
    sb = sb + co[0]; sb = sb + co[2]*ux; sb = sb + co[4]*uy
    eb = eb + co[1]; eb = eb + co[3]*ux; eb = eb + co[5]*uy
   if nord >= 12:
    ux2 = power(ux,2)
    uy2 = power(uy,2)
    sb = sb + co[6]*ux2; sb = sb + co[8]*ux*uy; sb = sb + co[10]*uy2
    eb = eb + co[7]*ux2; eb = eb + co[9]*ux*uy; eb = eb + co[11]*uy2
   if nord >= 20:
    ux3 = ux2*ux
    uy3 = uy2*uy
    sb = sb + co[12]*ux3; sb = sb + co[14]*ux2*uy;
    sb = sb + co[16]*ux*uy2; sb = sb + co[18]*uy3
    eb = eb + co[13]*ux3; eb = eb + co[15]*ux2*uy
    eb = eb + co[17]*ux*uy2; eb = eb + co[19]*uy3
   if nord >= 30:
    ux4 = ux2*ux2
    uy4 = uy2*uy2
    sb = sb + co[20]*ux4; sb = sb + co[22]*ux3*uy
    sb = sb + co[24]*ux2*uy2; sb = sb + co[26]*ux*uy3
    sb = sb + co[28]*uy4; eb = eb + co[21]*ux4
    eb = eb + co[23]*ux3*uy; eb = eb + co[25]*ux2*uy2
    eb = eb + co[27]*ux*uy3; eb = eb + co[29]*uy4
   if nord >= 42:
    ux5 = ux3*ux2
    uy5 = uy3*uy2
    sb = sb + co[30]*ux5; sb = sb + co[31]*ux4*uy; sb = sb + co[32]*ux3*uy2
    sb = sb + co[33]*ux2*uy3; sb = sb + co[34]*ux*uy4; sb = sb + co[35]*uy5
    eb = eb + co[36]*ux5; eb = eb + co[37]*ux4*uy; eb = eb + co[38]*ux3*uy2
    eb = eb + co[39]*ux2*uy3; eb = eb + co[40]*ux*uy4; eb = eb + co[41]*uy5
   if nord >= 56:
    ux6 = ux3*ux3
    uy6 = uy3*uy3
    sb = sb + co[42]*ux6; sb = sb + co[43]*ux5*uy; sb = sb + co[44]*ux4*uy2
    sb = sb + co[45]*ux3*uy3
    sb = sb + co[46]*ux2*uy4; sb = sb + co[47]*ux*uy5; sb = sb + co[48]*uy6
    eb = eb + co[49]*ux6; eb = eb + co[50]*ux5*uy; eb = eb + co[51]*ux4*uy2
    eb = eb + co[52]*ux3*uy3
    eb = eb + co[53]*ux2*uy4; eb = eb + co[54]*ux*uy5; eb = eb + co[55]*uy6
   if oo: eb=eb[0]; sb=sb[0]
   return [sb,eb]


def vtkmbasis(co,ux,uy):
   if len(shape(ux)) < 0 or len(shape(uy)) == 0:
       sb,eb = mbasis(co,ux,uy)
   elif len(ux) < 2 or len(uy) < 2:
       sb,eb = mbasis(co,ux,uy)
   else:
       I=ones(ux.shape,ux.typecode())
       Z=zeros(ux.shape,ux.typecode())
       nord = len(co)
       sb = zeros(ux.shape,ux.typecode())
       eb = zeros(ux.shape,ux.typecode())
       if nord == 4:
        sb = VTKAdd(sb, co[0]); sb = VTKAdd(sb,VTKMultiply(ux,+co[2])); sb = VTKAdd(sb,VTKMultiply(uy,+co[3]))
        eb = VTKAdd(eb, co[1]); eb = VTKAdd(eb,VTKMultiply(uy,-co[2])); eb = VTKAdd(eb,VTKMultiply(ux,+co[3]))
       if nord >= 6:
        sb = VTKAdd(sb, co[0]); sb = VTKAdd(sb,VTKMultiply(ux,co[2])); sb = VTKAdd(sb,VTKMultiply(uy,co[4]))
        eb = VTKAdd(eb, co[1]); eb = VTKAdd(eb,VTKMultiply(ux,co[3])); eb = VTKAdd(eb,VTKMultiply(uy,co[5]))
       if nord >= 12:
        ux2 = VTKMultiply(ux,ux)
        uy2 = VTKMultiply(uy,uy)
        uxy = VTKMultiply(ux,uy)
        sb = VTKAdd(sb,VTKMultiply(ux2,co[6])); sb = VTKAdd(sb,VTKMultiply(uxy,co[8])); sb = VTKAdd(sb,VTKMultiply(uy2,co[10]))
        eb = VTKAdd(eb,VTKMultiply(ux2,co[7])); eb = VTKAdd(eb,VTKMultiply(uxy,co[9])); eb = VTKAdd(eb,VTKMultiply(uy2,co[11]))
       if nord >= 20:
        ux3 = VTKMultiply(ux2,ux)
        uy3 = VTKMultiply(uy2,uy)
        ux2y = VTKMultiply(ux2,uy)
        uy2x = VTKMultiply(ux,uy2)
        sb = VTKAdd(sb,VTKMultiply(ux3,co[12])); sb = VTKAdd(sb,VTKMultiply(ux2y,co[14]))
        sb = VTKAdd(sb,VTKMultiply(uy2x,co[16])); sb = VTKAdd(sb,VTKMultiply(uy3,co[18]))
        eb = VTKAdd(eb,VTKMultiply(ux3,co[13])); eb = VTKAdd(eb,VTKMultiply(ux2y,co[15]))
        eb = VTKAdd(eb,VTKMultiply(uy2x,co[17])); eb = VTKAdd(eb,VTKMultiply(uy3,co[19]))
       if nord >= 30:
        ux4 = VTKMultiply(ux2,ux2)
        uy4 = VTKMultiply(uy2,uy2)
        uy3x = VTKMultiply(uy3,ux)
        ux3y = VTKMultiply(ux3,uy)
        ux2y2 = VTKMultiply(ux2,uy2)
        sb = VTKAdd(sb,VTKMultiply(ux4,co[20])); sb = VTKAdd(sb,VTKMultiply(ux3y,co[22]))
        sb = VTKAdd(sb,VTKMultiply(ux2y2,co[24])); sb = VTKAdd(sb,VTKMultiply(uy3x,co[26]))
        sb = VTKAdd(sb,VTKMultiply(uy4,co[28])); eb = VTKAdd(eb,VTKMultiply(ux4,co[21]))
        eb = VTKAdd(eb,VTKMultiply(ux3y,co[23])); eb = VTKAdd(eb,VTKMultiply(ux2y2,co[25]))
        eb = VTKAdd(eb,VTKMultiply(uy3x,co[27])); eb = VTKAdd(eb,VTKMultiply(uy4,co[29]))
       if nord >= 42:
        ux5 = VTKMultiply(ux3,ux2)
        uy5 = VTKMultiply(uy3,uy2)
        ux3y2 = VTKMultiply(ux3,uy2)
        ux4y = VTKMultiply(ux4,uy)
        ux2y3 = VTKMultiply(ux3,uy3)
        uy4x = VTKMultiply(uy4,ux)
        sb = VTKAdd(sb,VTKMultiply(ux5,co[30])); sb = VTKAdd(sb,VTKMultiply(ux4y,co[31])); sb = VTKAdd(sb,VTKMultiply(ux3y2,co[32]))
        sb = VTKAdd(sb,VTKMultiply(ux2y3,co[33])); sb = VTKAdd(sb,VTKMultiply(uy4x,xco[34])); sb = VTKAdd(sb,VTKMultiply(uy5,co[35]))
        eb = VTKAdd(eb,VTKMultiply(ux5,co[36])); eb = VTKAdd(eb,VTKMultiply(ux4y,co[37])); eb = VTKAdd(eb,VTKMultiply(ux3y2,co[38]))
        eb = VTKAdd(eb,VTKMultiply(ux2y3,co[39])); eb = VTKAdd(eb,VTKMultiply(uy4x,co[40])); eb = VTKAdd(eb,VTKMultiply(uy5,co[41]))
#####   if nord >= 56:
#####    ux6 = ux3*ux3
#####    uy6 = uy3*uy3
#####    sb = VTKAdd(sb + co[42]*ux6; sb = VTKAdd(sb + co[43]*ux5*uy; sb = VTKAdd(sb + co[44]*ux4*uy2
#####    sb = VTKAdd(sb + co[45]*ux3*uy3
#####    sb = VTKAdd(sb + co[46]*ux2*uy4; sb = VTKAdd(sb + co[47]*ux*uy5; sb = VTKAdd(sb + co[48]*uy6
#####    eb = VTKAdd(eb + co[49]*ux6; eb = VTKAdd(eb + co[50]*ux5*uy; eb = VTKAdd(eb + co[51]*ux4*uy2
#####    eb = VTKAdd(eb + co[52]*ux3*uy3
#####    eb = VTKAdd(eb + co[53]*ux2*uy4; eb = VTKAdd(eb + co[54]*ux*uy5; eb = VTKAdd(eb + co[55]*uy6
   return [sb,eb]

def rd2xy(soln,ra,dec,imsize=None):
	f=open(soln,'r')
	s=f.readlines()
	f.close()

	nterms=int(split(s[0])[0])
	xc = float(split(s[0])[2])
	yc = float(split(s[0])[3])
        rot = 0.0
        if split(s[0])[4:]: rot = float(split(s[0])[4])
        if find(split(s[1])[0],":")+1:
	   rc = hrrad(scanSexagesimal(split(s[1])[0]))
	   dc = degrad(scanSexagesimal(split(s[1])[1]))
        else:
	   rc = float(split(s[1])[0])
	   dc = float(split(s[1])[1])

	co=[]
	for n in range(0,nterms): co.append(float(strip(s[n+2])))
        co=array(co)

	squ,eta = normalcoord(ra,dec,rc,dc)

        squ,eta = squ*cos(degrad(rot)) - eta*sin(degrad(rot)),  \
                squ*sin(degrad(rot)) + eta*cos(degrad(rot))

        if len(co) == 4:
           deter = co[2]*co[2] + co[3]*co[3]
           x = ( co[2]*(squ-co[0]) - co[3]*(eta-co[1]))/deter - 0.5 + xc
           y = ( co[3]*(squ-co[0]) - co[2]*(eta-co[1]))/deter - 0.5 + yc
        elif len(co) >= 6:
           deter = co[2]*co[5] - co[3]*co[4]
           x = ( co[5]*(squ-co[0]) - co[3]*(eta-co[1]))/deter - 0.5 + xc
           y = (-co[4]*(squ-co[0]) + co[2]*(eta-co[1]))/deter - 0.5 + yc

	scale = 180*sqrt(co[2]**2+co[3]**2)/pi * 3600
	impa = (math.atan2(co[3],co[2]) + pi) % (2*pi)

        if len(co) > 6:
          if 0:
             if imsize:
                dyc,dxc = imsize
                ux = arange(0.0,dxc+dxc/16.,dxc/16)
                uy = arange(0.0,dyc+dxc/16.,dyc/16)
             else:
                dxc = max(abs(x-xc))
                dyc = max(abs(y-yc))
                if dxc < 100: dxc = 100
                if dyc < 100: dyc = 100
                ux = arange(min(x)-dxc,max(x)+dxc,dxc/8)
                uy = arange(min(y)-dyc,max(y)+dyc,dyc/8)
             nx = len(ux)
             ny = len(uy)
             oo = ones((ny,nx))
             yy = (oo*uy[::,NewAxis]).flat - yc
             xx = (oo*ux[NewAxis,::]).flat - xc
             uu,vv = mbasis(co,xx,yy)
             nterm = [4,6,12,20,30,42,56]
             nords = [0,1,2,3,4,5,6]
             mm = nords[nterm.index(len(co))]
             bb = abasis(mm,uu,vv)
             cc = svdfit(bb,concatenate([xx,yy]))
             x,y = mbasis(cc,squ,eta)
             x = x + xc
             y = y + xc
          else:
             nd = 8.0
             if imsize:
                dyc,dxc = imsize
                ux = arange(-dxc/nd,imsize[1]+0.5*dxc/nd,dxc/32)
                uy = arange(-dyc/nd,imsize[0]+0.5*dxc/nd,dyc/32)
             else:
                dxc = max(abs(x-xc))
                dyc = max(abs(y-yc))
                if dxc < 100: dxc = 100
                if dyc < 100: dyc = 100
                ux = arange(min(x)-dxc,max(x)+dxc,dxc/nd)
                uy = arange(min(y)-dyc,max(y)+dyc,dyc/nd)
             x4,y4 = asarray(len(uy)*[ux]), transpose(len(ux)*[uy])
             x4,y4 = x4.flat, ravel(y4)
             tra,tdec = xy2rd(soln,x4,y4,text=0)
             tsqu,teta = normalcoord(tra,tdec,rc,dc)
             dsqu = (max(tsqu)-min(tsqu))/nd
             deta = (max(teta)-min(teta))/nd
             kx,ky = 3,3
#             xknots = concatenate([ (kx+1)*[min(tsqu)-2*dsqu], arange(min(tsqu)-1*dsqu,max(tsqu)+1*dsqu,dsqu), (kx+1)*[max(tsqu)+2*dsqu]])
#             yknots = concatenate([ (ky+1)*[min(teta)-2*deta], arange(min(teta)-1*deta,max(teta)+1*deta,deta), (ky+1)*[max(teta)+2*deta]])
             xknots = concatenate([ (kx+1)*[min(tsqu)], (kx+1)*[max(tsqu)]])
             yknots = concatenate([ (ky+1)*[min(teta)], (ky+1)*[max(teta)]])
             if imsize:
                xb = min(xknots); xe = max(xknots); yb = min(yknots); ye = max(yknots)
             else:
                xb = min(xknots); xe = max(xknots); yb = min(yknots); ye = max(yknots)
             xtck = bisplrep(tsqu,teta,x4-xc,task=-1,s=None,kx=kx,ky=ky,tx=xknots,ty=yknots,\
                xb=xb, xe=xe, yb=yb, ye=ye, nxest=len(xknots), nyest=len(yknots), full_output=0,quite=1)
             ytck = bisplrep(tsqu,teta,y4-yc,task=-1,s=None,kx=kx,ky=ky,tx=xknots,ty=yknots,\
                xb=xb, xe=xe, yb=yb, ye=ye, nxest=len(xknots), nyest=len(yknots), full_output=0,quite=1)
             if imsize:
               bra,bdec = xy2rd(soln,array([0.0,imsize[1],imsize[1],0]),array([0.0,0.0,imsize[0],imsize[0]]),text=0)
               bsqu,beta = normalcoord(bra,bdec,rc,dc)
               good = between(squ,min(bsqu),max(bsqu))* between(eta,min(beta),max(beta))
             if len(squ) * len(eta) == 1:
                x = asarray([bisplev(squ,eta,xtck) + xc])
                y = asarray([bisplev(squ,eta,ytck) + yc])
             else:
                x = bisplev2(squ,eta,xtck) + xc
                y = bisplev2(squ,eta,ytck) + yc
             if imsize:
                x=where(good,x,-900)
                y=where(good,y,-900)

     	return x, y, scale, impa

def refract(ra, dec, ha0=0.0, latd=19.8265, pres=615.0, temp=0.0, wave=0.8):

	lat = degrad(latd)
	ra = ra
	dec = dec
	ha = ha0 + (ra-ra[0])

	sina = sin(lat)*sin(dec) + cos(lat)*cos(dec)*cos(ha)
	cosz = sina
	sinz = sqrt(1.0-power(sina,2))
	tanz = sinz/cosz

	sinq=cos(lat)*sin(ha)/sinz
	cosq=(cos(pi/2-lat)-cosz*cos(pi/2-dec))/(sinz*sin(pi/2-dec))

	wavers = 1./(wave*wave)
	n = 64.328 + (29498.1/(146.-wavers)) + (255.4/(41.-wavers))
	n = 1.E-6*n

	tcorr = 1. + 0.003661*temp
	num = 720.88*tcorr
	n = n*((pres*(1.+(1.049-0.0157*temp)*1.E-6*pres))/num)
	r = n*206265.0* tanz

        r = pi*r/3600/180
	dra =  r*sinq/cos(dec)
        ddec = r*cosq
        dra = dra - dra[0]
        ddec = ddec - ddec[0]

        ara = ra + dra
        adec = dec + ddec

	return ara, adec

def getstars(stars,soln,rotate=0):
	(stargrid,rt)=readsim(stars)
	starid = stargrid[0]
	sra=[]; sdec=[]
	for sin in range(0,len(starid)):
		sra.append(hrrad(scanSexagesimal(stargrid[3][sin])))
		sdec.append(degrad(scanSexagesimal(stargrid[4][sin])))
	sra=array(sra,Float32)
	sdec=array(sdec,Float32)
	sx, sy, scale, ipa = rd2xy(soln,sra,sdec)

	return sx, sy, scale, ipa, starid, stargrid

def getmasks(masks,soln=None,ha=0,rotate=0,inst="lris",verb=0,wave=0.8):
 mid=[]
 rac=[]
 decc=[]
 maskinst=[]
 maskgrid=[]
 maskpa=[]
 maskpar=[]
 mxpos=[]
 mypos=[]
 for ncat in range(0,len(masks)):
   mask=masks[ncat]
   if verb: print "Mask:",mask
   maskgrid0=readsim(mask)[0]
   maskgrid.append(maskgrid0)
   mid.append(maskgrid[ncat][0])
   maskpa.append( degrad(float(maskgrid[ncat][5][0])))
   ra=[]; dec=[]; slitpa=[]
   rac.append(-999.0); decc.append(-999.0)
   for gin in range(0,len(mid[ncat])):
	ra.append(hrrad(scanSexagesimal(maskgrid[ncat][3][gin])))
	dec.append(degrad(scanSexagesimal(maskgrid[ncat][4][gin])))
	if mid[ncat][gin] == "CENTER":
	  rac[ncat]=ra[gin]
	  decc[ncat]=dec[gin]

          if maskgrid[ncat][6][gin] > 0:
             boxwid = 1.*maskgrid[ncat][6][gin]
             maskgrid[ncat][6][gin] = -999.
          else: boxwid = None
          if maskgrid[ncat][7][gin] > 0:
             wran = 1.*maskgrid[ncat][7][gin]
             maskgrid[ncat][7][gin] = -999.
          else: wran = None
          I = slitmask(inst,box=boxwid,wran=wran)

          if I.instrument == "deimos":
              xoff,yoff = I.XOFFSET*cos(maskpa[-1]) - I.YOFFSET*sin(maskpa[-1]), \
                      I.XOFFSET*sin(maskpa[-1]) + I.YOFFSET*cos(maskpa[-1])
              xoff = -degrad(xoff/3600)
              yoff = degrad(yoff/3600)
              rac[ncat],decc[ncat] = equicoord(xoff,yoff,rac[ncat],decc[ncat])

   	spa = maskgrid[ncat][5][gin]
#        print spa, maskpa[ncat]
	if maskgrid[ncat][5][gin] > -998: spa = slitpa.append(degrad(spa))
	else: slitpa.append(maskpa[ncat])

   maskinst.append(I)
   ra=array(ra,Float64)
   dec=array(dec,Float64)
   prior=array(maskgrid[ncat][1],Int32)
   mag=array(maskgrid[ncat][2],Float32)
   slitpa=array(slitpa,Float64)
   bot= array(maskgrid[ncat][6],Float32)
   top= array(maskgrid[ncat][7],Float32)
   if maskpa > -999:
	maskpar0=I.mapmask(mid[ncat],ha,ra,dec,prior,mag,slitpa,bot,top,verb=verb,wave=wave)
	maskpar.append(maskpar0)
   if soln:
     xpos0, ypos0, scale, impa = rd2xy(soln,ra,dec)
     mxpos.append(xpos0)
     mypos.append(ypos0)
   else:
     impa = 0
     mxpos.append(maskpar0[0])
     mypos.append(maskpar0[1])

 return mid, mxpos, mypos, maskpar, maskgrid, maskpa, rac, decc, impa, maskinst

def getcatalogs(catalogs,csts,xcols,ycols,tcolss,icolss,soln,rotate=0,imsize=None):
 id=[]
 catgrid=[]
 xpos=[]
 ypos=[]
 for ncat in range(0,len(catalogs)):
   if soln:
    catalog=catalogs[ncat]
    (catgrid0,catt)=readsim(catalog)
    catgrid.append(catgrid0)
    id.append(catgrid[ncat][0])
    ra=[]; dec=[]
    for gin in range(0,len(id[ncat])):
         ra.append(hrrad(scanSexagesimal(catgrid[ncat][3][gin])))
         dec.append(degrad(scanSexagesimal(catgrid[ncat][4][gin])))
        
    ra=array(ra,Float32)
    dec=array(dec,Float32)
    prior=array(catgrid[ncat][1],Int32)
    mag=array(catgrid[ncat][2],Float32)
    xpos0, ypos0, scale, ipa = rd2xy(soln,ra,dec,imsize=imsize)
    xpos.append(xpos0)
    ypos.append(ypos0)
   else:
    catalog=catalogs[ncat]
    cst=csts[ncat]
    xcol=xcols[ncat]
    ycol=ycols[ncat]
    tcols=tcolss[ncat]
    icols=icolss[ncat]
    catgrid0=readtable(catalog,tcols,icols,cst)
    catgrid.append(catgrid0)
    id.append(catgrid[ncat][0])
    xpos0, ypos0=catgrid[ncat][xcol]+0.5, catgrid[ncat][ycol]+0.5
    xpos.append(xpos0)
    ypos.append(ypos0)
  
 return id, xpos, ypos, catgrid

def normalcoord(ra,dec,rc,dc,conv=1.0):
      squ=cos(dec)*sin(ra-rc)/	\
      (sin(dc)*sin(dec)+cos(dc)*cos(dec)*cos(ra-rc))
      eta=(cos(dc)*sin(dec) - sin(dc)*cos(dec)*cos(ra-rc))/ \
      (sin(dc)*sin(dec)+cos(dc)*cos(dec)*cos(ra-rc))
      return squ*conv,eta*conv

def equicoord(u,v,rc,dc,text=0):
        t = u/(cos(dc)-v*sin(dc))
        tt=(sin(dc)+v*cos(dc))/(cos(dc)-v*sin(dc))*cos(arctan(t))
        dec = arctan(tt)
        ra = arctan(t)+rc

        if text:
           if len(shape(ra)) == 0:
             dec = formatSexagesimal(raddeg(dec), 4, 3600000)
             ra = formatSexagesimal(radhr(ra), 3, 3600000)
           else:
             dec = [formatSexagesimal(raddeg(dec), 4, 3600000) for dec in dec]
             ra = [formatSexagesimal(radhr(ra), 3, 3600000) for ra in ra]
        return ra,dec

def mkobj( oldmjd, id, ora, odec, pa):
   obj=Obj()
   obj.any.type=FIXED
   obj.f.epoch=oldmjd
   obj.f.ra=ora
   obj.f.dec=odec
   if pa == -999: pa = 0.0
   obj.f.positionAngle=pa
   return obj

def precess(ra,dec,epoch_in=1950,epoch_out=2000):
#   preference.whichEquatorial = 0
   oldmjd=fromYear(epoch_in)
   newmjd=fromYear(epoch_out)
   space=Circumstance()
   space.mjd = newmjd
   space.epoch = newmjd
   if len(shape(ra)) == 0: onlyone=1; ra,dec=[ra],[dec]
   else: onlyone = 0
   nobj = len(ra)
   gobj = [mkobj(oldmjd,"xx",ra[n],dec[n],0) for n in xrange(nobj)]
   cobj = [computeLocation(space, g) for g in gobj]
   pra = asarray([g.f.ra for g in gobj])
   pdec = asarray([g.f.dec for g in gobj])
   if onlyone: pra,pdec=pra[0],pdec[0]
   del gobj,cobj
   return pra, pdec

def lrismill(millfile,idmask,owner,x,y,ymin,ymax,theta,type,npts,slitwid):
   f=open(millfile,"w")
   print >>f,"# (     MaskName:     %25s)", idmask
   print >>f,"# (     MaskOwner:    %25s)", owner
   print >>f,"# (     Date:         %25s)", time.asctime()
   print >>f,"# (     Slit width:   %5.3f mm = %5.2f arcsec)", \
        slitwid, slitwid/MM_ARCS
   f.close()

def ast_vr(ra1,dec1,v1,ra2,dec2):
   vx = v1 * cos(ra1)  * cos(dec1)
   vy = v1 * sin(ra1)  * cos(dec1)
   vz = v1 * sin(dec1) * cos(dec1)
   cc = cos(dec2)  * cos(ra2)
   cs = sin(dec2)  * cos(ra2)
   s  = sin(dec2)
   v2 = (vx*cc + vy*cs + vz*s)
   return v2

def ast_vorbit(ra,dec,epoch):
   t = (fromYear(epoch) + 2400000.5 - 2415020.0)/36525.0

   manom = 358.47583 + t * (35999.04975 - t * (0.000150 + t * 0.000003))
   lperi = 101.22083 + t * (1.7191733 + t * (0.000453 + t * 0.000003))
   oblq = 23.452294 - t * (0.0130125 + t * (0.00000164 - t * 0.000000503))
   eccen = 0.01675104 - t * (0.00004180 + t * 0.000000126)

   manom = manom % 360.0
   lperi = lperi % 360.0

   r = ra
   d = dec
   manom = degrad(manom)
   lperi = degrad(lperi)
   oblq = degrad(oblq)

   tanom = manom + (2 * eccen - 0.25 * eccen**3) * sin (manom) + \
       1.25 * eccen**2 * sin (2 * manom) + \
       13./12. * eccen**3 * sin (3 * manom)

   slong = lperi + tanom + pi

   # L and B are the longitude and latitude of the star in the orbital
   # plane of the Earth (radians)

   l,b = ast_coord(0.0,0.0,-pi/2.0, pi/2.0 - oblq, r, d)
   vorb = ((2*pi/ 365.2564) * 149598500.0 / sqrt (1. - eccen**2)) / 86400.0
   v = vorb * cos (b) * (sin (slong - l) - eccen * sin (lperi - l))
   return v

def ast_coord(ao,bo,ap,bp,a1,b1):
    x = cos (a1) * cos (b1)
    y = sin (a1) * cos (b1)
    z = sin (b1)
    xp = cos (ap) * cos (bp)
    yp = sin (ap) * cos (bp)
    zp = sin (bp)
    sao = sin (ao)
    cao = cos (ao)
    sbo = sin (bo)
    cbo = cos (bo)
    temp = -xp * sao + yp * cao
    xp = xp * cao + yp * sao
    yp = temp
    temp = -x * sao + y * cao
    x = x * cao + y * sao
    y = temp
    temp = -xp * sbo + zp * cbo
    xp = xp * cbo + zp * sbo
    zp = temp
    temp = -x * sbo + z * cbo
    x = x * cbo + z * sbo
    z = temp
    sbp = zp
    cbp = yp
    temp = y * cbp + z * sbp
    y = y * sbp - z * cbp
    z = temp
    a2 = math.atan2(y, x)
    b2 = math.asin (z)
    return a2,b2

def ast_vbary(ra,dec,epoch):
   t = (fromYear(epoch) + 2400000.5 - 2415020.0)/36525.0
   oblq = 23.452294 - t * (0.0130125 + t * (0.00000164 - t * 0.000000503))
   omega = 259.183275 - t * (1934.142008 + t * (0.002078 + t * 0.000002))
   llong = 270.434164 + t * (481267.88315 + t * (-0.001133 + t * 0.0000019)) - omega
   lperi = 334.329556 + t * (4069.034029 - t * (0.010325 + t * 0.000012)) - omega
   em = 0.054900489
   inclin = 5.1453964
   lperi = degrad(lperi)
   llong = degrad(llong)
   anom = llong - lperi
   anom = anom + (2. * em - 0.25 * em**3) * sin (anom) + 1.25 * em**2 * \
       sin (2. * anom) + 13./12. * em**3 * sin (3. * anom)
   llong = anom + lperi
   r = degrad (ra * 15)
   d = degrad (dec)
   omega = degrad (omega)
   oblq = degrad (oblq)
   inclin = degrad (inclin)
   l,b = ast_coord (0.0, 0.0, -pi/2.0, pi/2.0 - oblq, r, d)
   lm,bm = ast_coord (omega, 0.0, omega-pi/2.0, pi/2.0-inclin, l, b)
   vmoon = (2*pi / 27.321661) * 384403.12040 / sqrt (1. - em**2) / 86400.
   v = vmoon * cos (bm) * (sin (llong - lm) - em * sin (lperi - lm))
   v = v / 81.53
   return v

def ast_vrotate(ra,dec,epoch,latitude,longitude,altitude):
    lat = degrad (latitude) 
    dlat = -(11. * 60. + 32.743000) * sin (2 * lat) + \
            1.163300 * sin (4 * lat) -0.002600 * sin (6 * lat)
    lat = lat + degrad (dlat / 3600.)
    r = 6378160.0 * (0.998327073 + 0.00167643800 * cos (2 * lat) - \
        0.00000351 * cos (4 * lat) + 0.000000008 * cos (6 * lat)) + \
        altitude
    vc = 2*pi * (r / 1000.)  / (23.934469591229 * 3600.)
    lmst = ast_mst (epoch, longitude)
    v = vc * cos (lat) * cos (degrad (dec)) * \
        sin (degrad ((ra - lmst) * 15.))
    return v

def ast_mst (epoch, longitude):
    jd = fromYear(epoch)
    ut = (jd - int (jd) - 0.5) * 24.
    t = (jd - 2451545.) / 36525.
    st = 24110.54841 + t * (8640184.812866 + t * (0.093104 - t * 6.2-6))
    st = (st / 3600. + ut - longitude / 15.) % 24.0
    if st < 0: st = st + 24
    return st

def radec2galactic(ra,dec,epoch=1950.0):
        degrad = lambda x: x*pi/180
        raddeg = lambda x: x*180/pi

        LONGNCP	 = degrad(123.00)
        RAGPOLE	 = degrad(192.25)
        DECGPOLE = degrad(27.4)
        GEPOCH	 = 1950.0
	rar,decr = precess(ra, dec, epoch, GEPOCH)

	# Precompute the necessary constants.
	drar = rar - RAGPOLE
	cosdecg = cos(DECGPOLE)
	sindecg = sin(DECGPOLE)
	cosdecr = cos(decr)

	# Compute the tansformation equations
	x = cosdecr * cos(drar)
	y =  cosdecr * sin(drar)
	z = sin (decr)
	temp = z * cosdecg - x * sindecg 
	z = z * sindecg + x * cosdecg 
	x = temp
	r = sqrt (x * x + y * y)

	# Compute lii and bii and convert to degrees.
	lii = LONGNCP  + arctan2(-y, x)
        if (lii < 0.0): lii = lii + 2*pi
	bii = arctan2(z, r)

        return lii,bii

