
from os import chmod
from basics import *
from FITS import *

class MIKErecord:
    def __init__(self,record,lampkey,flatkey,slitkey):
        self.record = record
        self.filename,self.inst,self.date,self.start,self.mjd,self.etime = self.record.split()[:6]
        self.side = self.inst.split("-")[1].lower()
        self.object = join(self.record.split()[6:],"").lower()
        if lampkey in self.object: self.type = "lamp"
        elif flatkey in self.object: self.type = "flat"
        elif slitkey in self.object: self.type = "slit"
        else: self.type = "other"
        self.mjd = float(self.mjd)
        self.etime = float(self.etime)
        self.used = False

class MIKEdb:
    def __init__(self,dbfile,blue,red,lampkey,flatkey,slitkey):
        self.file = open(dbfile,"r")
        self.read(blue,red,lampkey,flatkey,slitkey)
        self.file.close()
    def read(self,blue,red,lampkey,flatkey,slitkey):
        self.records = [MIKErecord(rec,lampkey,flatkey,slitkey) for rec in self.file.xreadlines() if rec[0] != "#"]
        self.records = [rec for rec in self.records if ("blue" in rec.inst.lower() and blue) or ("red" in rec.inst.lower() and red)]

CALIBTYPES = ["lamp","flat","slit"]

class MIKEobservations:
    def __init__(self,target,dbdata,onlyflat,useflat,twodim,logb,clean):
        self.target = target
        self.dbdata = dbdata
        self.onlyflat = onlyflat
        self.useflat = useflat
        self.twodim = twodim
        self.clean = clean
        self.logb = logb
        self.IdentifyObject()
        print self.spectra
        self.useflat = "%sflat.fits" % (self.dbdata.records[self.spectra[0]].side)
        if not os.path.exists(self.useflat):
           print self.useflat
           [self.AssociateCalib(n,"flat") for n in self.spectra]
           print self.dbdata.records[self.spectra[0]].useflat

    def IdentifyObject(self):
        self.spectra = nonzero([(self.target in rec.object and rec.type not in CALIBTYPES) for rec in self.dbdata.records])

    def Associate(self):
        print "slits"
        [self.AssociateCalib(n,"slit") for n in self.spectra]
        print "lamps"
        [self.AssociateCalib(n,"lamp") for n in self.spectra]

    def AssociateCalib(self,n,ctype):
        nrec = arange(len(self.dbdata.records))
        record = self.dbdata.records[n].record
        fname = self.dbdata.records[n].filename
        object = self.dbdata.records[n].object
        mjd = self.dbdata.records[n].mjd
        etime  = self.dbdata.records[n].etime
        date = self.dbdata.records[n].date
        inst = self.dbdata.records[n].inst
        use = [(rec.inst == inst and rec.type == ctype) for rec in self.dbdata.records]
        dmjd = [abs(rec.mjd - mjd) + 1e38 * (rec.inst != inst) + 1e38*(rec.type != ctype) for rec in self.dbdata.records]
        dmjd = where(equal(nrec,n),1e38,dmjd)
        print n,fname,object,mjd,min(dmjd),argmin(dmjd)
        if ctype == "lamp":
           self.dbdata.records[n].uselamp = argmin(dmjd)
           self.dbdata.records[n].lampdmjd = min(dmjd)
#           print ctype, self.dbdata.records[ self.dbdata.records[n].uselamp].filename
        elif ctype == "flat":
           self.dbdata.records[n].useflat = nonzero(use)
        elif ctype == "slit":
           self.dbdata.records[n].useslit = argmin(dmjd)
           self.dbdata.records[n].slitdmjd = min(dmjd)
#           print ctype, self.dbdata.records[ self.dbdata.records[n].useslit].filename

#        self.Prefix = pre
#        self.ReadInput()
#        self.Organize()
#        if not sum(self.specs) and sum(self.flats):
#           self.onlyflat = 1
#        if not sum(self.flats) and self.onlyflat:
#           print "Error! You are trying to make a flat field?"
#           exit()
#        if not sum(self.flats) and not self.useflat:
#           print "Error! Need to make or borrow a flat field."
#           exit()
#        self.OpenPipeline()
#        self.WriteSteps()
#        self.ClosePipeline()

    def ReadInput(self):
        F = open(self.input,"r")
        self.object = F.readline()[1:].strip()
        Fl = F.readlines()
        F.close()
        keywords = ["image","directory","procedures"]
        self.information = {}
        for kw in keywords: self.information[kw] = []
        for l in Fl:
           if l[0] != "#":
              v = l.replace(","," ").split()
              for n in range(2):
                 self.information[keywords[n]].append(v[n])
              self.information[keywords[2]].append([])
              for i in v[2:]: self.information[keywords[2]][-1].append(i)
        self.nimages = len(self.information["image"])
        self.specs = zeros((self.nimages,),Int32)
        self.flats = zeros((self.nimages,),Int32)
        self.slits = zeros((self.nimages,),Int32)
        self.xdists = zeros((self.nimages,),Int32)
        self.ydists = zeros((self.nimages,),Int32)
        self.wdists = zeros((self.nimages,),Int32)
        self.pipeline = "%s_%s.pipeline" % (self.object,self.Prefix)
        self.images = []

    def Organize(self):
        self.information["root"] = []
        for n in range(self.nimages):
           if "flat" in self.information["procedures"][n]: root = "flat"
           elif "milky" in self.information["procedures"][n]: root = "flat"
           elif self.object in self.information["procedures"][n]: root = "spec"
           elif "ydist" in self.information["procedures"][n]: root = "ydist"
           elif "xdist" in self.information["procedures"][n]: root = "xdist"
           elif "slits" in self.information["procedures"][n]: root = "slits"
           elif "orders" in self.information["procedures"][n]: root = "orders"
           self.information["root"].append( "%s_%s%s" % (self.object,self.Prefix,root))
           self.information["image"][n] = self.information["image"][n].zfill(4)
           if self.object in self.information["procedures"][n]: self.specs[n] = 1
           if "xdist" in self.information["procedures"][n]: self.xdists[n] = 1
           if "ydist" in self.information["procedures"][n]: self.ydists[n] = 1
           if "wdist" in self.information["procedures"][n]: self.wdists[n] = 1
           if "flat" in self.information["procedures"][n]: self.flats[n] = 1
           if "milky" in self.information["procedures"][n]: self.flats[n] = 1
           if "slits" in self.information["procedures"][n]: self.slits[n] = 1
           if "orders" in self.information["procedures"][n]: self.slits[n] = 1
        self.ispecs = nonzero(self.specs)
        self.iflats = nonzero(self.flats)
        self.islits = nonzero(self.slits)
        self.ixdists = nonzero(self.xdists)
        self.iydists = nonzero(self.ydists)
        self.iwdists = nonzero(self.wdists)
        if self.iflats == self.ispecs: self.onlyflat = True

    def OpenPipeline(self):
        self.pipe = open( self.pipeline,"w")

    def WriteLine(self,line=""):
        if line:
            if line[0] == "#":
               print >> self.pipe, "echo %s" % (line[1:])
        print >> self.pipe, line

    def WriteSteps(self):
        self.WriteLine("# Overscan Subtraction, Trimming, and Image Reorientation.")
        [self.Overscan(n) for n in range(self.nimages)]
        self.WriteLine()

        if len(self.iydists):
            self.WriteLine("# Measurement and Fitting of Y Distortion.")
            self.YDistortion()
            self.WriteLine()

            self.WriteLine("# Copy Y Distortion Information.")
            [self.Copyrect("-ydist",n) for n in self.islits if self.images[n] != self.ydistortion]
            [self.Copyrect("-ydist",n) for n in self.ixdists  if self.images[n] != self.ydistortion]
            [self.Copyrect("-ydist",n) for n in self.ispecs  if self.images[n] != self.ydistortion]
            self.WriteLine()

        if len(self.islits):
            self.WriteLine("# Find the Order Edges.")
            self.FindOrders()
            self.WriteLine()

            self.WriteLine("# Copy Order Information.")
            [self.CopyOrder(n,arc=1) for n in self.ixdists if self.images[n] != self.foundslit]
            [self.CopyOrder(n) for n in self.iydists if self.images[n] != self.foundslit]
            [self.CopyOrder(n) for n in self.ispecs if self.images[n] != self.foundslit]
            self.WriteLine()

        if len(self.ixdists):
            self.WriteLine("# Prepare Image for X Distortion.")
            [self.PrepareX(n) for n in self.ixdists]
            self.WriteLine()

            self.WriteLine("# Measurement and Fitting of X Distortion.")
            self.XDistortion()
            self.WriteLine()

            self.WriteLine("# Global Fitting of X Distortion Parameters.")
            self.GlobalXDistortion()
            self.WriteLine()

            self.WriteLine("# Regenerate X Distortion Mapping.")
            self.MapXDistortion()
            self.WriteLine()

            self.WriteLine("# Copy X Distortion Information.")
            [self.Copyrect("-xdist",n) for n in range(self.nimages) if self.images[n] != self.xdistortionb]
            self.WriteLine()

        if len(self.iwdists):
           self.wdistortion = self.images[self.iwdists[0]]
           self.WriteLine("# Identify the Order Numbers and Produce 1D Spectra for Wavelength Identifications")
           self.MatchOrders()

           self.WriteLine("# Generate Database of Individual Comparison Lamp Lines")
           self.FindLines( self.wdistortion,th=5)

           self.WriteLine("# Identify Individual Spectral Features and Fit for Dispersion Solution")
           self.MatchLamps()

           self.WriteLine("# Copy Dispersion Solution Information")
           [self.CopyECOrders(n) for n in range(self.nimages) if self.images[n] != self.wdistortion]
           self.WriteLine()
           [self.Copyrect("-wdist",n) for n in range(self.nimages) if self.images[n] != self.wdistortion]
           self.WriteLine()
           [self.Copyrect("-edist",n) for n in range(self.nimages) if self.images[n] != self.wdistortion]
           self.WriteLine()

           if "r" in self.Prefix.lower():
               self.WriteLine("# Generate 1D Spectra with Sky Lines")
               [self.Get1DSky(self.images[n]) for n in self.ispecs]
               self.WriteLine()

               self.WriteLine("# Generate Database of Individual Sky Lines")
               [self.FindLines(self.images[n],th=5) for n in self.ispecs]
               self.WriteLine()

               self.WriteLine("# Identify Individual Spectral Features and Fit for Wavelength Shift")
               [self.MatchSky(self.images[n]) for n in self.ispecs]
               self.WriteLine()

        if self.useflat:
             self.flat2d = self.useflat
        else:
             self.WriteLine("# Normalizing the Flatfields.")
             [self.NormFlat(n) for n in self.iflats]
             self.WriteLine()
       
             self.WriteLine("# Combine the Normalized Flatfields.")
             self.CombineFlats("flt")
             self.WriteLine()

        if not self.onlyflat:

             self.WriteLine("# Divide by Pixel-To-Pixel Flatfield.")
             [self.Flatten(n) for n in self.ispecs]
             self.WriteLine()

             self.WriteLine("# Generate Sky Apertures.")
             self.SkyAps()
             self.WriteLine()

             self.WriteLine("# Make New Masks.")
             [self.NewMask(n) for n in self.ispecs]

             self.WriteLine("# Fit the Sky Background")
             [self.SkyModel(n) for n in self.ispecs]
             self.WriteLine()

             self.WriteLine("# Subtract the Fitted Sky Background.")
             [self.SkySubtract(n) for n in self.ispecs]
             self.WriteLine()

             self.WriteLine("# Generate Object Positions.")
             self.PosObjects()
             self.AllPos()
             self.WriteLine()

             self.WriteLine("# Run Multiunrect.")
             self.Unrect()
             self.WriteLine()

             self.WriteLine("# Collate.")
             self.CollateRect()
             self.WriteLine()

             if not self.twodim:
                self.WriteLine("# Extract.")
                self.MultiExtract()
                self.WriteLine()

    def ClosePipeline(self):
        chmod(self.pipe.name,493)
        self.pipe.close()
           
    def Overscan(self,n):
        image = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(image,readonly=1)
        self.binning = int(F["binning"].split("x")[1])
        self.sbinning = int(F["binning"].split("x")[0])
        F.close()
        linebias = 0
        if self.object in self.information["procedures"][n]: u = searchsorted(self.ispecs,n); linebias = 1
        elif "flat" in self.information["procedures"][n]: u = searchsorted(self.iflats,n); linebias = 1
        elif "milky" in self.information["procedures"][n]: u = searchsorted(self.iflats,n); linebias = 1
        elif "xdist" in self.information["procedures"][n]: u = searchsorted(self.ixdists,n); linebias = 1
        elif "ydist" in self.information["procedures"][n]: u = searchsorted(self.iydists,n); linebias = 1
        elif "slits" in self.information["procedures"][n]: u = searchsorted(self.islits,n); linebias = 0
        elif "orders" in self.information["procedures"][n]: u = searchsorted(self.islits,n); linebias = 0
        new = "%s%s.fits" % (self.information["root"][n],zfill(u+1,2))
        command = "mikebias","-c 3",image,new
        if linebias: command += "-linebias",
        command = join(command)
        self.images.append(new)
        self.WriteLine(command)

    def YDistortion(self,nx=4,ny=64,ns=4,ox=2,oy=4,pw=10):
        if "r" in self.Prefix.lower(): ox,oy = 2,4
        elif "b" in self.Prefix.lower(): ox,oy = 1,5
        n = self.iydists[0]
        image = self.images[n]
        command = "getrect","-ydist", image
        command += "-nx %d" % (nx), "-ny %d" % (ny), "-x %d" % (ox),"-dump"
        command += "-nsub2 %d" % (ns), "-y %d" % (oy), "-peakwin %d" % (pw)
        command = join(command)
        self.WriteLine(command)
        self.ydistortion = image

    def Copyrect(self,dist,n):
        image = self.images[n]
        if dist == "-ydist": d = self.ydistortion
        elif dist == "-xdist": d = self.xdistortionb
        elif dist in ["-wdist","-edist"]: d = self.wdistortion
        command = join(["copyrect",d,dist,image])
        self.WriteLine(command)

    def CopyECOrders(self,n):
        image = self.images[n]
        command = join(["copyecords",self.wdistortion,image])
        self.WriteLine(command)

    def FindOrders(self,minlen=20,pmin=0.2,pmax=999.0,oo=3):
        n = self.islits[0]
        image = self.images[n]
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        rn = F["enoise"]
        g = F["egain"]
        binning = int(F["binning"].split("x")[0])
        F.close()
        if len(self.islits) > 1:
           image = "%s_%sborder.fits" % (self.object,self.Prefix)
           command = ["efits"]
           command += [self.images[n] for n in self.islits]
           command += ["\'sum(allimages)\'", image]
           command = join(command)
           self.WriteLine(command)
           self.WriteLine()
        command = "findorders",image,"-d",image.replace(".fits","orders.ps/vcps"),"-rn %f" % (rn),"-g %f" % (g)
        command += "-len %d" % (minlen*2.0/binning),
        command += "-min %f" % (pmin), "-max %f" % (pmax),
        command += "-extend", "-oo %d" % (oo),
        command = join(command)
        self.WriteLine(command)
        self.foundslit = image

    def CopyOrder(self,n,arc=0):
        image = self.images[n]
        if arc:
           fw = 10*2.0/self.sbinning
           command = "copyslit","-fwhm %f" % (fw), "-ydist","-fft",self.foundslit,image
           command = join(command)
           self.foundslit = image
        else:
           command = "importslits","-f",self.foundslit,image
           command = join(command)
        self.WriteLine(command)

    def PrepareX(self,n):
        image = self.images[n]
        imageb = image.replace(".fits","b.fits")
        command = "efits",image
        command += "\'i1-VTKMedian(i1,1,51,numret=1)\'",imageb
        command = join(command)
        self.WriteLine(command)
        command = "efits",imageb
        command += "\'i1-VTKMedian(i1,11,3,numret=1)\'",imageb
        command = join(command)
        self.WriteLine(command)
        self.xdistortion = image
        self.xdistortionb = imageb

    def XDistortion(self,nx=1,dy=3,nsub2=1,ox=0,oy=1,b=2,nwt=100):
        image = self.xdistortionb
        command = "getrect","-xdist","-ydist",image
        command += "-nx %d" % (nx), "-dy %d" % (dy), "-x %d" % (ox),
        command += "-nsub2 %d" % (nsub2), "-y %d" % (oy)
        command += "-b %d" % (b), "-normwt %f" % (nwt),
        command = join(command)
        self.WriteLine(command)

    def MatchLamps(self,xo=5,oo=4):
        image = self.wdistortion
        command = "mikeMatchLamps",image
        command += "-x %d" % (xo), "-o %d" % (oo)
        command = join(command)
        self.WriteLine(command)
        self.WriteLine()
        if self.clean:
          command = "rm", image.replace(".fits","_EC???.fits")
          command = join(command)
          self.WriteLine(command)
          self.WriteLine()

    def MatchSky(self,image):
        image1d = image.replace(".fits","specs.fits")
        command = "mikeMatchSky",image1d,"-o",image,"-go"
        command = join(command)
        self.WriteLine(command)
        if self.clean:
           command = "rm", image1d.replace(".fits","_EC???.fits")
           command = join(command)
           self.WriteLine(command)
           self.WriteLine()

    def MatchOrders(self):
        image = self.wdistortion
        command = "mikeMatchOrders",image
        command = join(command)
        self.WriteLine(command)
        self.WriteLine()
        self.wdistortion = image.replace(".fits","specs.fits")

    def Get1DSky(self,image):
        command = "mikeGet1DOrders",image
        command = join(command)
        self.WriteLine(command)

    def FindLines(self,image,fwhm=10.0,th=5):
        if image != self.wdistortion: image = image.replace(".fits","specs.fits")
        command = "mikeFindLines",image
        command += "-fwhm %f" % (fwhm/self.binning), "-th %f" % (th)
        command = join(command)
        self.WriteLine(command)
        if image == self.wdistortion: self.WriteLine()

    def GlobalXDistortion(self,maxo=3,s=8):
        image = self.xdistortionb
        command = "globalrect",image
        command += "-maxo %d" % (maxo),"-go","-s %f" % (s),"-d %s" % (image.replace(".fits","glob.ps/vcps"))
        command = join(command)
        self.WriteLine(command)

    def MapYDistortion(self,s=8):
        image = self.ydistortion
        command = "maprect","-ydist",image
        command = join(command)
        self.WriteLine(command)

    def MapXDistortion(self,s=8):
        image = self.xdistortionb
        command = "maprect","-xdist",image
        command = join(command)
        self.WriteLine(command)

    def NormFlat(self,n):
        image = self.images[n]
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        rn = F["enoise"]
        g = F["egain"]
        F.close()
        command = "procMilky",image
        command = join(command)
        self.WriteLine(command)

    def slowNormFlat(self,n,kx=3,ky=3,b=0,dkx=16,pct=85,skycut=10,check=0,spass=4,extend=2):
        image = self.images[n]
        command = "modrect",image
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        rn = F["enoise"]
        g = F["egain"]
        binning = int(F["binning"].split("x")[1])
        F.close()
        skyint = 4096/binning
        dkx = dkx/binning
        command += "-xdist","-ydist","-rn %f" % (rn), "-g %f" % (g)
        command += "-extend %d" % (extend),
        command += "-skyint %d" % (skyint), "-kx %d" % (kx), "-ky %d" % (ky)
        command += "-goofy 0","-b %d" % (b), "-dkx %f" % (dkx), "-pct %f" % (pct)
        command += "-skycut %f" % (skycut), "-check %d" % (check), "-spass %d" % (spass)
        command = join(command)
        self.WriteLine(command)
        flm = image.replace(".fits","m.fits")
        flp = image.replace(".fits","p.fits")
        flt = flp.replace(".fits","t.fits")
        flu = flt.replace(".fits","u.fits")
        flg = flt.replace(".fits","g.fits")
        fly = flg.replace(".fits","d.fits")
        p2p = image.replace(".fits","flt.fits")
        command = "efits",image,flm,
        command += "\"divz(i1,i2)\"",flp
        command = join(command)
        self.WriteLine(command)
        command = "unrect","-ydist",flp
        command = join(command)
        self.WriteLine(command)
        command = "efits",flt,
        command += "\"1.0*between(i1,0.7,1.3)\"",flu
        command = join(command)
        self.WriteLine(command)
        command = "efits",flt,flu,
        command += "\"divz(VTKGauss(i1*i2,numret=1,sigmax=50,sigmay=0),VTKGauss(i2,numret=1,sigmax=50,sigmay=0))\"",flg
        command = join(command)
        self.WriteLine(command)
        command = "rerect","-ydist",flg
        command = join(command)
        self.WriteLine(command)
        command = "efits",flp,fly,
        command += "\"divz(i1,i2)\"",p2p
        command = join(command)
        self.WriteLine(command)

    def CombineFlats(self,suffix):
        finalflat= "%s_%s_%s.fits" % (self.object,self.Prefix,suffix)
        n = self.iflats[0]
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        rn = F["enoise"]
        g = F["egain"]
        F.close()
        allimages = []
        simages = ["i%d" % (j+1) for j in range(len(self.iflats))]
        limages = join(simages,",")
        limages1 = "i%d" % (len(self.iflats)+1)
        limages2 = simages[-1]
        for n in self.iflats: allimages += self.images[n].replace(".fits","%s.fits" % (suffix)),
        if suffix == "flt":
            medflat= "%s_%s_med%s.fits" % (self.object,self.Prefix,suffix)
            command = ["efits","-p"]
            command += allimages
            command += ["\'median([%s])\'" % (limages), finalflat]
            command = join(command)
            self.WriteLine(command)
            rmsflat= "%s_%s_%s.fits" % (self.object,self.Prefix,"rms")
            command = ["efits","-p"]
            command += allimages
            command += [finalflat,"\'1.49*median(abs(subtract([%s],%s)))\'" % (limages,limages1)]
            command += [rmsflat]
            command = join(command)
            self.WriteLine(command)
#            command = ["efits","-r"]
#            command += allimages
#            command += [ medflat,rmsflat]
#            command += ["\'wt=sqrt(exptimes[:-2,NewAxis,NewAxis])*less((allimages[:-2]-allimages[-2])**2,5*allimages[-1]);"]
#            command += ["result=divz(sum(wt*allimages[:-2]),sum(wt))\'"]
#            command += [finalflat]
#            command = join(command)
#            self.WriteLine(command)
        else:
            command = ["efits","-p"]
            command += allimages
            command += ["\'mean(allimages)\'", finalflat]
            command = join(command)
            self.WriteLine(command)
        if suffix == "flt": self.flat2d = finalflat
        elif suffix == "slt": self.flat1d = finalflat
        elif suffix == "blz": self.blaze = finalflat

    def FlatError(self,suffix):
        if suffix == "flt":   finalflat = self.flat2d
        elif suffix == "slt": finalflat = self.flat1d
        elif suffix == "blz": finalflat = self.blaze 
        simages = ["i%d" % (j+1) for j in range(len(self.iflats))]
        limages = join(simages,",")
        limages1 = "i%d" % (len(self.iflats)+1)
        limages2 = simages[-1]
        command = "efits",
        for n in self.iflats: command += self.images[n].replace(".fits","%s.fits" % (suffix)),
        command += finalflat,"\'1.49*median(abs(subtract([%s],%s)))\'" % (limages,limages1),
        command += "%s_%s_%s.fits" % (self.object,self.Prefix,"rms"),
        command = join(command)
        self.WriteLine(command)

    def Flatten(self,n):
        command = "flat2d",self.flat2d,self.images[n]
        command = join(command)
        self.WriteLine(command)
#        if self.clean:
#           command = "rm",self.images[n]
#           command = join(command)
#           self.WriteLine(command)
        self.images[n] = self.images[n].replace(".fits","f.fits")

    def SkyAps(self,nsig=2.0,oo=5):
        command = "mkorderaps","-ydist","-nsig %f" % (nsig),"-oo %d" % (oo),
        for n in self.ispecs: command += self.images[n],
        command += "-maskout",
        command = join(command)
        self.WriteLine(command)

    def SkyModel(self,n):
        image = self.images[n]
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        rn = F["enoise"]
        g = F["egain"]
        sbinning = int(F["binning"].split("x")[0])
        binning = int(F["binning"].split("x")[1])
        F.close()
        skyint = 4096/binning
        if "r" in self.Prefix.lower():
           kx, ky, dkx = 3, 0, 1.0
        else:
           kx, ky, dkx = 3, 0, 1.0
        maskimage = image.replace(".fits","l.fits")
        pct = 40
        ncut = 0.0
        b = 2
        check = 0
        skycut = 3
        spass = -5
        command = "skyrect",image
        command += "-xdist","-ydist","-rn %f" % (rn), "-g %f" % (g)
        command += "-mask %s" % (maskimage),
        command += "-skyint %d" % (skyint), "-kx %d" % (kx), "-ky %d" % (ky)
        command += "-b %d" % (b), "-dkx %f" % (dkx),
        command += "-skycut %f" % (skycut), "-spass %d" % (spass)
        command = join(command)
        self.WriteLine(command)
        if self.clean:
           command = "rm",maskimage,
           command = join(command)
           self.WriteLine(command)

    def NewMask(self,n,jfit=0):
        image = self.images[n]
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        rn = F["enoise"]
        g = F["egain"]
        sbinning = int(F["binning"].split("x")[0])
        binning = int(F["binning"].split("x")[1])
        F.close()
        th = 10.0
        usemod = image.replace(".fits","v.fits")

        command = "efits",image,"\'VTKMedian(i1,5,5,numret=1)\'", usemod
        command = join(command)
        self.WriteLine(command)

        empnoise = image.replace(".fits","n.fits")
        command = "efits",image,usemod,"\'1.49*VTKMedian(abs(i1-i2),5,5,numret=1)\'", empnoise
        command = join(command)
        self.WriteLine(command)

        oldmask = image.replace(".fits","k.fits")
        sigmask = image.replace(".fits","z.fits")
        newmask = image.replace(".fits","l.fits")
        usemask = image.replace(".fits","u.fits")

        command = "efits",image,usemod,empnoise,"\'divz(i1-i2,i3)\'", sigmask
        command = join(command)
        self.WriteLine(command)

        command = "efits","-b 16",sigmask,oldmask,"\'i2*greater(VTKConvolve(less(abs(i1),%f).astype(Float32),k=3,numret=1),0.75)\'" % (th), newmask
        command = join(command)
        self.WriteLine(command)

        command = "efits",image,newmask,"\'i1*i2\'",usemask
        command = join(command)
        self.WriteLine(command)
        self.WriteLine()

        if self.clean or 1:
            command = "rm",oldmask,sigmask,usemask,empnoise,usemod
            command = join(command)
            self.WriteLine(command)
            self.WriteLine()

    def SkySubtract(self,n):
        image = self.images[n]
        skymod = image.replace(".fits","m.fits")
        skysub = image.replace(".fits","o.fits")
        command = "efits",self.images[n],skymod,"\'(i1-i2)\'",skysub
        command = join(command)
        self.WriteLine(command)
        self.images[n] = skysub

    def PosObjects(self,ex=1.7625,oo=5):
        self.apdb = "%s_%s" % (self.object,self.Prefix)
        command = "mkorderaps","-ydist","-bigsky","-oo %d" % (oo),"-ex -nsig %f" % (ex),"-root ",self.apdb,
        for n in self.ispecs: command += self.images[n],
        command = join(command)
        self.WriteLine(command)

    def AllPos(self):
        self.apdb = "%s.aps" % (self.apdb)
        command = "cat",
        for n in self.ispecs: command += self.images[n].replace("fo.fits","fo.aps"),
        command += ">", self.apdb
        command = join(command)
        self.WriteLine(command)

    def SumRect(self,kx=1,ky=1,xint=16,yint=64,fwhm=5,spass=3,mx=9,my=3,ncut=5.0,ccut=5.0):
        n = self.ispecs[0]
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        rn = F["enoise"]
        g = F["egain"]
        binning = int(F["binning"].split("x")[1])
        F.close()
        w1, w2 = 0, 4096/binning
        if self.twodim: command = "sumrect",
        else: command = "multiextrect",
        command += "-quiet","-xdist","-ydist","-wdist","-defsample",self.apdb
        if self.logb: command += "-logb",
        command += "-kx %d -ky %d" % (kx,ky),
        command += "-xint %d -yint %d" % (xint,yint),
        command += "-fwhm %f" % (fwhm), "-spass %d" % (spass),
        command += "-mx %d -my %d" % (mx,my), "-ncut %f" % (ncut), "-ccut %f" % (ccut),
        command += "-rn %f -g %f" % (rn,g),
        command += "-extra %s" % (self.images[self.ixdists[0]]),
        command = join(command)
        self.WriteLine(command)

    def Unrect(self,fwhm=4,spass=1,mx=15,my=1,ncut=5.0,ccut=0.0,xsub=1,ysub=3):
        n = self.ispecs[0]
        oimage = "%s/%s%s.fits" % (self.information["directory"][n],self.Prefix,self.information["image"][n])
        F = FITS(oimage,readonly=1)
        rn = F["enoise"]
        g = F["egain"]
        binning = int(F["binning"].split("x")[1])
        F.close()
        if len(self.ispecs) == 1:
           spass = 1
           ccut = ncut = 0.0
           mx = my = 1
        command = "multiunrect",
        command += "-xdist","-ydist","-wdist","-defsample",self.apdb
        if self.logb: command += "-logb",
        command += "-fwhm %f" % (fwhm), "-spass %d" % (spass),
        command += "-mx %d -my %d" % (mx,my), "-ncut %f" % (ncut), "-ccut %f" % (ccut),
        command += "-rn %f -g %f" % (rn,g),
        command += "-xsub %d -ysub %d" % (xsub,ysub),
        command += "-extra %s" % (self.images[self.ixdists[0]]),
        if len(self.ispecs) == 1: command += "-uniform",
        command = join(command)
        self.WriteLine(command)

    def CollateRect(self):
        command = "mikejoin","%s_%s???sum.fits" % (self.object,self.Prefix)
        command += "-r","-o %s_%s_sum.fits" % (self.object,self.Prefix),
        command = join(command)
        self.WriteLine(command)
        command = "mikejoin","%s_%s???sumsky.fits" % (self.object,self.Prefix)
        command += "-r","-o %s_%s_sky.fits" % (self.object,self.Prefix),
        command = join(command)
        self.WriteLine(command)
        command = "mikejoin","%s_%s???sumsig.fits" % (self.object,self.Prefix)
        command += "-r","-o %s_%s_sig.fits" % (self.object,self.Prefix),
        command = join(command)
        self.WriteLine(command)
        command = "mikejoin","%s_%s???sumext.fits" % (self.object,self.Prefix)
        command += "-r","-o %s_%s_ext.fits" % (self.object,self.Prefix),
        command = join(command)
        self.WriteLine(command)

    def MultiExtract(self,nsig=2.0):
        command = "extractmultispec","%s_%s???sum.fits" % (self.object,self.Prefix)
        command += "-extra","-nsig %f" % (nsig),"-o %s_%s_multispec.fits" % (self.object,self.Prefix),
        command += "-d %s_%s_multispec.ps/vcps" % (self.object,self.Prefix),
        if self.clean: command += "-clean",
        command = join(command)
        self.WriteLine(command)

